/***************************************************************************
                          components.h  -  description
                             -------------------

    functionality        : for constructing and handling component matrices
    relevant classes     : component_matrix
    relevant functions   : get_partition

    begin                : Thu Dec 6 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef COMPONENTS_H
#define COMPONENTS_H

#include "rules.h"

class and_tree;
//class condition;


inline int   get_partition(condition cond)
 {
  if (cond.Operator== EMPTY) return -1;
  if  (cond.Operator== CONST_EQ) return 0;
if  (cond.Operator== CONST_NEQ) return 0;
if  (cond.Operator== VAR_EQ)  return 1;
if  (cond.Operator== VAR_NEQ) return 1;
if  (cond.Operator== NUM_LT)  return 2;
if  (cond.Operator== NUM_GE) return 2;
if  (cond.Operator== NUM_LT_WU)  return 2;
if  (cond.Operator== NUM_GE_WU) return 2;
if  (cond.Operator== NUM_MISSING)  return 2;
return -1;
}

class component_matrix {
private:
  int NumAttributes;
  int NumSplits;
  vector<double > V;  // With vector there would be no need for copy constructor of assignment operator.

public:
  component_matrix() {
  }

  component_matrix(int n,int m, double *Source)  :  V(Source, Source + n*m)  { // for valarray V(Source, n*n) {
    NumAttributes = n;
    NumSplits =m;
  }

  component_matrix(int n, int m,double Value= 0.0)  :  V(n*m, Value)  { // for valarray V(Value, n*n) {
    NumAttributes = n;
    NumSplits =m;
  }

  double & AccessRP(int a, int b) {
    return V[b*NumAttributes+ a];
  }

  const double & AccessRP(int a, int b)  const {  // The same as before
    return (const double &)V[b*NumAttributes + a];
  }

  void fill_matrix_cond(int ActualArg, condition ActualCond, int depth);

  void fill_matrix_rule(rule Rule, int depth);

  void Add(const component_matrix &C);

  // puts the matrix to 0
  void Reset();

  // puts the f% of elements to 1 and the rest to 0 randomly
  void Randomise1(float f);

  void fill_matrix(vector <and_tree*> nodes);

  void Show() ;

  double EuclideanDistance(const component_matrix &m2);

   double ManhattanDistance(const component_matrix &m2);

};
#endif

