/***************************************************************************
                          criteria.cpp  -  description
                             -------------------

    functionality        : for computing some criteria and defines structures related to these options
    relevant classes     : split_partition, split_selection_criteria ...
    relevant functions   :

    begin                : Tue Oct 23 2001
    copyright            : (C) 2001 by C. Ferri, Jos Hernndez-Orallo
    email                : {cferri, jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "trees.h"
#include "criteria.h"
#include "utils.h"
#include "cost.h"
#include "roc.h"

#include <vector>
// #define TREE_DEBUG

// Computes the different criteria according to the values: SSC, ...
void Compute_Criteria_AND(and_tree &AT, const dataset &Train, const options &Options ) {
      int nC= Train.NumClasses();

     float Card= AT.Cardinality;

    vector<double> StratifiedDistribution;

    if (Card!= 0) {
       if ((Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION) ||
           (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION_NO_SMOOTHING)) {
        vector<double> Dist(nC);
        double Acc= 0;
         for(int j=0; j < nC; j++) {
           double jValue;
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION)
              jValue= AT.SmoothDistribution[j];
            else
              jValue= AT.ClassDistribution[j];
           Dist[j]= jValue * Options.WeightVector[j];
           Acc+= Dist[j];
         }
         if (Acc) {
           for(int j=0; j < nC; j++) {  // Normalitzar
             Dist[j]/=Acc;
           }
         }
        StratifiedDistribution= Dist;
      }
    }

    if ((Options.SSC == GAIN) || (Options.SSC == GAIN_RATIO) || (Options.SSC == C45))  {
      float IE = 0;

      if (Card != 0) {
         for(int j=0; j < nC; j++) {
          float pEj;
          if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
           pEj= (float)AT.SmoothDistribution[j];
          else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
           pEj= (float)AT.ClassDistribution[j] / Card;
          else if ((Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION) ||
                    (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION_NO_SMOOTHING))
           pEj= StratifiedDistribution[j];
          else if (Options.ProbabilityInSplittingCriteria == FROM_COSTS)
            pEj= (float)AT.CostProb[j];
          if (pEj != 0.0)
            IE+= pEj * log2(pEj);
        }
      }
      AT.Crit.Gain.InfoE = -IE;
        #ifdef TREE_DEBUG
          cout << "InfoE of Node AND #" << AT.Num << " computed: " << AT.Crit.Gain.InfoE << endl;
        #endif

      if  ((Options.SSC == GAIN_RATIO) || (Options.SSC == C45)) {
        // nothing special to do
      }
    }
    else if ((Options.SSC == CART) || (Options.SSC == DKM)) {
      float GINI = 1;
      if (Card != 0) {
         for(int j=0; j < nC; j++) {
          float pEj;
          if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
           pEj= (float)AT.SmoothDistribution[j];
          else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
           pEj= (float)AT.ClassDistribution[j] / (float)Card;
           else if ((Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION) ||
                 (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_WITH_STRATIFICATION_NO_SMOOTHING))
           pEj= StratifiedDistribution[j];
           else if (Options.ProbabilityInSplittingCriteria == FROM_COSTS) {
            pEj= (float)AT.CostProb[j];
          }
          GINI+= Sqr(pEj);
        }
        GINI= 1 - GINI;
      }
     if (Options.SSC == CART) {
      AT.Crit.Cart.GiniE = GINI;
     }
     else { // if (Options.SSC == DKM) {
       AT.Crit.Cart.GiniE = pow(GINI, 1/float(nC));
      }

        #ifdef TREE_DEBUG
          cout << "GiniE of Node AND #" << AT.Num << " computed: " << AT.Crit.Cart.GiniE << endl;
        #endif
    }

    else if (Options.SSC==DESC_MDL)	{
   	
     	unsigned int j,k;
 	    indext h;
 	    vector<vindex> tix ;
      for (j=0;j<Train.num_arguments;j++)  // atenci que axi els arguments tamb tenen
   																		 //la classe
       {
         k=Train.exetable[j].size(); // numero de indexs diferents per a eixe arg o classe
     	  vindex vix =vindex(k);
       	tix.push_back(vix);
       }

	    for (h=0;h<AT.Covered.size();h++)// atenci que aci els arguments tamb tenen la classe
  		  for (j=0;j<Train.num_arguments;j++)
  		  {if (Train.KindOf(j)==DISCRETE_KIND)
			  tix[j].set_index(Train.Nth_arg(AT.Covered[h],j).Value_d);}
         AT.Crit.GMDL.cost=compute_cost(Train,tix,Train.exetable,AT.Cardinality);

         // I QUE PASSA AMB NUMERIC_KIND I IGNORE_KIND
    }
    else if (Options.SSC==SPLIT_EXPECTED_ERROR) {
     // Nothing to do
    }
    else if (Options.SSC==LOCAL_ROC_AREA) {
     // Nothing to do
    }
    else if (Options.SSC==MSE_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==LOGLOSS_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==AUCH_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==AUCS_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==MGINI_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==GENENTROPY_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==SQDIF_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==ROCV_SPLIT) {
     // Nothing to do
    }
    else if (Options.SSC==ONE_POINT_LOCAL_ROC_AREA) {
     // Nothing to do
    }
    else {
     cout << "Unknown Split criterion in function Compute_Criteria_AND" << endl;
   }
 }



// #define SMOOTH(a,b,c)   ((c)?(((a)+1)/(b+1)):(((a))/(b)))


inline double SMOOTH(double a, double b, double nNodes, const options &Options) {
  if (Options.NodeSmoothing.Method == NO_SMOOTHING) {
    return a/b;
  }
  else if (Options.NodeSmoothing.Method == LAPLACE) {
    return (a+1)/(b+nNodes);  // Authentic Laplace Smoothing
  }
  else if (Options.NodeSmoothing.Method == K_ESTIMATE) {
    double k= Options.NodeSmoothing.m;
   // k= 1000;
    return (a+k)/(b+k);  // This is a smoothing not generally used
  }
  else if ((Options.NodeSmoothing.Method == M_ESTIMATE) || (Options.NodeSmoothing.Method == M_ESTIMATE_UNIFORM)) {
    double m= Options.NodeSmoothing.m;
   // m= 1000;
    return (a+m*(1/nNodes)) / (b+m);
  }
  else {
    cout << "Node Smoothing Method Unknown" << endl;
    return 0;
  }
}


// Computes different criteria according to values: SSC, ...
void Compute_Criteria_OR(or_tree &OT, const dataset &Train, const options &Options ) {
/*    if (SSC == LEFT_FIRST) {
      Split_Optimality=
    }
*/

    if ((Options.SSC == GAIN) || (Options.SSC == GAIN_RATIO) || (Options.SSC == C45)) {
      float G = ((and_tree *)OT.Parent)->Crit.Gain.InfoE;
      indext Card = ((and_tree *)OT.Parent)->Cardinality;
      float AG= 0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
      if (Card != 0) {
        for(It= OT.Children.begin(); It != ItEnd; It++) {
          and_tree *Child= (and_tree *)*It;
          #ifdef TREE_DEBUG
          if (Child == NULL)  {
            cout << "Child NULL en Compute_Criteria de or_node";
          }
          #endif
         AG+= SMOOTH(Child->Cardinality, (float)Card, OT.Children.size(), Options) * Child->Crit.Gain.InfoE;
          // AG+= (Child->Cardinality / (float)Card) * Child->Crit.Gain.InfoE;
        }
      }
      OT.Crit.Gain.GainET = G - AG;
      if (Options.SSC == GAIN) {
         OT.Split_Optimality = OT.Crit.Gain.GainET;
      }
      else if  ((Options.SSC == GAIN_RATIO) || (Options.SSC == C45)) {
        float AS= 0;
        if (Card != 0) {
          ItEnd= OT.Children.end();
          for(It= OT.Children.begin(); It != ItEnd; It++) {
            and_tree *Child= (and_tree *)*It;
            #ifdef TREE_DEBUG
            if (Child == NULL)  {
              cout << "Child NULL en Compute_Criteria de or_node";
            }
            #endif
            // float p= Child->Cardinality / (float)Card;
            float p= SMOOTH(Child->Cardinality, (float)Card, OT.Children.size(), Options);
            if (p!=0) {
              AS+= p * log2(p);
            }
          }
        }
        OT.Crit.Gain.SplitET = -AS;
        if (AS != 0) {
          OT.Crit.Gain.GainRatioET = OT.Crit.Gain.GainET / OT.Crit.Gain.SplitET;
        }
        else {
          OT.Crit.Gain.GainRatioET = OT.Crit.Gain.GainET;
        }
        OT.Split_Optimality = OT.Crit.Gain.GainRatioET;
      }
    }
    else if ((Options.SSC == CART)  || (Options.SSC == DKM)) {
      float G = ((and_tree *)OT.Parent)->Crit.Cart.GiniE;
      indext Card = ((and_tree *)OT.Parent)->Cardinality;
      float AG= 0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
      if (Card != 0) {
        for(It= OT.Children.begin(); It != ItEnd; It++) {
          and_tree *Child= (and_tree *)*It;
          #ifdef TREE_DEBUG
          if (Child == NULL)  {
            cout << "Child NULL en Compute_Criteria de or_node";
          }
          #endif
          // float p= Child->Cardinality / (float)Card;
          double p= SMOOTH(Child->Cardinality, (float)Card, OT.Children.size(), Options);
          AG+= p * Child->Crit.Cart.GiniE;
        }
      }
      OT.Crit.Cart.GainET = G - AG;
      OT.Split_Optimality = OT.Crit.Cart.GainET;
    }
    else if (Options.SSC==DESC_MDL) {
	
	    float AC=0;
      AC+= calcula_partv(((and_tree *)OT.Parent)->OpenVariables);
	    list<tree *>::iterator It, ItEnd;
	    indext Card = ((and_tree *)OT.Parent)->Cardinality;
      ItEnd= OT.Children.end();
      if (Card != 0) {
        for(It= OT.Children.begin(); It != ItEnd; It++) {
          and_tree *Child= (and_tree *)*It;
          AC+= (Child->Crit.GMDL.cost);
        }
        OT.Crit.GMDL.cost = AC;
        OT.Split_Optimality =1/AC;
      }
	  }
    else if (Options.SSC==SPLIT_EXPECTED_ERROR) {
      OT.Split_Optimality= -OT.BackupError;
    }	
    else if (Options.SSC==LOCAL_ROC_AREA) {
       double Area;

       vector<and_tree *> Vector;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }


       vector<vector<double> > Points;
       compute_ROC_points(Vector, Points, ROC_WITH_TRAIN, Options,0,1);
       compute_ROC_AREA(Points, Area);
   //    cout << "ROC Area: " << Area << endl;
       OT.Split_Optimality= Area;
    }
/*  ESTE ES EL L'ANTIC MSE QUE NO PERMET SMOOTHING
    else if (Options.SSC==MSE_SPLIT) {
      double MSE1;

      vector<and_tree *> Vector;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

      MSE1= compute_MSE(Vector, Options, MSE_WITH_TRAIN);
      MSE1= MSE1/Train.NumClasses();
      MSE1= 1 - MSE1;

      OT.Split_Optimality= MSE1;
    }
*/
    else if (Options.SSC==MSE_SPLIT) {
      double MSE=0;
      int nC= Train.NumClasses();
      double TotalTestValue= 0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;

        long Card= Child->Cardinality;
        double MSE0= 0;

        for(int j=0; j<nC; j++) {
          long TestValue;
          TestValue= Child->ClassDistribution[j];
/*
          TestValue++; // DOUBLE SMOOTHING!!
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card + nC);
*/
/*
          const int K= 1;
          TestValue+= K; // K SMOOTHING!!
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card + K);
*/
          // WITHOUT DOUBLE SMOOTHING
          double pTestValue=0.0;
          if (Card > 0)
            pTestValue = TestValue / double(Card);

          for(int k=0; k<nC; k++) {

            double ProbTrain;
            if (Card == 0) {  // We assume equal probability for all the classes
              ProbTrain = 1 / double(nC);
            }
            else {
              if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
                ProbTrain= (float)Child->SmoothDistribution[k];
              else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
                ProbTrain= (float)Child->ClassDistribution[k] / (float)Card;
            }

            if (j!=k) {
              MSE0+= pTestValue * Sqr(0-ProbTrain);
            }
            else {
              MSE0+= pTestValue * Sqr(1-ProbTrain);
            }
          }
          TotalTestValue+= TestValue;
        }

        // double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
         double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);

        MSE+= pNode * MSE0;

      }

      MSE/= /* TotalTestValue * */ nC;

      OT.Split_Optimality= 1 - MSE;

    }
    else if (Options.SSC==LOGLOSS_SPLIT) {
      double MSE;

      vector<and_tree *> Vector;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

      MSE= compute_LogLoss(Vector, Options, MSE_WITH_TRAIN);
      // MSE= MSE/2;
      MSE= 1 - MSE;

      OT.Split_Optimality= MSE;
    }
    else if (Options.SSC==AUCH_SPLIT) {
      double AUCH;

      vector<and_tree *> Vector;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

      AUCH=compute_AUC_by_Hand(Vector,ROC_WITH_TRAIN,Options);

      OT.Split_Optimality= AUCH;
    }
    else if (Options.SSC==AUCS_SPLIT) {
      double AUCS;

      vector<and_tree *> Vector;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        Vector.push_back(Child);
      }

      AUCS=compute_AUC_by_Srinivasan(Vector,ROC_WITH_TRAIN,Options);

      OT.Split_Optimality= AUCS;
    }
    else if (Options.SSC==MGINI_SPLIT) {
      int nC= Train.NumClasses();
      double MGINI=0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> p(nC);
        vector<double> q(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          for(int j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
           // q[j]= (float)Child->ClassDistribution[j] / (float)Card;
            q[j]= p[j];
          }
          //           double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
          double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
           double ValNode= 0;
          for(int j=0; j < nC; j++) {
            for(int k=0; k < nC; k++) {
              if (j!=k) {
                ValNode+= p[j]*q[k];
              }
            }
          }
          MGINI+= pNode * ValNode;
        }
      }

       OT.Split_Optimality= - MGINI;
    }
    else if (Options.SSC==GENENTROPY_SPLIT) {
      int nC= Train.NumClasses();
      double GENENT=0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> p(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          double NodeVal=0;
          for(int j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
       //     NodeVal+= sqrt(p[j]);
            NodeVal+= pow(p[j], 0.5);
          }

       //           double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
        double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);

          GENENT+= pNode * NodeVal;
        }
      }

       OT.Split_Optimality= - GENENT;
    }
    else if (Options.SSC==SQDIF_SPLIT) {
      int nC= Train.NumClasses();
      double SQDIF=0;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        vector<double> p(nC);
        long Card= Child->Cardinality;

        if (Card != 0) {
          for(int j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
          }
          double p0= p[0];
          double p1= p[1];
          //           double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
          double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
           SQDIF+= pNode * Sqr(p0-p1);
        }
      }

       OT.Split_Optimality= SQDIF;
    }
    else if (Options.SSC==ROCV_SPLIT) {
      int nC= Train.NumClasses();
      double GENENT=0;
      vector<double> NCard;
      vector<vector<double> > pmat;

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        long Card= Child->Cardinality;

        vector<double> p(nC);

        if (Card != 0) {
          for(int j=0; j < nC; j++) {
            if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING)
              p[j]= (float)Child->SmoothDistribution[j];
            else if (Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_NO_SMOOTHING)
              p[j]= (float)Child->ClassDistribution[j] / (float)Card;
          }
        }
        //           double pNode= (Card/(double)((and_tree *) OT.Parent)->Cardinality);
        double pNode= SMOOTH(Card, (double)((and_tree *) OT.Parent)->Cardinality, OT.Children.size(), Options);
        NCard.push_back(pNode);

        pmat.push_back(p);
      }

      int NNodes= OT.Children.size();
      for(int k=0; k<NNodes; k++) {
        for(int l=0; l<NNodes; l++) {
          for(int i=0; i<nC; i++) {
            for(int j=0; j<nC; j++) {
              if (i<j) {
                GENENT+= NCard[k]*NCard[l]*pmat[k][i]*pmat[l][j];
              }
            }
          }
        }
      }
/*
      int NNodes= OT.Children.size();
        for(int l=0; l<NNodes; l++) {
          for(int i=0; i<nC; i++) {
            for(int j=0; j<nC; j++) {
              if (i<j) {
                GENENT+= NCard[l]*NCard[l]*pmat[l][i]*pmat[l][j];
              }
            }
          }
        }
*/
      OT.Split_Optimality=  +GENENT;

    }
    else if (Options.SSC==ONE_POINT_LOCAL_ROC_AREA) {
       double Area;

       vector<double> Confus(4);  // ONLY FOR TWO CLASSES. ALTHOUGH IT CAN BE GENERALISED

      list<tree *>::iterator It, ItEnd;
      ItEnd= OT.Children.end();
	    for(It= OT.Children.begin(); It != ItEnd; It++) {
        and_tree *Child= (and_tree *)*It;
        if (Child->Cardinality != 0.0) {
          if (Child->PredictedClass == 0) {
            Confus[0]+= Child->SmoothDistribution[0] * Child->Cardinality;
            Confus[1]+= Child->SmoothDistribution[1] * Child->Cardinality;
          }
          else {
            Confus[2]+= Child->SmoothDistribution[0] * Child->Cardinality;
            Confus[3]+= Child->SmoothDistribution[1] * Child->Cardinality;
          }
        }
      }

       vector<vector<double> > Points;
       vector<double> Point(2);
       Point[0]= 0;
       Point[1]= 0;
       Points.push_back(Point);

       double c0= Confus[1]+Confus[3];
       double c1= Confus[0]+Confus[2];

       if (c0 == 0)
         Point[0]= 0.5;  // this never happens
       else
         Point[0]= Confus[1] / c0;

       if (c1 == 0)
         Point[1]= 0.5;  // this never happens
       else
         Point[1]= Confus[0] / c1;
       Points.push_back(Point);

       Point[0]= 1.0;
       Point[1]= 1.0;
       Points.push_back(Point);


//       compute_ROC_points(Vector, Points, ROC_WITH_TRAIN, Options);
       compute_ROC_AREA(Points, Area);
   //    cout << "ROC Area: " << Area << endl;
       OT.Split_Optimality= Area;
    }

    else {
      cout << "Unknown Splitting criterion in function Compute_Criteria_OR" << endl;
    }


  if (Options.Components_In_Split_Criterion.Met == ADAPTATIVE_COMPONENT) {
     //     Options.Components.Show();
     //     Options.AccComponents.Show();
    // cout << "En este moment s'haurien de fer les coses b FISTRO" << endl;

     and_tree *a1Pa;
     a1Pa= (and_tree*)(OT.Children.front());

     float v1= Options.AccComponents.AccessRP(a1Pa->ActualArg, get_partition(a1Pa->ActualCond));

/*
     if (v1 == 0)
       cout << "matrix zeros FISTRO en criteria.cpp" << endl;
*/

     float f1= pow(2, - v1 * Options.Components_In_Split_Criterion.Factor);

     OT.Split_Optimality= OT.Split_Optimality * f1;
  }
  else if (Options.Components_In_Split_Criterion.Met == RANDOM_COMPONENTS) {
     //     Options.Components.Show();
     //     Options.AccComponents.Show();
   //  cout << "En este moment s'haurien de fer les coses b FISTRO" << endl;

     and_tree *a1Pa;
     a1Pa= (and_tree*)(OT.Children.front());

     float v1= Options.RandomComponents.AccessRP(a1Pa->ActualArg, get_partition(a1Pa->ActualCond));

/*
     if (v1 == 0)
       cout << "matrix zeros FISTRO en criteria.cpp" << endl;
*/

     float f1= pow(2, - v1 * Options.Components_In_Split_Criterion.Factor);
/*
     if (f1 != 1)
       cout << "iep" << endl;
*/
     OT.Split_Optimality= OT.Split_Optimality * f1;
  }



  if (Options.TestCostMethod != NO_TEST_COSTS) {

    if (Options.TestCostUse == TEST_COSTS_NO_USE) {
      // Nothing to do
    }
    else {
      and_tree *a1;
      a1= (and_tree*)(OT.Children.front());
      int Arg= a1->ActualArg;
      double TestCost= Options.TestCostVector[Arg];
      if (Options.TestCostUse == TEST_COSTS_LINEAR_PLUS1_WITHOUT_REPETITION) {
        or_tree *o2 = &OT;
        while(true) {
          and_tree *a2= (and_tree *)o2->Parent;

          if (a2->Parent == NULL)  // If it is the root. bye. The root does not have condition.
            break;

          if (a2->ActualArg == Arg) {  // this argument has already been used. We don't have to take it into account.
            TestCost= 0;
            break;
          }

          o2= (or_tree *)a2->Parent;
        }
      }


      // else if (Options.TestCostUse == TEST_COSTS_LINEAR_PLUS1_WITH_REPETITION) {
        // NO CONTROLLEM SI EIXE ARGUMENT JA S'HA UTILITZAT DALT
      double Denom= pow(TestCost + 1, Options.TestCostRelevanceInSplitting);

      OT.Split_Optimality= OT.Split_Optimality / Denom;

      // }
    }
  }





    #ifdef TREE_DEBUG
       cout << "Optimality of Node OR #" << OT.Num << " computed: " << OT.Split_Optimality << endl;
    #endif

 }


