/***************************************************************************
                          evaluate.cpp  -  description
                             -------------------

    functionality        : for the statistics for evaluating trees including cost and confusion matrix
    relevant classes     : class_matrix, evaluation_statistics
    relevant functions   :

    begin                : Tue Nov 20 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#include "evaluate.h"
#include "options.h"
#include "exemples.h"

#include <stdio.h>

void evaluation_statistics::Show(const options &Options) {
  int c;
  cout << endl << "Statistics over test set of length: " << TestCardinality << ":"  << endl;


  if ((Options.Output.ShowStatistics == BOTH_STATISTICS)
      || (Options.Output.ShowStatistics == ABSOLUTE_STATISTICS)) {
     cout << endl;
     cout << "Absolute Results:" << endl;
     for(c=0; c < NumClasses; c++) {
       cout << "TP" << c << ": " << AbsTruePositives[c] << ", ";
     }
     cout << endl;
     for(c=0; c < NumClasses; c++) {
       cout << "FP" << c << ": " << AbsFalsePositives[c] << ", ";
     }
     cout << endl;

     cout << "Absolute Accuracy: " << AbsAccuracy << endl;
  }

  if ((Options.Output.ShowStatistics == BOTH_STATISTICS)
      || (Options.Output.ShowStatistics == RELATIVE_STATISTICS)) {
     cout << endl;
     cout << "Relative Results:" << endl;
     for(c=0; c < NumClasses; c++) {
       cout << "TP" << c << ": " << TruePositives[c] << ", ";
     }
     cout << endl;
     for(c=0; c < NumClasses; c++) {
       cout << "FP" << c << ": " << FalsePositives[c] << ", ";
     }
     cout << endl;
  }

  cout << "Relative Accuracy: " << Accuracy << endl;
 		



   if ((Options.Output.ShowConfusionMatrix == SHOW_CONFUSION_MATRIX) ||
       ( (Options.Output.ShowConfusionMatrix == SHOW_ONLY_IF_COSTS) &&
         (Options.Weights.Method != NO_COSTS) )
      )
   {
 		 cout << endl << "Confusion Matrix" << endl;
    	 	ConfusionMatrix.Show();
     cout << endl;
   }
    	 	
    	 	
    	 	  	 	
    	 	

  if (Options.Weights.Method != NO_COSTS)  {

  	  cout << "Cost Matrix" << endl;
    		 CostMatrix.Show();

     	cout << endl << "Resulting Matrix" << endl;
     	ResultMatrix.Show();

 		 cout << endl << "Total Cost:" << TotalCost << endl;
 	}
 	
 	
 	if (AUC != -1) {
    cout << endl << "AUC (example by example) in Hand & Till's way: " << AUC << endl;
 	}

  if (Options.TestCostMethod != NO_TEST_COSTS)  {
     cout << endl << "Total Test Cost: " << TestCost << endl;
     cout << "Test Cost per Example: " << TestCostPerExample << endl;
  }

}




void evaluation_statistics::Evaluate(const dataset & Test, const vector<discrete> &True_Results,
                                     const vector<discrete> & Results, options &Options,
                                     const vector<unsigned long> &ArgumentUseVector ) {
    int NClasses= Test.NumClasses();

    TestCardinality= Test.Cardinality();
    if (TestCardinality == 0)
      return; // No cardinality. What are we doing here?

    indext k;
    for(k=0; k < Test.Cardinality(); k++) {
      discrete Pred, Real;
      Pred= Results[k];
      Real= True_Results[k];
      if (Real == Pred) {
         AbsAccuracy++;
         AbsTruePositives[Real]++;
      }
      else {
         AbsFalsePositives[Real]++;
      }
      ConfusionMatrix.AccessRP(Real,Pred)++;
    }

    PredCardinality= k;  // TOTS COBERTS
    for(int c=0; c < NClasses; c++) {
       TruePositives[c] = AbsTruePositives[c] / float(TestCardinality);
       FalsePositives[c] = AbsFalsePositives[c] / float(TestCardinality);
    }
    Accuracy = AbsAccuracy / float(TestCardinality);
    if (Options.Weights.Method!= NO_COSTS) {
    	ResultMatrix = ConfusionMatrix * CostMatrix;
   		TotalCost= ResultMatrix.Sum();
 		}

 		TestCost= 0;

    if (Options.TestCostMethod != NO_TEST_COSTS)  {
    	for(int a=0; a < Test.NumArguments(); a++) {
 		    TestCost+= Options.TestCostVector[a] * ArgumentUseVector[a];
 		  }
 		  TestCostPerExample = TestCost / double(Test.Cardinality());
 		}
 		
}


#define EXAMPLESIZE 8000
#define COMENTARI '%'



void llig_matriu_costs(const string & Matrixfile, class_matrix & C) {

 FILE *in=NULL;
// char filein[800];
 char exemple[EXAMPLESIZE];
// word i=1;
// bool iswe=false;
 //strcpy (filein,fitxer);
  int NClasses= C.NumClasses;

  long Tam= NClasses*NClasses;
//  vector<double> Valors(Tam);
  long i= 0;
	
  if ((in=fopen(Matrixfile.c_str(),"rt"))==NULL) {
    cout << "Error opening cost file: \"" << Matrixfile << "\"" << endl;
    cout << "If test dataset is not used, please use syntax as follows: " << endl;
    cout << "./smiles samples/mydataset.all - samples/mydataset.cost" << endl;
    cout << "If you don't want to use a cost file, disable the option" << endl;
    exit(-1);
  }
	
	bool Eixir= false;

   // tracta_exemple(exemple);
  while ((!Eixir) && (fgets((char *) exemple,EXAMPLESIZE,in)))  // agarra una linia en "exemple"
    {
      exemple[strlen(exemple)-1]='\0'; // eliminem /n
      if (exemple[strlen(exemple)-1]=='\r')
        exemple[strlen(exemple)-1]='\0';// eliminem /r si hi ha
      	
      if (exemple[0]!=COMENTARI) {


        char *aux,buf[EXAMPLESIZE],*tok;
        // printf ("cadena %s",exemple);
        strcpy(buf,exemple);
         aux=buf;
        //printf ("\n cadena %s",exemple);
         while ((!Eixir) && ((tok=strtok(aux,","))!=NULL))
         {
     	
     	      C.V[i]= atof(tok);
	
          	aux=NULL;

            i++;
            if (i>=Tam)
              Eixir= true;
         }
       }
    }
 fclose(in);


}



void llig_vector_testcosts(const string & TestCostfile, vector<double> &V) {

  FILE *in=NULL;
 char exemple[EXAMPLESIZE];

  long Tam= V.size();
  long i= 0;
	
  if ((in=fopen(TestCostfile.c_str(),"rt"))==NULL) {
    cout << "Error opening testcost file: \"" << TestCostfile << "\"" << endl;
    cout << "If either the test dataset or the cost matrix is not used, please use syntax as follows: " << endl;
    cout << "./smiles samples/mydataset.all - - samples/mydataset.testcost" << endl;
    cout << "If you don't want to use a testcost file, disable the option" << endl;
    exit(-1);
  }
	
	bool Eixir= false;

   // tracta_exemple(exemple);
  while ((!Eixir) && (fgets((char *) exemple,EXAMPLESIZE,in)))  // agarra una linia en "exemple"
    {
      exemple[strlen(exemple)-1]='\0'; // eliminem /n
      if (exemple[strlen(exemple)-1]=='\r')
        exemple[strlen(exemple)-1]='\0';// eliminem /r si hi ha
      	
      if (exemple[0]!=COMENTARI) {


        char *aux,buf[EXAMPLESIZE],*tok;
        // printf ("cadena %s",exemple);
        strcpy(buf,exemple);
         aux=buf;
        //printf ("\n cadena %s",exemple);
         while ((!Eixir) && ((tok=strtok(aux,","))!=NULL))
         {
     	
     	      V[i]= atof(tok);
	
          	aux=NULL;

            i++;
            if (i>=Tam)
              Eixir= true;
         }
       }
    }
 fclose(in);


}




double Similarity(const vector<unsigned long> &V, int NClasses, const options &Options) {
   if (Options.Comb2Single.SimilarityMethod == KAPPA) {
     return Kappa(V, NClasses);
   }
   else if (Options.Comb2Single.SimilarityMethod == KAPPA1) {
     return Kappa1(V, NClasses);
   }
   else if (Options.Comb2Single.SimilarityMethod == QSTAT) {
     return QStat(V, NClasses);
   }
   else {
     cout << "ERROR in function Similarity. Similarity Method now known" << endl;
     exit(-1);
   }
}
