/***************************************************************************
                          evaluate.h  -  description
                             -------------------

    functionality        : for the statistics for evaluating trees including cost and confusion matrix
    relevant classes     : class_matrix, evaluation_statistics
    relevant functions   :

    begin                : Nov 10 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef EVALUATE_H
#define EVALUATE_H



#include <iostream>
#include <iomanip>
#include <valarray>

#include "utils.h"
#include "estructures.h"
#include "exemples.h"
#include <vector>


void llig_vector_testcosts(const string & TestCostfile, vector<double> &V);

class options;

class class_matrix {
private:
  int NumClasses;
  vector<double> V;  // With vector there would be no need for copy constructor of assignment operator.

friend void llig_matriu_costs(const string &Matrixfile, class_matrix & C);

public:
//  cost_matrix() {};
/*
  class_matrix(const class_matrix &c) {
  }

  const class matrix & operator = (const class_matrix &c) {
    NumClasses = c.NumClasses;

  }
*/
  class_matrix(int n, double *Source)  :  V(Source, Source + n*n)  { // for valarray V(Source, n*n) {
    NumClasses = n;
  }

  class_matrix(int n, double Value= 0.0)  :  V(n*n, Value)  { // for valarray V(Value, n*n) {
    NumClasses = n;
  }

  class_matrix(int n, vector<numeric> & Source) : V(n*n)  {  // fills the matrix from a weight vector
    NumClasses = n;
    for(int i=0; i < NumClasses; i++) {  // changing row
      for(int j=0; j < NumClasses; j++) {  // fill a column
        if (i==j) {
          V[j*NumClasses + i]= -Source[i];   // the minus sign is because the vector is in weights and the matrix is in costs
        }
        else {
          V[j*NumClasses + i]= Source[i];
        }
      }
    }

  }

  double & AccessRP(int a, int b) {
    return V[b*NumClasses + a];
  }

  const double & AccessRP(int a, int b)  const {  // The same as before
    return (const double &)V[b*NumClasses + a];
  }

    void Show() {
    const int ColW= 5;
    int i;
    cout << "P\\R" << setw(ColW);
    for(i=0; i < NumClasses; i++) {
     cout << "  C" << setw(ColW-3) << i << "|" ;
   }
    for(i=0; i < NumClasses*NumClasses; i++) {
       if ((i % NumClasses) == 0)
          cout << endl << "  C" << setw(ColW-3) << i / NumClasses;
       cout << setw(ColW) << V[i] << "|";
      }
  }

  class_matrix operator * (const class_matrix &m2) {
    class_matrix m1= class_matrix(NumClasses) ;
    for(int i=0; i < NumClasses*NumClasses; i++) {
      m1.V[i] = V[i] * m2.V[i];
    }
    return m1;
  }

  double EuclideanDistance(const class_matrix &m2) {
    double Dist= 0.0;
    for(int i=0; i < NumClasses*NumClasses; i++) {
      Dist += Sqr(V[i] - m2.V[i]);
   }
   return sqrt(Dist);
 }

   double ManhattanDistance(const class_matrix &m2) {
    double Dist= 0.0;
    for(int i=0; i < NumClasses*NumClasses; i++) {
      Dist += abs(V[i] - m2.V[i]);
   }
   return Dist;
 }


  double Sum() {
    double Accum= 0.0;
    for(int i=0; i < NumClasses*NumClasses; i++) {
      Accum+= V[i];
    }
    return Accum;
  }

  double Min() {
     double M= V[0];
     for(int i=1; i < NumClasses*NumClasses; i++) {
       if (V[i] < M)
         M= V[i];
    }
    return M;
  }

  double PartialMagnitude() const {  // computes the module of the matrix, ignoring the diagonal
    double Accum= 0.0;
    for(int j=0; j<NumClasses; j++) {
       for(int i=0; i < NumClasses; i++) {
          if (i!=j)
            Accum+= Sqr(V[j*NumClasses+i]);
       }
    }
    return sqrt(Accum);
  }

  double PartialDotProduct(const class_matrix &m2) const { // computes the dot product of two matrices, ignoring the diagonal
     double Accum= 0.0;
     for(int j=0; j<NumClasses; j++) {
       for(int i=0; i < NumClasses; i++) {
          if (i!=j)
            Accum+= V[j*NumClasses+i] * m2.V[j*NumClasses+i];
       }
    }
    return Accum;
  }

  double PartialAngle(const class_matrix &m2) const { // computes the angle between two matrices, ignoring the diagonal
    double Divisor=(PartialMagnitude() * m2.PartialMagnitude());
    double Numerator=PartialDotProduct(m2) ;
    double Ret=0; // PER FER ALGO
    if (Numerator > Divisor)
      Ret= 0;
    else if (Divisor)
      Ret= acos(Numerator/ Divisor);
    return Ret;
  }

  double Inverse() {
  }

  void Unite() {
    double S=Sum();
    if (S!=0) {
      for(int i=0; i < NumClasses*NumClasses; i++) {
        V[i] /= S;
      }
   }
   else  // CUIDAO
      for(int i=0; i < NumClasses*NumClasses; i++) {
        V[i] = 1 / double(NumClasses*NumClasses);
      }
  }

  void Standardise() {
    double M= Min();
    for(int i=0; i < NumClasses*NumClasses; i++) {
      V[i] -= M;
    }
    Unite();
  }

  double Irregularity() {
    double Sum0= Sum();
    if (Sum0 == 0)
      return 0;
    double Sum1=0.0, Sum2= 0.0;
    int m= NumClasses*NumClasses;
    for(int i=0; i < m; i++) {
      double Stand= V[i] / Sum0;                /// Y SI SUM0 = 0????
      Sum1 += Stand;
       Sum2 += Sqr(Stand);
   }
   double ret= sqrt( abs(m*Sum2 - Sqr(Sum1))  / (m*(m-1)) );
   return ret;
  }

  double Vertical_Irregularity() {
   double Acc= 0;
    int m= NumClasses;

   for(int j=0; j<m; j++) {
     double Sum1=0.0, Sum2= 0.0;

     double Sum0= 0.0;
     for(int i=0; i < m; i++) {
      Sum0+= V[j*m+i];
     }

     for(int i=0; i < m; i++) {
       double Stand= V[j*m+i] / Sum0;
       Sum1 += Stand;
       Sum2 += Sqr(Stand);
     }
     Acc+= sqrt( abs((m*Sum2 - Sqr(Sum1)))  / (m*(m-1)));
   }
   return Acc/m;

  }

  double Horizontal_Irregularity() {
   double Acc= 0;
   int m= NumClasses;

   for(int j=0; j<m; j++) {
     double Sum1=0.0, Sum2= 0.0;

     double Sum0= 0.0;
     for(int i=0; i < m; i++) {
      Sum0+= V[i*m+j];
     }

     for(int i=0; i < m; i++) {
       double Stand= V[i*m+j] / Sum0;
       Sum1 += Stand;
       Sum2 += Sqr(Stand);
     }
     Acc+= sqrt(abs( (m*Sum2 - Sqr(Sum1)))  / (m*(m-1)));
   }
   return Acc/m;

  }

  double Sensitivity() {
     double Acc= 0;
     int m= NumClasses;

    for(int j=0; j<m; j++) {
     double Acc2= 0;
     for(int i=0; i < m; i++) {
        for(int k=0; k < m; k++) {
           Acc2+= abs(V[i*m+j] - V[k*m+j]);
        }
     }
     Acc+= Acc2  / (m*(m-1));
   }
   return Acc/m;
  }


  double RandomExpectancy() {
   return Sum()/(NumClasses*NumClasses);
  }

/*  slice_array<float> GetRowOfClass(int n) {
     slice s= slice(n*NumClasses, NumClasses, NumClasses);   // start, length, stride
     return V(s);
  }
*/

  vector<double> WeightVector() {  // (rows summed and inversed)
    vector<double> TotalCosts(NumClasses), Weights(NumClasses);

    double Acc= 0;
    int NumZeros= 0;
    int m= NumClasses;

    for(int j=0; j<m; j++) {
     double Acc2= 0;
     for(int i=0; i < m; i++) {
        Acc2+= V[j*m+i];
     }
     if (Acc2 == 0) {
       NumZeros++;
     }
     if (!NumZeros)
       Acc+= 1/Acc2;
     TotalCosts[j]= Acc2;;
   }

   if (NumZeros) {
     for(int j=0; j<m; j++) {
       if (TotalCosts[j] == 0)
         Weights[j]= 1 / double(NumZeros);
       else
         Weights[j]= 0;
     }
   }
   else {
     for(int j=0; j<m; j++) {
       Weights[j]= (1 / TotalCosts[j]) / Acc;
     }
   }

    return Weights;
  }

};



inline double Kappa1a(const vector<unsigned long> &V, int NClasses) {
  double Ret=0.0;
  for (int i=0; i<NClasses; i++) {
    Ret+= V[i+i*NClasses];
  }
  return Ret;
}

inline long CardOfMatrix(const vector<unsigned long> &V, int NClasses) {
  long Ret=0;
  for (int i=0; i<NClasses*NClasses; i++) {
    Ret+= V[i];
  }
  return Ret;
}

inline double Kappa1(const vector<unsigned long> &V, int NClasses) {
  long Card= CardOfMatrix(V,NClasses);
  if (Card == 0)
    return 0;
  else {
    double Ret= Kappa1a(V, NClasses);
    Ret /= Card;
    return Ret;
  }
}

inline double Kappa2(const vector<unsigned long> &V, int NClasses) {
  long Card = CardOfMatrix(V,NClasses);
  if (Card == 0)
    return 0;
  double Ret=0.0;
  for (int i=0; i<NClasses; i++) {
    double Ret1=0.0;
    for (int j=0; j<NClasses; j++) {
      Ret1+= V[i+j*NClasses];
    }
    Ret1 /= Card;
    double Ret2=0.0;
    for (int j=0; j<NClasses; j++) {
      Ret2+= V[j+i*NClasses];
    }
    Ret2 /= Card;
    Ret+= Ret1 * Ret2;
  }
  return Ret;

}

inline double Kappa(const vector<unsigned long> &V, int NClasses) {
  if (V.size() == 0)
    return 0;
  double K1= Kappa1(V, NClasses);
  double K2= Kappa2(V, NClasses);

  double ret;
  if (K2 == 1) {
    // cout << "iep";
    ret= K1;  // This happens when we have only examples of one class.
  }
  else {
   ret= (K1 - K2) / (1 - K2);
  }

  return ret;
}


inline double ProdEq(const vector<unsigned long> &V, int NClasses) {
 double Ret=1.0;
  for (int i=0; i<NClasses; i++) {
    Ret*= V[i+i*NClasses] + 1;
  }
  return Ret;
}


inline double ProdNEq(const vector<unsigned long> &V, int NClasses) {
 double Ret=1.0;
  for (int i=0; i<NClasses; i++) {
    for (int j=0; j<NClasses; j++) {
      if (i != j)
        Ret*= V[i+j*NClasses] + 1;
    }
  }
  return Ret;
}


inline double QStat(const vector<unsigned long> &V, int NClasses) {
  if (V.size() == 0)
    return 0;
  double K1= ProdEq(V, NClasses);
  double K2= ProdNEq(V, NClasses);
  if ((K1 + K2) == 0)
    return 0;
  else
    return (K1 - K2) / (K1 + K2);
}

double Similarity(const vector<unsigned long> &V, int NClasses, const options &Options);

struct evaluation_statistics {
  int NumClasses;
  long AbsAccuracy;
  float Accuracy;
  float AUC; // done by example
//  float MSE;

  vector<long> AbsTruePositives;
  vector<long> AbsFalsePositives;
  vector<float> TruePositives;
  vector<float> FalsePositives;
  long TestCardinality;
  long PredCardinality;
  class_matrix ConfusionMatrix;
  class_matrix ResultMatrix;
  class_matrix CostMatrix;
  double TotalCost;
  double TestCost;
  double TestCostPerExample;

//   evaluation_statistics(){};

  evaluation_statistics(int NClasses, class_matrix C)
  : AbsTruePositives(NClasses), AbsFalsePositives(NClasses),
    TruePositives(NClasses), FalsePositives(NClasses),
    ConfusionMatrix(NClasses), ResultMatrix(NClasses),
     CostMatrix(C)
    {
    NumClasses = NClasses;
    AbsAccuracy= 0;
    Accuracy= 0.0;
    TestCardinality= 0;
    PredCardinality= 0;
    TotalCost= 0.0;
    AUC= -1.0;
    TestCost= -1.0;
    TestCostPerExample= -1.0;
  }


  evaluation_statistics(int NClasses)
  : AbsTruePositives(NClasses), AbsFalsePositives(NClasses),
    TruePositives(NClasses), FalsePositives(NClasses),
    ConfusionMatrix(NClasses), ResultMatrix(NClasses),
     CostMatrix(NClasses, 1.0)
    {
    NumClasses = NClasses;
    AbsAccuracy= 0;
    Accuracy= 0.0;
    TestCardinality= 0;
    PredCardinality= 0;
    TotalCost= 0.0;
    AUC= -1.0;
    TestCost= -1.0;
    TestCostPerExample= -1.0;
  }

  void Show(const options &Options);



  void Evaluate(const dataset & Test, const vector<discrete> &True_Results,
                const vector<discrete> &Results,
                options &Options, const vector<unsigned long> & ArgumentUseVector);

};

#endif
