/***************************************************************************
                          main.cpp  -  description
                             -------------------

    functionality        : main program: reads options and arguments, creates multitree, learns it and evaluates results
    relevant classes     :
    relevant functions   : main

    begin                : jue oct 18 10:42:30 CEST 2001
    copyright            : (C) 2001-2002 by Cesar Ferri, Jose Hernandez, M. Jose Ramirez
    email                : {cferri,jorallo,mramirez}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/


/*************************************************************************************

PROGRAM HISTORY:

   v.0.1   October  2001: New system design begins as a successor of FLIP (FLIP3.0)
   v.0.5   November 2001: Basic Facilities: multi-tree structure, parsing utilities
   v.0.6   November 2001: First Evaluation measures: testset
   v.0.7   December 2001: Cost-sensitive facilities: Use of cost and confusion matrices
   v.0.9   December 2001: Different Solutions: component matrices
   v.1.0   January  2002: Options file. General improvement of inputs and outputs
                          The system receives its current name: SMILES
                          FIRST RELEASED VERSION

   v.1.1   January  2002: ROC Facilities
   v.1.2   January  2002: Cross-validation and basic missing numeric values handling
   V.1.2.5 January  2002: K-Fold Cross-validation implemented
   v.1.3   January  2002: First Post-pruning method implemented (PEP Method)
   v.1.4   January  2002: Classes can be made equivalent in order to reduce no. of classes
   v.1.4.1 January  2002: Repeated k-fold cross-validation
   v.1.4.2 January  2002: Some memory leaks are fixed.
   v.1.5   January  2002: When only one tree is learnt the new option "ONE_AND_FORGET" is used.
                          This options deletes the or_nodes that are not further used and memory is freed.
                          SECOND RELEASED VERSION.

   v.1.6   February 2002: MSE measure and MSE split criterion implemented
   v.1.6.1 February 2002: LogLoss split criterion implemented
   v.1.6.2 February 2002: All memory leaks corrected.
   v.1.6.3 February 2002: Some correlations are computed using the Compute_Correlation function.
                          The new source file "utils.cpp" is added to the project.
   v.1.6.4 February 2002: SqDiff split criterion implemented
   V.1.6.5 February 2002: MGINI_SPLIT and GENENTROPY_SPLIT criteria implemented
   v.1.6.6 February 2002: Split Nodes Smoothing implemented (with new option "Nodes Smoothing")
   v.1.6.7 February 2002: New Smoothing Method: k-estimate
   v.1.6.8 February 2002: GINI Criterion fixed (CART) for more than 2 classes. DKM only recommended for 2 classes.
   v.1.6.9 February 2002: MSE_SPLIT remade with better results!

   v.1.7.0 March    2002: M AUC measure (Hand's measure) called AUCH implemented for evaluating classifier with more than 2 classes.
   v.1.7.1 March    2002: New split criterion: AUCH_SPLIT using Hand's measure. Useful for more than 2 classes.

   v.1.8.0 March    2002: Invented Datasets created.
                          Use of "Comb2Single" option with Kappa similarity method.
                          Two bugs fixed: SECOND_RANDOM sometimes selected NULL or_trees
                                          and Setvdist() in "exemples.h" used word instead of long and gave problems with long datasets;
   v.1.8.1 March    2002: Combination Accuracy and Comb2Single Accuracy (means and StdDev) are computed when cross-validation.
                          First Solution is also shown when combination is enabled (in order to allow an easier comparison).
   v.1.8.2 March    2002: Combination Methods: majority crisp and majority absolute stochastic corrected
   v.1.9.0 March    2002: New Combination Methods: MAX, MIN, GEOMEAN, ...
   v.1.9.1 March    2002: Smoothing in Combination
   v.1.9.2 March    2002: Postpruning now enabled also for combination (a new function PostPrune has been created)
   v.1.9.3 March    2002: Combination options arranged.
   v.1.9.4 April    2002: First Solution (shown when combination)corrected. However, it doesn't work with post-pruning.
   v.1.9.5 April    2002: New options: PruneInCombination, SIMILARITY MEASURES for Comb2Single and AllowTestWithoutOneClass
   v.1.9.6 April    2002: Mean NumRules are also output of each kind of solution.
   v.1.9.7 April    2002: The memory of the random datasets were not freed. To bugs: A temp object delete missing in the dataset constructor and no delete in main.
                          There was also a similar memory link bug in "tracta_exemple".
   v.1.9.8 May      2002: Computed the number of possible solutions of the multitree
                          The AUC value can be computed by Hand or by our way for any number of classes by using the M approach (it is equivalent)
   v.1.9.9 June     2002: Treat missing numeric values (extend numeric partition with an extra leaf for unknown value)
                          In the !TYPES directive, '0U' or '0u' means that this numeric value may contain missing values.
                          For traditional numeric types represented with 0, this doesn't affect.
                          (this is very useful for datasets such as hepatitis, where many of the examples were lost)
   v.1.9.10 June    2002: Now SMILES ignores some attributes when instead of the type, an 'I' or 'i' is put in the !TYPES directive.
   v.1.9.11 June    2002: New vector combination method: squared.
   v.1.9.12 June    2002: Suspended Nodes Forgetting. New options enabled to free memory from suspended nodes that probably will never be woken.
                          Currently there are three options:
                              - suspended nodes forgetting=maintain all: as before. Everything is kept in memory.
                              - suspended nodes forgetting=maintain log random: only a log(number of children) are maintained
                              - suspended nodes forgetting=maintain const random: a constant number of children is maintained
                              - suspended nodes forgetting=maintain log random with depth: only a log(number of children/DEPTH) are maintained
                          WITH THIS, MEMORY AND TIME REQUIREMENTS FOR MULTITREE ARE SIGNIFICANTLY IMPROVED

   v.2.0B   June    2002: Version 2.0 Beta release.
                          THIRD RELEASED VERSION.
   v.2.0    June    2002: Version 2.0 stable release.

   v.2.1    June    2002: New way to compute AUC for more than 2 classes: Fawcett's method.
                          New splitting criterion AUCS
   v.2.1.1  July    2002: Reliability is now calculated for each prediction. Two ways are implemented:
                            - Frequency-based reliability: as the nodes frequencies or the combined vector frequencies.
                            - Laplace reliability: same as before but using Laplace smoothing.
                          These two options can be chosen through the "options.cfg" file.
   v.2.1.2  July    2002: Reliability is used to compute 2-class AUC for combined and other solutions!!!!
                          Due to its granularity it has more variance than previous measures of AUC.
                          This AUC for more than 2 classes is still not implemented.
   v.2.1.3  July    2002: Now SMILES accepts a ".testcost" file, which keeps information about the cost of attribute tests.
   v.2.1.4  July    2002: AUC of combination for 2 classes is fixed. Now it matches with AUC computed with nodes.
   v.2.1.5  July    2002: AUC of combination for more than 2 classes is implemented.
                          It uses a ClassProbabilityVector in predictions that can be computed as raw frequencies
                          or using Laplace. There is a new option "class probability vector method" in the option file.
                          In the case of using combination, some methods that modify the vector, such as "DIFFERENCE" can mangle the AUC and make it nonsense.
   v.2.1.6  July    2002: Prediction now works better for combination that uses "DIFFERENCE" Or other methods that mangle the probability vector.
   v.2.1.7  July    2002: AUC by Hand (computed by nodes) is now implemented in a way that this measure is independent of the casual ordering of two or more nodes that have the same ratio.
                          Consequently, now AUC computed by nodes matches exactly with AUC computed by examples.
   v.2.1.8  July    2002: In "roc.cpp" a double type is used for cardinalities and now smoothing works better.
                          This has the following consequences:
                            - MSE and LogLoss can be slightly different if Options.ProbabilityInSplittingCriteria == FROM_FREQUENCY_SMOOTHING
                            - AUC by nodes with LAPLACE SMOOTHING and AUC by examples with LAPLACE SMOOTHING now match.
   v.2.2    July    2002: testcost learning (modification of optimality measure implemented).
                          For learning the repeated use of the same argument is not checked and hence is reckonening.
                          For computing the total test cost, we do check the repeated use of the same argument, and it is not reckoned, giving an exact measurement of testcost.
                          Note that testcost learning can also be used to give more relevance to some attributes than others.
                          Currently, there are three methods:
                             Options.TestCostUse == TEST_COSTS_NO_USE
                             Options.TestCostUse == TEST_COSTS_LINEAR_PLUS1_WITHOUT_REPETITION
                             Options.TestCostUse == TEST_COSTS_LINEAR_PLUS1_WITH_REPETITION
   v.2.2.1  July    2002: new way of selecting the Best Tree from the multitree:  TEST_COST_BEST
                          This method selects the tree with the lowest Test Cost (repeated use of the same attribute in the same branch are discarded).
   v.2.2.2  July    2002: the way in which TEST_COST_BEST now is weighted with the cardinality of each node. Now it computes the real TEST_COST
   v.2.2.3  July    2002: A new Best Tree Method: OCCAM_AND_TEST_COST_BEST that combines both OCCAM and TEST_COST through the use of a factor.
                          The weight of each is determined by the new value TestCostRelevanceInSelectBest that must be from 0 to 1.
   v.2.2.4  July    2002: A new weight that tells how much TEST_COSTS are used to modify splitting criteria. TestCostRelevanceInSplitting that must be from 0 to infinite.
   v.2.2.5  July    2002: Two bugs fixed: "test cost method=no test costs" didn't work and there was a problem with numeric attributes with unknown values.
   v.2.2.6  July    2002: More friendly messages are output when the cost and testcost files are not provided correctly. The good way is something such as:
                                   ./smiles samples/liver.all.train - - liver.testcost
   v.2.2.7  July    2002: Some minor improvement in the parser. Now if a line does not contain the class, SMILES gives a proper message and exits.
   v.2.2.8  August  2002: The standard deviation of kfold crossvalidation for FirstAUC was not computed well. Now fixed.
                          Testcost is also shown for Archetype solution.
   v.2.2.9  August  2002: There was a bug in the archetype extraction. Suspended Nodes were considered. Although it seems it didn't affect on results, it did on efficiency.
   v.2.3.0  August  2002: Implementaion of the Combination to Single Solution (Archetype) Use of Other Criteria.
                          Three factors are defined:
                             Options.Comb2Single.SimilarityImportanceFactor;  // Similarity
                             Options.Comb2Single.TestCostImportanceFactor;    // TestCost
                             Options.Comb2Single.OccamImportanceFactor;       // OCCAM
                          These factors affect how the Archetype is extracted.
                          If all the factors except similarity are left to 0, then it is just a semantical extraction as before.
                          But if these factors are used, then we can extract an archetype taking into account its length and its testcost too.
   v.2.3.1  August  2002: Minor improvements to the interface.
   v.2.3.2  September 02: Test Costs not shown in statistics if they don't exist.
                          Now AUC computed and shown for archetype.
   v.2.3.3  September 02: Bug fixed. In "Obtain_Best_wrt_Combination2", a negative number for computing a maximum was not low enough. Now a boolean "FirstTime" is used instead.
                          The symptom was that suspended nodes were open.
                          The cause really was that in the Kappa function, the ret= (K1 - K2) / (1 - K2), and K2 could be 1;


KNOWN BUGS:
   Currently the program has no recognised bugs.
   It has no memory leaks either: (seen with memprof and memusage).
   The program works correctly if inputs conform to the specification.
   The application may hang if the input files are not in the correct format. (PER EXEMPLE EL FITXER segmentation.all.data)
   Problems may also occur if options are contradictory.

   An intriguing thing is that the Best Solution with TestCost does not match exactly with the archetype when all the weight is given to TestCost.
   The same happens if we select Occam Best Solution and archetyp with Occam only.

**************************************************************************************/


/*************************************************************************************

TO-DO'S:

   THE FUNCTION "SMOOTH" IN "criteria.cpp" has not been implemented in all the splitting criteria.

   make the parser more robust. If files don't follow the format (such as segmentation.all.data) the parser should detect it.

   treat inconsistencies,i.e. when two equal examples have different class (CLASHES) (the thing to do is to prune the node when all the examples are equal but with different classes).

   disable AUC measures in cross-validation when ROC options are not enabled.

   Check splitting criteria that cannot be run when the number of classes is greater than 2

   Modificar el tema de les arees ROC mximes

   Change names of TP --> Acc and FP --> Err because of confusion

   S'HAURIEN DE CONTROLLAR ELS BLANCS QUE PUGUEN HI HAURE DARRERE en tracta_linia de "options-file.cpp"

   Inputs & Outputs in XML	(PMML and RuleML)

**************************************************************************************/






#define SYSTEMNAME  "SMILES"
#define RELEASETEXT "v.2.3.3"
#define RELEASEDATE "5-September-2002"
#define SAMPLEDIR   "samples"




#include "parser.h"
#include "exemples.h"
#include "rules.h"
#include "estructures.h"
#include "temps.h"
#include "options.h"
#include "trees.h"
#include "components.h"
#include "roc.h"
#include <string>
#include <iostream>
#include <cstdlib>



// #define COMPUTE_FAWCETT_AUC


double ROC_Things(int Roc_With, int NClasses, const options &Options, vector<vector<and_tree *> > &Solutions,
                int i, double &AccArea, double &MaxArea); // forward declaration



int main(int argc, char **argv) {
  /*
    Compute_Some_Sample_Points_ROC_By_Programa(); // function just to compute areas of roc by hand. Ignore this function
    exit(0);
  */
    cout << endl;

    cout << "**** " << SYSTEMNAME << " " << RELEASETEXT << " (Release Date: " << RELEASEDATE << ")" << " ****" << endl;
    cout << endl;

//    Pause("Pausa per a generar nmeros pseudoaleatoris diferents");
/*    long Semilla2= time(0);
    cout << "Semilla: " << Semilla2 << endl;
    srand(Semilla2);
*/
    options Options;
    default_options(Options);


    string ftrain, ftest, fcost, ftestcost;
    string defaultextension = ".all";
    string defaultdataset = SAMPLEDIR "/" "echocardiogram-UM";

             // NEW WITH MISSING
             // "autos-drivewheels-USEMISSING.all.train";
             // "hepatitis-USEMISSING.all.train";


             // bidimensional:
             //   monks
             //   tic-tac
             //   house-votes
             //   agaricus
             //   adult              // no es pot en ell si no es fa una mostra (5% per exemple)
             //   breast-wdbc.all
             // bidimensional with some missing values
             //   breast-wpbc.all
             //   ionosphere.all
             //   liver.all
             //   pima.all
             //   chess-kr-vs-kp.all
             //   sonar.all
             //   breast-cancer-w.all
             //   cyl-bands.all      // no es pot utilitzar. T ms de 256 valors per a un atribut
             //   hepatitis.all
             //   thyroid-hypo.all
             //   thyroid-sick-eu.all
             //   spam.all             // es lent d'aprendre. Amb cross-validation no sol poder acabar
             // more than 2 classes with equivalences to reduce them to 2 classes
             //   tae2c.all
             //   cars2c.all
             //   nursery2c.all
             //   pendigits2c0.all       // es lent d'aprendre, especialment per a CART
             //   page-blocks2c.all
             //   yeast2c.all
             //   letter2cA.all          // es lent d'aprendre
             //   optdigits2c0.all
             //
             // more than 2 dimensions
             //   nursery.all
             //   cars.all
             //   tae.all
             //   pendigitis.all
             //   drug  (non-UCI)
             //   page-blocks.all
             //   yeast.all
             //   letter.all
             //   optdigits.all

             //  NOUS MULTICLASS
             //    balance-scale.data
             //    dermatology.data
             //    echocardiogram.data
             //    ecoli.data
             //    new-thyroid.data
             //    post-operative.data
             //    processed.cleveland.data
             //    soybean-small.data
             //  NOUS BINARIS
             //    SPECT.data
             //    SPECTF.data

             // TOY PROBLEMS:
             //   playt
             //   invented
             //   water



	  trau_opcions(argv,argc, Options, ftrain, ftest, fcost, ftestcost, defaultdataset, defaultextension);

	
	  cout << "ftrain: " << ftrain << endl;
	  cout << "ftest: " << ftest << endl;
	  cout << "fcost: " << fcost << endl;
	  cout << "ftestcost: " << ftestcost << endl << endl;	  	  	
	
    hardwired_options(Options);



    long LimitKFold;
    int KFold;

    if ((Options.CrossValidation == CROSS_VALIDATION) || (Options.CrossValidation == KFOLD_CROSS_VALIDATION)) {
      LimitKFold= Options.kFoldValidation;
      KFold= LimitKFold;
    }
    else if (Options.CrossValidation == REPEATED_KFOLD_CROSS_VALIDATION) {
      KFold= Options.kFoldValidation;;
      LimitKFold= KFold * (long)Options.RepeatKFold;
    }
    else
      LimitKFold= 1;


    vector<double> AccuracyVectorFirst(LimitKFold);
    vector<long>   NumRulesVectorFirst(LimitKFold);
		vector<double> TestCostPerExampleVectorFirst(LimitKFold);
		vector<double> AUCVectorFirst(LimitKFold);
		
		vector<double> AccuracyVectorComb(LimitKFold);
    vector<double> AUCVectorComb(LimitKFold);
    vector<double> AccuracyVectorComb2Single(LimitKFold);
    vector<double> AUCVectorComb2Single(LimitKFold);
    vector<long>   NumRulesVectorComb2Single(LimitKFold);
    vector<double> TestCostPerExampleVectorComb2Single(LimitKFold);

    vector<double> AccuracyVector(LimitKFold);
    vector<long>   NumRulesVector(LimitKFold);
    vector<double> MSEVector(LimitKFold);	  	
    vector<double> LogLVector(LimitKFold);
		vector<double> TestCostPerExampleVector(LimitKFold);
    vector<double> AUC0Vector(LimitKFold);
    vector<double> AUC1Vector(LimitKFold);
    vector<double> AUC2Vector(LimitKFold);
    vector<double> AUC4Vector(LimitKFold);
    vector<double> AUCHVector(LimitKFold);
  	#ifdef COMPUTE_FAWCETT_AUC
    vector<double> AUCSVector(LimitKFold);
    #endif
    vector<double> NumSolutions(LimitKFold);
		vector<double> NumTrees(LimitKFold);
	
		vector<double> Times(LimitKFold);		
//    vector<double> NumNodes(LimitKFold);
    vector<vector<double> > AccClassVector(LimitKFold);

    int NClasses;

    for(int k=0; k<LimitKFold; k++)
    {
      if (k==0) {
        cout << "Training Set: " << "\"" << ftrain << "\"" << endl;
      }

      dataset train(ftrain, Options);
      NClasses= train.NumClasses();

      if (k==0) {
        cout << "No. of Attributes: " << train.NumArguments() << endl;
        cout << "Cardinality: " << train.Cardinality() << " examples " << endl;

        if (train.Cardinality() == 0) {
          cout << "Learning cannot be done with 0 examples." << endl;
          exit(-1);
        }

        for(int i= 0; i < train.NumClasses(); i++) {
           cout << "Class: " << i << " (\"" << train.NameClassValue(i) << "\"). Distribution: " << train.vdist[i]
         // << ". Weight: " << train.weights[i]
         << endl;
        }
      }


      cout << endl;

      if (LimitKFold != 1) {
        cout << endl << endl << "***************************************************************" << endl;
        cout << "    KFold Cross-Validation: Iteration " << k+1 << " from a total of " << LimitKFold << endl;
        cout << "***************************************************************" << endl;
 	    }


       // train.show();


 	
   	  dataset *Validation_Test;

 	
 	
      if ((Options.SampleTrainingSet == SAMPLE_TRAINING_SET) ||
          (Options.Multitree.BestTreeCrit == CROSS_COVERAGE) ||
          (Options.Comb.Comb == CROSS_COVERAGE_COMBINATION) ||
          (Options.CrossValidation == CROSS_VALIDATION)||
          (Options.CrossValidation == KFOLD_CROSS_VALIDATION) ||
          (Options.CrossValidation == REPEATED_KFOLD_CROSS_VALIDATION) ) 	       {

        Validation_Test= new dataset;

        if ((Options.CrossValidation == KFOLD_CROSS_VALIDATION) ||
           (Options.CrossValidation == REPEATED_KFOLD_CROSS_VALIDATION) ){
 	 		    train.RandomExtract(1-Options.SampleTrainingSetProportion,*Validation_Test, k % KFold);
 		       // we can't use "RandomExtractr" for KFOLD_CROSS_VALIDATION because the seed must always be the same! 	 		
 	 		  }
 	 		  else {  // anyother option
 	 		   	train.RandomExtract(1-Options.SampleTrainingSetProportion,*Validation_Test, 0);      // always gets the same result.
 	 		   	// train.RandomExtractr(1-Options.SampleTrainingSetProportion,*Validation_Test, 0);  // the randomgenerator is initialised with time()
 	 		  }
  	
 	 		
        //Validation_Test.show();
         cout << endl << "Training set sampled." << endl;
         cout << "New card of TrainSet: " << train.Cardinality() << " distributed as follows: " << endl;
         for(int i= 0; i < train.NumClasses(); i++) {
           cout << "Class: " << i << ". Distribution: " << train.vdist[i]
           // << ". Weight: " << train.weights[i]
           << endl;

         }
         long PartialCardTest= 0;
         bool ZeroClass= false;
         cout << "The rest is for the validation set with cardinality: " << Validation_Test->Cardinality() << " distributed as follows: " << endl;
         for(int i= 0; i < Validation_Test->NumClasses(); i++) {
           PartialCardTest= Validation_Test->vdist[i];
           cout << "Class: " << i << ". Distribution: " << PartialCardTest
           // << ". Weight: " << train.weights[i]
           << endl;
           if (PartialCardTest == 0)
             ZeroClass= true;
         }

         cout << endl << endl;

         if (Options.AllowTestWithoutOneClass)
           ZeroClass= false;

         if (ZeroClass) {
           cout << endl << endl << "TEST SET WITH NO CARDINALITY FOR A CLASS" << endl;
           cout << endl << "WE GO BACK FROM ITERATION " << k << " TO ITERATION ";
           k= (k / KFold) * KFold;
           cout << k << endl  << endl;
           k--;  // because it is incremented when gone to the for clause
           continue;  // goes to begin of k for
         }

         // train.show();
       }



      // Invented Dataset
       dataset *Invented;

       if (Options.Comb2Single.Method == INVENTED_DATASET) {
          //  train.show();

          cout << endl << endl;
          long nEx= Options.Comb2Single.InventedDataSetLength;

          Invented = new dataset(train, nEx);
          cout << "Invented dataset of " << nEx << " examples created: " << endl;
          // Invented->show();
       }


      if (Options.TestCostMethod == NO_TEST_COSTS) {
        // we create the vector because some evaluation metrics are computed
        for (int i= 0; i < train.NumArguments(); i++) {
          Options.TestCostVector.push_back(1.0);
        }

      }
      else {
        // Creates the vector
        for (int i= 0; i < train.NumArguments(); i++) {
          Options.TestCostVector.push_back(1.0);
        }
        if (Options.TestCostMethod == UNIFORM_TEST_COSTS) {
          // the unitary vector is created. Nothing to do.
        }
        else if (Options.TestCostMethod == TEST_COSTS_FROM_FILE) {
          llig_vector_testcosts(ftestcost, Options.TestCostVector);
        }

        cout << endl << "Test Cost Vector to be used: " << endl;
        for (int i= 0; i < train.NumArguments(); i++) {
          cout << "Argument " << i << ": " << Options.TestCostVector[i] << endl;
        }

      }

      if (Options.Weights.Method == INVERSE_FREQUENCY_WEIGHTS) {
        train.SetWeightsToInvFreq(Options.Weights.Parameter) ;
      }
      else if (Options.Weights.Method == WEIGHTS_FROM_FILE) {
                                            // Given the weights (a, b, c), sets a matrix:
                                     										// (-a, b, c)
                                     										// (a, -b, c)
                                     										 // (a, b, -c)
                                     										 // This is equivalent to oversampling
                                     										
        cout << "Weights read from training set file (if the file doesn't contain weights they will be all 1)" << endl;                               										
        for (int i=0; i<NClasses; i++) {
          cout << "Weight class " << i << ": " << train.weights[i] << endl;
        }
        cout << endl;
      }
      else if (Options.Weights.Method == UNIFORM_WEIGHTS) {
        for (int i=0; i<NClasses; i++) {
          train.weights[i]= 1;
        }
      }

      if ((Options.Weights.Method == INVERSE_FREQUENCY_WEIGHTS) ||
          (Options.Weights.Method == WEIGHTS_FROM_FILE) ||
          (Options.Weights.Method == UNIFORM_WEIGHTS) ) {
        class_matrix C(NClasses, train.weights);
        Options.CostMatrix= C;
      }
      else if (Options.Weights.Method == NO_COSTS) {
        class_matrix C(NClasses, 1);  // MATRIU AMB TOT 1
        Options.CostMatrix= C;
      }
      else if (Options.Weights.Method == COSTS_FROM_MATRIX) {
        class_matrix C(NClasses, 1);

        llig_matriu_costs(fcost, C);

        // Fiquem una matriu per programa
         /*   double v[]
                  = {   -2,  4000.5,   0   ,      1,  5.3,
                          1,  -2,   2.5,   3.2,     1,
                       3.2,    3000,    -1,      7,  3.2,
                       0.5,  1000.7, 11.2, -70.5, 0.3,
                        3.2, 9200.3,   4.3,   3.7,  -20  }; */
         /*
         double v[]
                 = {   0,  4000.5,   0   ,      1,  5.3,
                          1,  0,   2.5,   3.2,     1,
                       3.2,    3000,    0,      7,  3.2,
                       0.5,  1000.7, 11.2, 0, 0.3,
                        3.2, 9200.3,   4.3,   3.7,  0  };

        class_matrix C(NClasses, v);
        */


        Options.CostMatrix= C;
      }

      Options.WeightVector = Options.CostMatrix.WeightVector();

      if (Options.Weights.Method != NO_COSTS) {
        cout << "The cost matrix to be used:" << endl;
        Options.CostMatrix.Show();
        cout << endl ;
      }


      cout << endl ;


      Options.Components= component_matrix(train.NumArguments(), NUM_PARTITIONS);
      Options.AccComponents= component_matrix(train.NumArguments(), NUM_PARTITIONS);
      Options.RandomComponents= component_matrix(train.NumArguments(), NUM_PARTITIONS);

      if (k==0) {
        if (!Options.Validate(NClasses)) {  // THIS FUNCTION HAS TO BE IMPROVED
          cout << "Invalid Options" << endl;
          return -1;
        }
        cout << "Valid options. " << endl << endl;
      }






      {  // space where the multitree is created
         cout << "Creating the multitree." << endl;


         multitree LAMAREDELCORDER(train, Options);

         cout << "Learning begins..." << endl;


        start_st() ;
        LAMAREDELCORDER.Learn(train,*Validation_Test);
        stop_st() ;

        cout << endl;

        float MeanDepth;
        if (Options.Multitree.MaxNumTrees == 1)
        MeanDepth = 0;
        else {
          MeanDepth= Options.SecondTreeDepthAccStat / (float)(Options.Multitree.MaxNumTrees - 1);
          cout << "MeanDepth of Second Solution Start Positions: " << MeanDepth << endl;
        }


        //    Pause();
        // sleep(5);
        //  cout << "Generem l'eixida" << endl;
        //  string s= LAMAREDELCORDER.ShowTreeBest("f", train,Validation_Test, false);
        //  cout << "Este s l'arbre que s'ha construit" << endl;
        //  cout << s;




        if ((Options.Output.ShowAllMultiTreeRules != NO_SHOW_ALL_RULES)
         // && (Options.Multitree.Construction != ONE_ONLY)   ONE_AND_FORGET
          ) {
          string s2= LAMAREDELCORDER.ShowAllRules("f", train, false);

          if (Options.Output.ShowAllMultiTreeRules == SHOW_ALL_RULES) {
            cout << endl;
            cout << "These are all the rules of the multitree" << endl;
            cout << s2;
          }
          else if (Options.Output.ShowAllMultiTreeRules == ALL_FILES_TO_FILE) {  // Not implemented
            cout << endl << "Outputting the multitree to a file is still not implemented" << endl;
          }
        }

        cout << endl;

        if (Options.PostPruningMethod != NO_POSTPRUNING) {
          // Marks the nodes as POSTPRUNE if it is enabled
          cout << "Begins postpruning... ";
          cout.flush();
          LAMAREDELCORDER.PostPrune();
          cout << " finished OK." << endl << endl;
        }

        // BEGINS SELECTION AND EVALUATION
        vector<and_tree *> Solution1FromComb;

        if (Options.Comb2Single.Method != NO_EXTRACTION) {
          cout << "Extracting Solution Most Similar to Combination... ";
          cout.flush();
          LAMAREDELCORDER.Assign_Difference_wrt_Combination(train, *Invented, Options);

          LAMAREDELCORDER.Obtain_Best_wrt_Combination(Solution1FromComb, Options);
          cout << " finished OK." << endl << endl;
        }

        // EVALUATION

        dataset *Test;

        if ((Options.CrossValidation != CROSS_VALIDATION) &&
            (Options.CrossValidation != KFOLD_CROSS_VALIDATION) &&
            (Options.CrossValidation != REPEATED_KFOLD_CROSS_VALIDATION)
           ) {
          if (ftest == "") {
            cout << "No test set defined. The training test will also be used for test" << endl;
            ftest= ftrain;
          }

          cout << endl;
          cout << "The test set: " << "\"" << ftest << "\"" << " will be used to evaluate the results" << endl;

          Test= new dataset(ftest, train, Options);
        }
        else {
          if ((Options.SampleTrainingSetProportion >= 1) || (Options.SampleTrainingSetProportion <= 0)) {
            cout << "For using cross-validation the sample proportion must be strictly between 0 and 1, exclusively" << endl;
          }

          cout << "The Test Set to be used to evaluate the results is extracted from the training set with sample percentage: " << 1-Options.SampleTrainingSetProportion << endl;
          Test= Validation_Test;
        }


        cout << endl << "Predicting and preparing statistics" << endl;

        evaluation_statistics EvalComb(NClasses, Options.CostMatrix);

        if (Options.Comb.Comb != NO_COMBINATION) {
          cout << endl << endl << "Showing first the results of the combination method of all branches: " << endl;
          cout << endl << endl << endl << endl << "COMBINATION RESULTS: ";
     	    LAMAREDELCORDER.Evaluate(*Test, EvalComb, Options);
          EvalComb.Show(Options);
          cout << endl;
        }

        cout << endl;

        vector<discrete> Sample_Results;
        Test->Obtain_Results(Sample_Results);
        long Tam= Sample_Results.size();


      	if ((Options.Output.ComputeROCPoints == COMPUTE_ROC_POINTS)||( Options.Output.ComputeROCArea == COMPUTE_ROC_AREA_BY_HAND)) {  // Currently this is only necessary for ROC things
      	  cout << endl << "Filling new test probabilities of the leaves of the multi-tree with the Test set" << endl;
      	  LAMAREDELCORDER.FillTestDistribution(*Test);
    	    // Fill all leaf nodes with their "TestDistribution"
          // Nodes with no examples under them are NOT INITIALISED
      	}

        evaluation_statistics EvalComb2Single(NClasses, Options.CostMatrix);

        if (Options.Comb2Single.Method != NO_EXTRACTION) {

          cout << endl << "Showing now the results of the single solution most similar to combination: " << endl;
          cout << endl << endl << endl << "ONE COMB-LIKE SOLUTION RESULTS: ";

          LAMAREDELCORDER.Mark(Solution1FromComb);
          vector<discrete> Res;
          vector<unsigned long> ArgumentUseVector(train.NumArguments());

          LAMAREDELCORDER.Result_as_Marked(*Test, Options, Res, ArgumentUseVector);

          EvalComb2Single.Evaluate(*Test, Sample_Results, Res, Options, ArgumentUseVector);

          // EvalComb2Single.AUC is not computed, so we have to compute it now
          double AUC=compute_AUC_by_Hand(Solution1FromComb, ROC_WITH_TRAIN_ORDER_TEST_FILL,  Options);
          AUCVectorComb2Single[k]= AUC;

					TestCostPerExampleVectorComb2Single[k]= EvalComb2Single.TestCostPerExample;

          long SizeSol= Solution1FromComb.size();
          cout << "Num. of Rules: " << SizeSol << endl;

          EvalComb2Single.Show(Options);

          cout << endl;

        }


        // Now the invented (random) dataset is no longer useful. We can delete it
        if (Options.Comb2Single.Method == INVENTED_DATASET) {
          delete Invented;
          cout << endl << "Invented dataset destroyed successfully." << endl;
        }

        vector<and_tree *>  SolutionFirst;
        evaluation_statistics EvalFirst(NClasses, Options.CostMatrix);

        if (Options.Comb.Comb != NO_COMBINATION) {
        // if (Options.ComputeSolutionFirst == YES) {

            cout << endl << endl;
            if (Options.Multitree.MaxNumTrees == 1) {
              cout << "NOTE: If you are using only 1 tree, disable combination and the output results will be simpler and easier to understand" << endl;
            }

            cout << "FIRST SOLUTION: (solution as if 1 tree were generated). \n  Note: if postpruning is enabled this solution is not the same as with 1 tree." << endl;

            LAMAREDELCORDER.ObtainFirst(SolutionFirst);

            LAMAREDELCORDER.Mark(SolutionFirst);
            vector<discrete> Res;
            vector<unsigned long> ArgumentUseVector(train.NumArguments());

            LAMAREDELCORDER.Result_as_Marked(*Test, Options, Res, ArgumentUseVector);

            EvalFirst.Evaluate(*Test, Sample_Results, Res, Options, ArgumentUseVector);

            long SizeSol= SolutionFirst.size();
            cout << "Num. of Rules: " << SizeSol << endl;

            EvalFirst.Show(Options);

            // EvalFirst.AUC is not computed, so we have to compute it now
            double AUC=compute_AUC_by_Hand(SolutionFirst, ROC_WITH_TRAIN_ORDER_TEST_FILL,  Options);
            AUCVectorFirst[k] = AUC;
            cout << "AUC Hand: " << AUCVectorFirst[k] << endl;


            // cout << endl; LAMAREDELCORDER.Show(SolutionFirst, "f");

        // }
        }

        int NArbres=  Options.SelectKBest.NumSolutions;

        if (NArbres > 1) {
          cout << endl << "Let us select the " << NArbres << " best trees." << endl;
        }

        vector<vector<and_tree *> > Solutions;

        LAMAREDELCORDER.ObtainNBestDifferent(NArbres, Solutions);

        if (NArbres > 1) {
          cout << endl << endl << "From " << NArbres << " seeked, " << Solutions.size() << " single olutions have been found." << endl;
          cout << "Showing their properties:" << endl;
        }
        else {
          cout << endl << endl << "Showing below the properties of the best single tree" << endl;
        }


      	vector<evaluation_statistics> EvalVector;

  	  	
        vector<vector<discrete> > Results;

        vector<component_matrix> Comps;
        double AccAccu= 0;
        double MaxAccu= 0;
        double AccArea= 0;
        double MaxArea= 0;

        for (int i=0; i<NArbres; i++) {

          long SizeSol= Solutions[i].size();
          cout << endl << endl;
          cout << "SOLUTION " << i << ": " << SizeSol << " rules " << endl;


          /*
          long Total0=0, Total1=0;
          for(int j=0; j<SizeSol; j++) {
            cout << "Node: " << j << endl;
            cout << " Train Dist: " << Solutions[i][j]->ClassDistribution[0] << "," << Solutions[i][j]->ClassDistribution[1] << endl;
            if (Solutions[i][j]->TestDistribution.size() == 0)
              cout << " Test  Dist: EMPTY" << endl;
            else {
              cout << " Test  Dist: " << Solutions[i][j]->TestDistribution[0] << "," << Solutions[i][j]->TestDistribution[1] << endl;
              Total0+= Solutions[i][j]->TestDistribution[0];
               Total1+= Solutions[i][j]->TestDistribution[1];
            }
          }
          cout << endl << "Total test examples: " << Total0+Total1 << ", from which C0: " << Total0 << " and C1: " << Total1 << endl;
          */

          component_matrix Comp(Test->NumArguments(), NUM_PARTITIONS);
          Comp.fill_matrix(Solutions[i]);
          Comps.push_back(Comp);

          if (Options.Output.ShowSolutionsComponents == SHOW_COMPONENTS) {
            cout << "Component Matrix: " << endl;
            Comp.Show();
          }

          LAMAREDELCORDER.Mark(Solutions[i]);
          vector<discrete> EmptyVector;
          Results.push_back(EmptyVector); // Se fa una cpia

          vector<unsigned long> ArgumentUseVector(train.NumArguments());
          LAMAREDELCORDER.Result_as_Marked(*Test, Options, Results[i], ArgumentUseVector);

          evaluation_statistics Eval2(NClasses, Options.CostMatrix);
          Eval2.Evaluate(*Test, Sample_Results, Results[i], Options, ArgumentUseVector);

          EvalVector.push_back(Eval2);

					TestCostPerExampleVector[k]= Eval2.TestCostPerExample;

          EvalVector[i].Show(Options);

          float Accu= EvalVector[i].Accuracy;

          AccAccu+= Accu;
          if (Accu > MaxAccu)
            MaxAccu= Accu;



          if (Options.Output.ShowAllKBestSolutions == SHOW_SOLUTIONS) {
            cout << "Let us show the solution: " << i << endl;

            LAMAREDELCORDER.Show(Solutions[i], "f");

            cout << endl;
          }
          else if (Options.Output.ShowAllKBestSolutions == ALL_SOLUTIONS_TO_FILE) {
            cout << "Outputting the solutions to files is still not implemented" << endl;  // Not implemented
          }

          if (Options.Output.ComputeROCArea == COMPUTE_ROC_AREA_BY_HAND) {
            AUCHVector[k]=compute_AUC_by_Hand(Solutions[i], ROC_WITH_TRAIN_ORDER_TEST_FILL,  Options);
						cout<<"AUC Hand: "<<AUCHVector[k]<<endl;

						#ifdef COMPUTE_FAWCETT_AUC
              AUCSVector[k]=compute_AUC_by_Srinivasan(Solutions[i], ROC_WITH_TRAIN_ORDER_TEST_FILL,  Options);
						  cout<<"AUC Fawcett: "<<AUCSVector[k]<<endl;
            #endif
          }

					if (Options.ShowNumberSolutions==SHOW_NUMBER_SOLUTIONS)
				  {
							NumSolutions[k]=LAMAREDELCORDER.Compute_Solutions();
							if (NArbres > 1) {
							  cout<<"Number of possible solutions in the Multitree: "<<NumSolutions[k]<<endl;
					    }
					}
					NumTrees[k]=LAMAREDELCORDER.SuspendedNodesExplored();							

						
					if (true) {	  // Options.MSE_LOGLOSS
            // Computing the MSE measure
            float MSE= compute_MSE(Solutions[i], Options, MSE_WITH_TEST);
            MSE= MSE/NClasses;
            MSE= 1 - MSE;

            MSEVector[k]= MSE;	  	

            /* if (k==68)
              cout << "iep" << endl;*/

            float LogLoss= compute_LogLoss(Solutions[i], Options, MSE_WITH_TEST);
            float LogLoss2= pow(2, -LogLoss);
       //     LogLoss= exp(-LogLoss);
       //     cout << LogLoss;

            LogLVector[k]= LogLoss2;	
					}
					
					
						
          // ROC THINGS
          if (Options.Output.ComputeROCPoints == COMPUTE_ROC_POINTS) {

            //     TODO: Modificar el tema de les arees ROC mximes



            AUC0Vector[k]= ROC_Things(ROC_WITH_TRAIN, NClasses, Options, Solutions, i, AccArea, MaxArea);



            double TP= EvalVector[i].AbsTruePositives[0];
            double FP= EvalVector[i].AbsFalsePositives[1];

            double TPCol= (TP + EvalVector[i].AbsFalsePositives[0]);
            double FPCol= (FP + EvalVector[i].AbsTruePositives[1]);

            double TPR= TP / TPCol;
            double FPR= FP / FPCol;


/*
            if (TP == 0) {
              cout << "TRUE POSITIVES = 0, AUC1, AUC2 and AUC4 assigned to 0.5: " << endl;
              AUC1Vector[k]= 0.5;
              AUC2Vector[k]= 0.5;
              AUC4Vector[k]= 0.5;
            }
            else
*/
             if (TPCol == 0) {
              cout << "TP + TN = 0, AUC1, AUC2 and AUC4 assigned to FPR: " << FPR << endl;
              AUC1Vector[k]= FPR;
              AUC2Vector[k]= FPR;
              AUC4Vector[k]= FPR;
            }
            else if (FPCol == 0) {
              cout << "FP + FN = 0, AUC1, AUC2 and AUC4 assigned to TPR: " << TPR << endl;
              AUC1Vector[k]= TPR;
              AUC2Vector[k]= TPR;
              AUC4Vector[k]= TPR;
            }
            else {
              AUC1Vector[k]= ROC_Things(ROC_WITH_TEST, NClasses, Options, Solutions, i, AccArea, MaxArea);

              AUC2Vector[k]= ROC_Things(ROC_WITH_TRAIN_ORDER_TEST_FILL, NClasses, Options, Solutions, i, AccArea, MaxArea);

              //     cout << "  TP: " << TP << ", FP: " << FP << endl;
              cout << endl << "ROC analysis with just one labelling" << endl;
              cout << "  TPR: " << TPR << ", FPR: " << FPR << endl;

              vector<vector<double> > punts;

              vector<double> V(2);
              V[0]= 0; V[1]= 0;
              punts.push_back(V);
              V[0]= FPR; V[1]= TPR;
              punts.push_back(V);
              V[0]= 1; V[1]= 1;
              punts.push_back(V);

   	  	      double Area;
	            compute_ROC_AREA(punts, Area);
              cout << "Area Under the ROC Curve: " << Area << endl;
              AUC4Vector[k]= Area;
           }

          }

        }

        cout << endl << "END OF RESULTS" << endl;

        if (NArbres > 1) {

          float Accuracy= AccAccu / NArbres;

          cout << endl << "Calculating mean differences between solutions" << endl;
          double AccDisc= 0;
          double AccEucDist= 0;
          double AccManDist= 0;
          for (int i=0; i<NArbres; i++) {
            for (int j=0; j<NArbres; j++) {
              if (i !=j) {
                long Discrepancy= 0;
                for (long k=0; k < Tam; k++) {
                  if (Results[i][k] != Results[j][k])
                    Discrepancy++;
                }
                AccDisc+= Discrepancy / float(Tam);

                AccEucDist+= Comps[i].EuclideanDistance(Comps[j]);
                AccManDist+= Comps[i].ManhattanDistance(Comps[j]);
              }
            }
          }

          double Den= NArbres * (NArbres - 1);
          cout << "Mean Discrepancy: " << AccDisc / Den << endl;
          cout << "Mean Syntactic Euclidean Distance: " << AccEucDist / Den << endl;
          cout << "Mean Syntactic Manhattan Distance: " << AccManDist / Den << endl;
          cout << "Mean Accuracy: " << Accuracy << endl;
          cout << "Maximum Accuracy: " << MaxAccu << endl;
          if (Options.Output.ComputeROCPoints == COMPUTE_ROC_POINTS) {
            cout << "Mean ROC Area: " << AccArea / NArbres << endl;
            cout << "Maximum ROC Area: " << MaxArea << endl;
          }
          cout << "MeanDepth 2nd Opened OR: " << MeanDepth << endl;


        }


        cout << endl << endl;


        delete Test;
        cout << "Test dataset destroyed successfully" << endl;


        if (LimitKFold > 1) {
          vore_st("\n====> Time used (for learning in this split): ");
          Times[k]= donam_st();


          AccuracyVectorFirst[k] = EvalFirst.Accuracy;
          NumRulesVectorFirst[k]= SolutionFirst.size();
          TestCostPerExampleVectorFirst[k] = EvalFirst.TestCostPerExample;
          AccuracyVectorComb[k]= EvalComb.Accuracy;  // for combination
          AUCVectorComb[k]= EvalComb.AUC;  // for combination
          AccuracyVectorComb2Single[k]= EvalComb2Single.Accuracy;  // for combination to single (using invented)
          NumRulesVectorComb2Single[k]= Solution1FromComb.size();

          // AccuracyVectorComb[k]= 0;  // for combination
          // AccuracyVectorComb2Single[k]= 0;  // for combination to single (using invented)

          AccuracyVector[k]= EvalVector[0].Accuracy; // for best solution (e.g. using Occam).
          NumRulesVector[k]= Solutions[0].size();

          AccClassVector[k].clear(); // This is because k can go backwards
          for(int l=0; l < NClasses; l++) {
            double den= double(EvalVector[0].AbsTruePositives[l] + EvalVector[0].AbsFalsePositives[l]);
            if (den != 0)
              AccClassVector[k].push_back(EvalVector[0].AbsTruePositives[l] / den);
            else
              AccClassVector[k].push_back(0);
          }
        }



      } // finishes space where the multitree has been defined



      if (LimitKFold > 1) {
        cout << "The multitree has been destroyed successfully" << endl << endl;
      }

    }  // end of cross-validation for








    if (LimitKFold == 1) {
      cout << "The multitree has been destroyed successfully" << endl;

      vore_st("\n====> Time used (for learning): ");
    }
    else {
      double AccFirstAcc= 0;
      double DevFirstAcc= 0;
      double AccFirstAUC= 0;
      double DevFirstAUC= 0;
      double AccFirstNumRules= 0;
      double AccTestCostFirst= 0;
      double DevTestCostFirst= 0;
      double AccCombAcc= 0;
      double DevCombAcc= 0;
      double AccCombAUC= 0;
      double DevCombAUC= 0;
      double AccComb2SingleAcc= 0;
      double DevComb2SingleAcc= 0;
      double AccComb2SingleAUC= 0;
      double DevComb2SingleAUC= 0;
      double AccComb2SingleNumRules= 0;
      double AccComb2SingleTestCost= 0;
      double DevComb2SingleTestCost= 0;
      double AccAcc= 0;
      double DevAcc= 0;
      double AccNumRules= 0;
      double AccAUC0= 0;
      double DevAUC0= 0;
      double AccAUC1= 0;
      double DevAUC1= 0;
      double AccAUC2= 0;
      double DevAUC2= 0;
      double AccAUC4= 0;
      double DevAUC4= 0;
      double AccMSE= 0;
      double DevMSE= 0;
      double AccLogL= 0;
      double DevLogL= 0;
      double AccTestCost= 0;
      double DevTestCost= 0;
      double AccAUCH= 0;
      double DevAUCH= 0;
			#ifdef COMPUTE_FAWCETT_AUC
      double AccAUCS= 0;
      double DevAUCS= 0;
      #endif
      double NumS=0;
      double DevNumS=0;
      double NumT=0;
      double DevNumT=0;
      double AccTimes=0;
      double DevTimes=0;

      vector<double> AccClass(NClasses);
      vector<double> DevAccClass(NClasses);

      cout << endl << endl << endl << "oooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooooo" << endl;
      cout << endl << "Separate Results of " << LimitKFold << "-fold cross-validation for different solutions" << endl;
      for (int k=0; k<LimitKFold; k++) {
        cout << "Iteration: " << k << endl;

	      if (Options.ShowNumberSolutions== SHOW_NUMBER_SOLUTIONS)
 					  {
              cout << " Number of trees   : " << NumTrees[k] << endl;
              NumT+=NumTrees[k];
              cout<<"  Solutions in Multitree   : "<<NumSolutions[k]<<endl;
              NumS+=NumSolutions[k];
						}
						
				
        cout << "  First Solution Accuracy  : " << AccuracyVectorFirst[k] << "    No. of rules: " << NumRulesVectorFirst[k] << endl;
        AccFirstAcc+= AccuracyVectorFirst[k];
        AccFirstNumRules+= NumRulesVectorFirst[k];

        cout << "  First Solution AUC       : " << AUCVectorFirst[k] << endl;
        AccFirstAUC+= AUCVectorFirst[k];


        cout << "  First Solution TestCost per Example: " << TestCostPerExampleVectorFirst[k] << endl;
        AccTestCostFirst+= TestCostPerExampleVectorFirst[k];	  	


        cout << "  Combination Accuracy     : " << AccuracyVectorComb[k] << endl;
        AccCombAcc+= AccuracyVectorComb[k];

        cout << "  Combination AUC          : " << AUCVectorComb[k] << endl;
        AccCombAUC+= AUCVectorComb[k];


        cout << "  Comb2Single Accuracy     : " << AccuracyVectorComb2Single[k]  << "    No. of rules: " << NumRulesVectorComb2Single[k] << endl;
        AccComb2SingleAcc+= AccuracyVectorComb2Single[k];
        AccComb2SingleNumRules+= NumRulesVectorComb2Single[k];

        cout << "  Comb2sinlge AUC          : " << AUCVectorComb2Single[k] << endl;
        AccComb2SingleAUC+= AUCVectorComb2Single[k];

        cout << "  Comb2Single TestCost per Example: " << TestCostPerExampleVectorComb2Single[k] << endl;
        AccComb2SingleTestCost+= TestCostPerExampleVectorComb2Single[k];	  	


        cout << "  Best Solution Accuracy   : " << AccuracyVector[k]  << "    No. of rules: " << NumRulesVector[k] << endl;
        AccAcc+= AccuracyVector[k];
        AccNumRules+= NumRulesVector[k];


        for (int l=0; l < NClasses; l++) {
          cout << "  Accuracy Class " << l << "         : " << AccClassVector[k][l] << endl;
          AccClass[l]+= AccClassVector[k][l];
        }



				if (Options.Output.ComputeROCPoints == COMPUTE_ROC_POINTS)
				{
	        cout << "  AUC0 (train)             : " << AUC0Vector[k] << endl;
  	      AccAUC0+= AUC0Vector[k];
    	    cout << "  AUC1 (test)              : " << AUC1Vector[k] << endl;
      	  AccAUC1+= AUC1Vector[k];
       	 cout << "  AUC2 (train-order + test): " << AUC2Vector[k] << endl;
       	 AccAUC2+= AUC2Vector[k];
       	 cout << "  AUC4 (train-1-lab + test): " << AUC4Vector[k] << endl;
       	 AccAUC4+= AUC4Vector[k];
				}

			  if ( Options.Output.ComputeROCArea == COMPUTE_ROC_AREA_BY_HAND)
	       {
				 	cout << "  AUC by Hand              : " << AUCHVector[k] << endl;
        	AccAUCH+= AUCHVector[k];
  				#ifdef COMPUTE_FAWCETT_AUC
	  		 	cout << "  AUC by Fawcett           : " << AUCSVector[k] << endl;
        	AccAUCS+= AUCSVector[k];        	
				  #endif
				 }

        cout << "  MSE                      : " << MSEVector[k] << endl;
        AccMSE+= MSEVector[k];	  	

        cout << "  LogLoss                  : " << LogLVector[k] << endl;
        AccLogL+= LogLVector[k];	  	

        if (Options.TestCostMethod != NO_TEST_COSTS) {
          cout << "  TestCost per Example     : " << TestCostPerExampleVector[k] << endl;
          AccTestCost+= TestCostPerExampleVector[k];	  	
        }

        cout << "  Time                  : " << Times[k] << endl;
        AccTimes+= Times[k];	  	

        cout << endl;
      }


      // Standard Deviation: sqrt( [n * (Sum sqr(x)) - sqr(Sum x)] / n(n-1) )


      for (int k=0; k<LimitKFold; k++) {
        DevFirstAcc+= Sqr(AccuracyVectorFirst[k]);

        DevFirstAUC+= Sqr(AUCVectorFirst[k]);  // Fixed

        DevTestCostFirst+= Sqr(TestCostPerExampleVectorFirst[k]);

        DevCombAcc+= Sqr(AccuracyVectorComb[k]);
        DevCombAUC+= Sqr(AUCVectorComb[k]);

        DevComb2SingleAcc+= Sqr(AccuracyVectorComb2Single[k]);
        DevComb2SingleAUC+= Sqr(AUCVectorComb2Single[k]);

        DevComb2SingleTestCost+= Sqr(TestCostPerExampleVectorComb2Single[k]);

        DevAcc+= Sqr(AccuracyVector[k]);

        DevAUC0+= Sqr(AUC0Vector[k]);
        DevAUC1+= Sqr(AUC1Vector[k]);
        DevAUC2+= Sqr(AUC2Vector[k]);
        DevAUC4+= Sqr(AUC4Vector[k]);
        DevAUCH+= Sqr(AUCHVector[k]);
				#ifdef COMPUTE_FAWCETT_AUC
        DevAUCS+= Sqr(AUCSVector[k]);
        #endif
        DevMSE+= Sqr(MSEVector[k]);
        DevLogL+= Sqr(LogLVector[k]);
        DevTestCost+= Sqr(TestCostPerExampleVector[k]);

        DevNumS+=Sqr(NumSolutions[k]);
        DevNumT+=Sqr(NumTrees[k]);

        for (int l=0; l < NClasses; l++) {
          DevAccClass[l]+= Sqr(AccClassVector[k][l]);
        }

        DevTimes+=Sqr(Times[k]);
      }



      DevFirstAcc= LimitKFold * DevFirstAcc - Sqr(AccFirstAcc);
      DevFirstAcc/= LimitKFold * (LimitKFold - 1);
      DevFirstAcc= sqrt(abs(DevFirstAcc));

      DevFirstAUC= LimitKFold * DevFirstAUC - Sqr(AccFirstAUC);
      DevFirstAUC/= LimitKFold * (LimitKFold - 1);
      DevFirstAUC= sqrt(abs(DevFirstAUC));

      DevTestCostFirst= LimitKFold * DevTestCostFirst - Sqr(AccTestCostFirst);
      DevTestCostFirst/= LimitKFold * (LimitKFold - 1);
      DevTestCostFirst= sqrt(abs(DevTestCostFirst));


      DevCombAcc= LimitKFold * DevCombAcc - Sqr(AccCombAcc);
      DevCombAcc/= LimitKFold * (LimitKFold - 1);
      DevCombAcc= sqrt(abs(DevCombAcc));

      DevCombAUC= LimitKFold * DevCombAUC - Sqr(AccCombAUC);
      DevCombAUC/= LimitKFold * (LimitKFold - 1);
      DevCombAUC= sqrt(abs(DevCombAUC));

      DevComb2SingleAcc= LimitKFold * DevComb2SingleAcc - Sqr(AccComb2SingleAcc);
      DevComb2SingleAcc/= LimitKFold * (LimitKFold - 1);
      DevComb2SingleAcc= sqrt(abs(DevComb2SingleAcc));

      DevComb2SingleAUC= LimitKFold * DevComb2SingleAUC - Sqr(AccComb2SingleAUC);
      DevComb2SingleAUC/= LimitKFold * (LimitKFold - 1);
      DevComb2SingleAUC= sqrt(abs(DevComb2SingleAUC));



      DevComb2SingleTestCost= LimitKFold * DevComb2SingleTestCost - Sqr(AccComb2SingleTestCost);
      DevComb2SingleTestCost/= LimitKFold * (LimitKFold - 1);
      DevComb2SingleTestCost= sqrt(abs(DevComb2SingleTestCost));


      DevAcc= LimitKFold * DevAcc - Sqr(AccAcc);
      DevAcc/= LimitKFold * (LimitKFold - 1);
      DevAcc= sqrt(abs(DevAcc));


      DevAUC0= LimitKFold * DevAUC0 - Sqr(AccAUC0);
      DevAUC0/= LimitKFold * (LimitKFold - 1);
      DevAUC0= sqrt(abs(DevAUC0));
      DevAUC1= LimitKFold * DevAUC1 - Sqr(AccAUC1);
      DevAUC1/= LimitKFold * (LimitKFold - 1);
      DevAUC1= sqrt(abs(DevAUC1));
      DevAUC2= LimitKFold * DevAUC2 - Sqr(AccAUC2);
      DevAUC2/= LimitKFold * (LimitKFold - 1);
      DevAUC2= sqrt(abs(DevAUC2));
      DevAUC4= LimitKFold * DevAUC4 - Sqr(AccAUC4);
      DevAUC4/= LimitKFold * (LimitKFold - 1);
      DevAUC4= sqrt(abs(DevAUC4));
      DevAUCH= LimitKFold * DevAUCH - Sqr(AccAUCH);
      DevAUCH/= LimitKFold * (LimitKFold - 1);
      DevAUCH= sqrt(abs(DevAUCH));
  		#ifdef COMPUTE_FAWCETT_AUC
      DevAUCS= LimitKFold * DevAUCS - Sqr(AccAUCS);
      DevAUCS/= LimitKFold * (LimitKFold - 1);
      DevAUCS= sqrt(abs(DevAUCS));
      #endif
      DevMSE= LimitKFold * DevMSE - Sqr(AccMSE);
      DevMSE/= LimitKFold * (LimitKFold - 1);
      DevMSE= sqrt(abs(DevMSE));
      DevLogL= LimitKFold * DevLogL - Sqr(AccLogL);
      DevLogL/= LimitKFold * (LimitKFold - 1);
      DevLogL= sqrt(abs(DevLogL));
      DevTestCost= LimitKFold * DevTestCost - Sqr(AccTestCost);
      DevTestCost/= LimitKFold * (LimitKFold - 1);
      DevTestCost= sqrt(abs(DevTestCost));


      DevNumS= LimitKFold * DevNumS - Sqr(NumS);
      DevNumS/= LimitKFold * (LimitKFold - 1);
      DevNumS= sqrt(abs(DevNumS));

      DevNumT= LimitKFold * DevNumT - Sqr(NumT);
      DevNumT/= LimitKFold * (LimitKFold - 1);
      DevNumT= sqrt(abs(DevNumT));


      for (int l=0; l < NClasses; l++) {
        DevAccClass[l] = LimitKFold * DevAccClass[l] - Sqr(AccClass[l]);
        DevAccClass[l] /= LimitKFold * (LimitKFold - 1);
        DevAccClass[l] = sqrt(abs(DevAccClass[l]));
      }

      DevTimes= LimitKFold * DevTimes - Sqr(AccTimes);
      DevTimes/= LimitKFold * (LimitKFold - 1);
      DevTimes= sqrt(abs(DevTimes));




      AccFirstAcc/= LimitKFold;
      AccFirstAUC /= LimitKFold;
      AccFirstNumRules /= LimitKFold;
      AccTestCostFirst/= LimitKFold;

      AccCombAcc/= LimitKFold;
      AccCombAUC/= LimitKFold;
      AccComb2SingleAcc/= LimitKFold;
      AccComb2SingleAUC/= LimitKFold;
      AccComb2SingleNumRules /= LimitKFold;
      AccComb2SingleTestCost /= LimitKFold;

      AccAcc/= LimitKFold;
      AccNumRules /= LimitKFold;
      AccAUC0/= LimitKFold;
      AccAUC1/= LimitKFold;
      AccAUC2/= LimitKFold;
      AccAUC4/= LimitKFold;
      AccAUCH/= LimitKFold;
			#ifdef COMPUTE_FAWCETT_AUC
      AccAUCS/= LimitKFold;
      #endif
      AccMSE/= LimitKFold;
      AccLogL/= LimitKFold;
      AccTestCost/= LimitKFold;
      NumS/=LimitKFold;
      NumT/=LimitKFold;


      for (int l=0; l < NClasses; l++) {
        AccClass[l] /= LimitKFold;
      }

      AccTimes/=LimitKFold;


      cout << endl;
      cout << "Mean Results: " << endl;
      if (Options.ShowNumberSolutions== SHOW_NUMBER_SOLUTIONS) {
        cout<<"  N. of susp. nodes explored : " << setw(12) << NumT << " +/- " << setw(12)<< DevNumT<<endl;
        cout<<"  Solutions in the Multitree : " << setw(12) << NumS << " +/- " << setw(12)<< DevNumS<<endl;
			}

      cout << " Results for 1st Solution: " << endl;
      cout << "  Accuracy of 1st Solution   : " << setw(12) << AccFirstAcc << " +/- " << setw(12) << DevFirstAcc << endl;
      cout << "  AUC (by nodes) of 1st Sol  : " << setw(12) << AccFirstAUC << " +/- " << setw(12) << DevFirstAUC << endl;
      cout << "  Mean # Rules               : " << setw(12) <<  AccFirstNumRules << endl;
      if (Options.TestCostMethod != NO_TEST_COSTS) {
        cout << "  Test Cost per Example      : " << setw(12) << AccTestCostFirst << " +/- " << setw(12) << DevTestCostFirst << endl;
      }
      cout << " Results for Combination: " << endl;
      cout << "  Accuracy of Combination    : " << setw(12) << AccCombAcc << " +/- " << setw(12) << DevCombAcc << endl;
      cout << "  AUC of Combination         : " << setw(12) << AccCombAUC << " +/- " << setw(12) << DevCombAUC << endl;

      if (Options.Comb2Single.Method != NO_EXTRACTION) {
        cout << " Results for Archetype: " << endl;
        cout << "  Accuracy of Archetype      : " << setw(12) << AccComb2SingleAcc << " +/- " << setw(12) << DevComb2SingleAcc << endl;
        cout << "  AUC (by nodes) of Archetype: " << setw(12) << AccComb2SingleAUC << " +/- " << setw(12) << DevComb2SingleAUC << endl;
        cout << "  Mean # Rules               : " << setw(12) <<  AccComb2SingleNumRules << endl;
        if (Options.TestCostMethod != NO_TEST_COSTS) {
          cout << "  Test Cost per Example      : " << setw(12) << AccComb2SingleTestCost << " +/- " << setw(12) << DevComb2SingleTestCost << endl;
        }
      }
      cout << " Results for Best Solution: " << endl;
      cout << "  Accuracy of Best           : " << setw(12) << AccAcc << " +/- " << setw(12) << DevAcc << endl;
      cout << "  Mean # Rules               : " << setw(12) <<  AccNumRules << endl;
      for (int l=0; l < NClasses; l++) {
        cout << "  Accuracy class " << l << "           : " << setw(12) << AccClass[l] << " +/- " << setw(12) << DevAccClass[l] << endl;
      }

      if (Options.Output.ComputeROCPoints == COMPUTE_ROC_POINTS)
			{
      	cout << "  AUC0 (train)               : " << setw(12) << AccAUC0 << " +/- " << setw(12)<< DevAUC0 << endl;
      	cout << "  AUC1 (test)                : " << setw(12) << AccAUC1 << " +/- " << setw(12)<< DevAUC1 << endl;
      	cout << "  AUC2 (train-order + test)  : " << setw(12) << AccAUC2 << " +/- " << setw(12)<< DevAUC2 << endl;
	      cout << "  AUC4 (train-1-lab + test)  : " << setw(12) << AccAUC4 << " +/- " << setw(12)<< DevAUC4 << endl;
			}
      if ( Options.Output.ComputeROCArea == COMPUTE_ROC_AREA_BY_HAND) {
 		     cout << "  AUC by Hand                : " << setw(12) << AccAUCH << " +/- " << setw(12)<< DevAUCH << endl;
			   #ifdef COMPUTE_FAWCETT_AUC
  		     cout << "  AUC by Fawcett           : " << setw(12) << AccAUCS << " +/- " << setw(12)<< DevAUCS << endl;
         #endif
      }
      cout << "  MSE                        : " << setw(12) << AccMSE << " +/- " << setw(12)<< DevMSE << endl;
      cout << "  LogLoss                    : " << setw(12) << AccLogL << " +/- " << setw(12)<< DevLogL << endl;
      if (Options.TestCostMethod != NO_TEST_COSTS) {
        cout << "  Test Cost per Example      : " << setw(12) << AccTestCost << " +/- " << setw(12)<< DevTestCost << endl;
      }
      cout << endl;
      cout << " Time Used                   : " << setw(12) << AccTimes << " +/- " << setw(12)<< DevTimes << endl;
      cout << endl;



/*
      double Corr_Acc_AUC= Compute_Correlation(AUC2Vector, AccAUC2, DevAUC2, AccuracyVector, AccAcc, DevAcc);
      cout << "Correlation between Accuracy and AUC2: " << Corr_Acc_AUC << endl;
      double Corr_MSE_Acc= Compute_Correlation(MSEVector, AccMSE, DevMSE, AccuracyVector, AccAcc, DevAcc);
      cout << "Correlation between Accuracy and MSE: " << Corr_MSE_Acc << endl;
      double Corr_Acc_LogL= Compute_Correlation(LogLVector, AccLogL, DevLogL, AccuracyVector, AccAcc, DevAcc);
      cout << "Correlation between Accuracy and LogLoss: " << Corr_Acc_LogL << endl;
      double Corr_AUC_MSE= Compute_Correlation(MSEVector, AccMSE, DevMSE, AUC2Vector, AccAUC2, DevAUC2);
      cout << "Correlation between AUC2 and MSE: " << Corr_AUC_MSE << endl;
      double Corr_AUC_LogL= Compute_Correlation(LogLVector, AccLogL, DevLogL, AUC2Vector, AccAUC2, DevAUC2);
      cout << "Correlation between AUC2 and LogLoss: " << Corr_AUC_LogL << endl;
      double Corr_MSE_LogL= Compute_Correlation(MSEVector, AccMSE, DevMSE, LogLVector, AccLogL, DevLogL);
      cout << "Correlation between MSE and LogLoss: " << Corr_MSE_LogL << endl;
      cout << endl;
*/
    }



    cout << endl;
    cout << "**** Smooth end of " << SYSTEMNAME << " execution ****" << endl;

    //	    Pause() ;

}











double ROC_Things(int Roc_With, int NClasses, const options &Options, vector<vector<and_tree *> > &Solutions,
                int i, double &AccArea, double &MaxArea) {


        unsigned long NNodes= Solutions[i].size();
        cout << endl;

	      if (Roc_With == ROC_WITH_TEST) {
          cout << "ROC Analysis wrt. test set" << endl;
        }
        else if (Roc_With == ROC_WITH_TRAIN_ORDER_TEST_FILL) {
          cout << "ROC Analysis wrt. training set order and test set derived propabilities" << endl;
        }
        else{
          cout << "ROC Analysis wrt. training set" << endl;
        }

        cout << "ROC Points from the assignments of the " << NNodes << " nodes with " << NClasses << " classes." << endl;
        cout << "This can take a long time for more than 2 classes..." << endl;
		    vector<vector<double> > punts;
	
	      if (Roc_With == ROC_WITH_TEST) {
	        compute_ROC_points(Solutions[i], punts, ROC_WITH_TEST, Options,0,1);
	      }
	      else if (Roc_With == ROC_WITH_TRAIN_ORDER_TEST_FILL) {
          compute_ROC_points(Solutions[i], punts, ROC_WITH_TRAIN_ORDER_TEST_FILL, Options,0,1);
        }
        else {
	        compute_ROC_points(Solutions[i], punts, ROC_WITH_TRAIN, Options,0,1);
	      }
	
	      unsigned long NPoints= punts.size();
	
	
	      cout << "Finally, the number of points is " << NPoints << " from a theoretical maximum of: " << pow(NClasses, NNodes) << endl;
	
        if (Options.Output.ShowROCPoints == SHOW_ROC_POINTS) {
	 	      for(unsigned long i1=0; i1 < punts.size(); i1++) {
	          for(unsigned long j1=0; j1 < punts[i1].size(); j1++)
	  		      cout << punts[i1][j1] << " ";
	  		    cout<<endl;
	        }
	      }
	
	      double Area;
	
	      if (NClasses == 2) {
          if (Options.Output.ComputeROCPoints == COMPUTE_ROC_POINTS) {
		
	          compute_ROC_AREA(punts, Area);
	          cout << "Area Under the ROC Curve: " << Area << endl;
	          AccArea+= Area;
            if (Area > MaxArea)
              MaxArea= Area;
	        }
	
	        if (Options.Output.GenerateROCCurveFile == GENERATE_ROC_CURVE_FILE) {
 	          const int Tam= 10;
	          char Buf[Tam];
	          strstream s(Buf, Tam);
	          s << i << ends;
	
	          string s2;
            if (Roc_With == ROC_WITH_TEST) {
              s2= "tst";
            }
            else if (Roc_With == ROC_WITH_TRAIN_ORDER_TEST_FILL) {
              s2= "tot";
            }
            else{
              s2= "trn";
            }


	    	    string rocfile= string("ROC") + s2 + string(Buf) + string(".ps");
	          cout << "Outputting the ROC curve of this solution to file: \"" << rocfile.c_str() << "\"" << endl;
 	   	      genera_ps(punts, rocfile.c_str());
 	   	    }
        }

        return Area;

}
