/***************************************************************************
                          options-file.cpp  -  description
                             -------------------

    functionality        : for parsing the option file
    relevant classes     :
    relevant functions   : ompli_options

    begin                : Wed Dec 26 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
// Ompli un data set amb tota la informacio a partir de un fitxer de dades




#include "options-file.h"
#define EXAMPLESIZE 8000

#define OFILE "options.cfg"
#define COMENTARI '%'




#define _EEM "expected error method"
#define _SMM "smoothing method"
#define _NSM "node smoothing method"
#define _CSM "class selection method"
#define _REC "reliability calculation"
#define _CPV "class probability vector calculation"
#define _NMI "numeric interval criterion"
#define _PSP "probability in splitting criteria"
#define _SPC "splitting criterion"
#define _PPM "pre-pruning method"	
#define _PST "post-pruning method"
#define _SPP "stump pruning limit"
//#define _MCN "multitree construction"
#define _SNF "suspended nodes forgetting"
#define _SNV "suspended nodes maintain const value"
#define _MTR "multitree number of trees"
#define _MBT "multitree best tree criterion"
#define _STO "second tree opening"

#define _TCM "test cost method"
#define _TCU "test cost use"
#define _WMT "weight method"
#define _CDP "cost derived probability"
#define _MSC "multitree solution combination"
#define _CST "combination smoothing"
#define _CVE "combination vector"
#define _CVM "combination vector method"
#define _CFM "combination fusion method"
#define _PIC "post-pruning in combination"

#define _SIM "similarity method for combination to single"
#define _C2S "combination to single solution method"
#define _IDL "invented dataset length"
#define _SIM2 "archetype similarity importance factor"
#define _TIM "archetype test cost importance factor"
#define _OIM "archetype occam importance factor"

#define _OUT "show rules mode"
#define _CDT "show class distribution"
#define _SAR "show all multitree rules"
#define _SAK "show all k-best solutions"
#define _SSC "show solutions components"
#define _SST "show statistics"
#define _SCM "show confusion matrix"

#define _CRP "compute ROC points"
#define _SRP "show ROC points"
#define _CRA "compute ROC area"
#define _GRF "generate ROC curve file"

#define _SKB "select k best"
#define _SCG "select k best components generation"
#define _KBN "k best number of solutions"
#define _SFR "select k best factor for random components"
#define _CRF "components random factor"
#define _MRC "components reckoning"
//#define _SP1 "split 1"
//#define _SP2 "split 2"
//#define _SP3 "split 3"
//#define _SP4 "split 4"	
#define _SNS "select k best number of solutions"

#define _STS "sample training set"
#define _STP "sample training set proportion"
#define _CRV "cross validation"
#define _KFV "k fold of cross validation"
#define _RKF "repeat kfold"
#define _TWC "allow test without one class"

#define _MNV "missing numeric values"
#define _SMS "show number of multitree possible solutions"


void funcEEM(char *valor,options &Options)
{
if(strcmp(valor,"no compute")==0) Options.ExpectedError.Method= NO_COMPUTE;
else if(strcmp(valor,"relative frequency with majority class")==0) Options.ExpectedError.Method= RELATIVE_FREQUENCY_WITH_MAJORITY_CLASS;
else if(strcmp(valor,"relative frequency with frequency probability")==0)    Options.ExpectedError.Method= RELATIVE_FREQUENCY_WITH_FREQ_PROB;
else if(strcmp(valor,"cost with minimum class")==0)  Options.ExpectedError.Method= COST_WITH_MINIMUM_CLASS;
else if(strcmp(valor,"cost with frequency probability")==0) Options.ExpectedError.Method=  COST_WITH_FREQ_PROB;
else if(strcmp(valor,"cost with cost probability")==0) Options.ExpectedError.Method= COST_WITH_COST_PROB;
else {cout<<"expected error method does not exist in the option file: " << valor << endl; exit(0);;}
}
	
void funcSMM(char *valor,options &Options)
{
if(strcmp(valor,"no smoothing")==0) Options.Smoothing.Method= NO_SMOOTHING;
else  if(strcmp(valor,"laplace")==0)   Options.Smoothing.Method= LAPLACE;
else  if(strcmp(valor,"k-estimate")==0) {Options.Smoothing.Method= K_ESTIMATE;Options.Smoothing.m= 5;}
else  if(strcmp(valor,"m-estimate")==0) {Options.Smoothing.Method= M_ESTIMATE;Options.Smoothing.m= 5;}
else  if(strcmp(valor,"m-estimate uniform")==0) {Options.Smoothing.Method= M_ESTIMATE_UNIFORM;Options.Smoothing.m= 5;}
else {cout<<"class smoothing method does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcNSM(char *valor,options &Options)
{
if(strcmp(valor,"no smoothing")==0) Options.NodeSmoothing.Method= NO_SMOOTHING;
else  if(strcmp(valor,"laplace")==0)   Options.NodeSmoothing.Method= LAPLACE;
else  if(strcmp(valor,"k-estimate")==0) {Options.NodeSmoothing.Method= K_ESTIMATE;Options.NodeSmoothing.m= 5;}
else  if(strcmp(valor,"m-estimate")==0) {Options.NodeSmoothing.Method= M_ESTIMATE;Options.NodeSmoothing.m= 5;}
else  if(strcmp(valor,"m-estimate uniform")==0) {Options.NodeSmoothing.Method= M_ESTIMATE_UNIFORM;Options.NodeSmoothing.m= 5;}
else {cout<<"node smoothing method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcCSM(char *valor,options &Options)
{
if(strcmp(valor,"majority class")==0) Options.Weights.ClassSelection= MAJORITY_CLASS;
else  if(strcmp(valor,"minimum cost selection")==0)   Options.Weights.ClassSelection= MIN_COST_CLASS;
else  if(strcmp(valor,"minimum cost class without smoothing")==0)  Options.Weights.ClassSelection = MIN_COST_CLASS_WITHOUT_SMOOTHING;
else  if(strcmp(valor,"stratification class")==0) Options.Weights.ClassSelection= STRATIFICATION_CLASS;
else  if(strcmp(valor,"stratification without smoothing")==0) Options.Weights.ClassSelection= STRATIFICATION_CLASS_WITHOUT_SMOOTHING;
else {cout<<"multitree class asignment method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcREC(char *valor,options &Options) {
  if(strcmp(valor,"frequency reliability")==0) Options.ReliabilityCalculation= FREQUENCY_RELIABILITY;
  else  if(strcmp(valor,"laplace reliability")==0)   Options.ReliabilityCalculation= LAPLACE_RELIABILITY;
  else {cout<<"reliability calculation method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcCPV(char *valor,options &Options) {
  if(strcmp(valor,"frequency class probability vector")==0)   Options.ClassProbabilityVectorCalculation= FREQUENCY_CLASS_PROBABILITY_VECTOR;
  else  if(strcmp(valor,"laplace class probability vector")==0)     Options.ClassProbabilityVectorCalculation= LAPLACE_CLASS_PROBABILITY_VECTOR;
  else {cout<<"class probability vector calculation method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcNMI(char *valor,options &Options) {
  if(strcmp(valor,"log")==0)Options.IntervalCriterion= numerical_interval_criterion(LOG, 5, 10.0); // Max number of intervals for a given argument = 2nd Argument + LOG BASE 3ND ARGUMENT OF NUMBER OF VALUES
  else  if(strcmp(valor,"max")==0) Options.IntervalCriterion=numerical_interval_criterion(MAX, 7); // Max number of intervals for a given argument
  else  if(strcmp(valor,"no limit")==0) Options.IntervalCriterion=numerical_interval_criterion(NO_LIMIT);
  else {cout<<" interval criterion does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcPSP(char *valor,options &Options)
{
if(strcmp(valor,"from frequency no smoothing")==0)  Options.ProbabilityInSplittingCriteria = FROM_FREQUENCY_NO_SMOOTHING;
else  if(strcmp(valor,"from frequency smoothing")==0) Options.ProbabilityInSplittingCriteria = FROM_FREQUENCY_SMOOTHING;
else  if(strcmp(valor,"from costs")==0) Options.ProbabilityInSplittingCriteria = FROM_COSTS;
else  if(strcmp(valor,"from frequency with stratification")==0) Options.ProbabilityInSplittingCriteria = FROM_FREQUENCY_WITH_STRATIFICATION;
else  if(strcmp(valor,"from frequency with stratification no smoothing")==0) Options.ProbabilityInSplittingCriteria = FROM_FREQUENCY_WITH_STRATIFICATION_NO_SMOOTHING;
else {cout<<"probability in splitting criteria method does not exist in the option file: " << valor << endl; exit(0);;}
}

	void funcSPC(char *valor,options &Options)
{
if(strcmp(valor,"desc mdl")==0)Options.SSC=DESC_MDL;
else  if(strcmp(valor,"left first")==0) Options.SSC=LEFT_FIRST;
else  if(strcmp(valor,"gain")==0) Options.SSC=GAIN;
else  if(strcmp(valor,"gain ratio")==0) Options.SSC=GAIN_RATIO;
else  if(strcmp(valor,"c4.5")==0) Options.SSC=C45;
else  if(strcmp(valor,"cart")==0) Options.SSC=CART;
else  if(strcmp(valor,"dkm")==0) Options.SSC=DKM;
else  if(strcmp(valor,"split expected error")==0) Options.SSC=SPLIT_EXPECTED_ERROR;
else  if(strcmp(valor,"local roc area")==0) Options.SSC=LOCAL_ROC_AREA;
else  if(strcmp(valor,"one point local roc area")==0) Options.SSC=ONE_POINT_LOCAL_ROC_AREA;
else  if(strcmp(valor,"mse")==0) Options.SSC= MSE_SPLIT;
else  if(strcmp(valor,"logloss")==0) Options.SSC= LOGLOSS_SPLIT;
else  if(strcmp(valor,"auch")==0) Options.SSC= AUCH_SPLIT;
else  if(strcmp(valor,"aucs")==0) Options.SSC= AUCS_SPLIT;
else  if(strcmp(valor,"sqdiff")==0) Options.SSC=SQDIF_SPLIT;
else  if(strcmp(valor,"mgini")==0) Options.SSC=MGINI_SPLIT;
else  if(strcmp(valor,"genentropy")==0) Options.SSC=GENENTROPY_SPLIT;
else  if(strcmp(valor,"rocv")==0) Options.SSC=ROCV_SPLIT;
else {cout<<"splitting criterion does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcPPM(char *valor,options &Options)
{
  if(strcmp(valor,"no pruning")==0)Options.PC=pruning_criterion(NO_PRUNING);
  else  if(strcmp(valor,"proportional")==0) Options.PC=pruning_criterion(PROPORTIONAL, 2.5, 3);   // El primer baix i el segon alt per a podar ms
                                                                          // en 2.5 i 3 dna 100% per al monks2
  else  if(strcmp(valor,"expected error pruning")==0) Options.PC=pruning_criterion(EXPECTED_ERROR_PRUNING, 1.05);  // Quan ms gran el factor ms poda.
  else  if(strcmp(valor,"MDL pruning")==0) Options.PC=pruning_criterion(MDL_PRUNING, 0.995);  // Quan ms gran el factor ms poda.
  else  if(strcmp(valor,"stump pruning")==0) Options.PC=pruning_criterion(STUMP_PRUNING, 3);  // Quan ms gran el factor ms poda.
  else if(strcmp(valor,"pep pruning")==0) Options.PC=pruning_criterion(PEP_PREPRUNING);   // Pessimistic Error Pruning (like Quinlan 1987 but pre-pruning)
  else {cout<<"pre-pruning criterion does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcPST(char *valor,options &Options) {
  if(strcmp(valor,"no pruning")==0)Options.PostPruningMethod= NO_POSTPRUNING;
  else if(strcmp(valor,"pep pruning")==0) Options.PostPruningMethod= PEP_POSTPRUNING;
   else {cout<<"post-pruning criterion does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcSPP(char *valor,options &Options)
{
int prof,i;
for(i=0;valor[i];i++)
 if (!isdigit(valor[i])) {cout<<"invalid limit of stump prepruning"<<endl;exit(1);}
prof=strtol(valor, (char **)NULL, 10);

Options.PC.Stump.MaxDepth= prof;
}


//void funcMCN(char *valor,options &Options)
//{
//if(strcmp(valor,"one only")==0)  Options.Multitree.Construction= ONE_ONLY;
//else if (strcmp(valor,"many and maintain")==0)     Options.Multitree.Construction= MANY_AND_MAINTAIN;
//else cout<<"multitree construction method does not exist in the option file"<<endl;
//}


void funcSNF(char *valor,options &Options) {  if (strcmp(valor,"maintain all")==0)  Options.Multitree.SuspendedNodesForgetting = MAINTAIN_ALL;
  else if (strcmp(valor,"maintain log random")==0)    Options.Multitree.SuspendedNodesForgetting = MAINTAIN_LOG_RANDOM;
  else if (strcmp(valor,"maintain const random")==0)    Options.Multitree.SuspendedNodesForgetting = MAINTAIN_CONST_RANDOM;
  else if (strcmp(valor,"maintain log random with depth")==0)   Options.Multitree.SuspendedNodesForgetting = MAINTAIN_LOG_RANDOM_WITH_DEPTH;
  else if (strcmp(valor,"maintain log random with squared depth")==0)  Options.Multitree.SuspendedNodesForgetting = MAINTAIN_LOG_RANDOM_WITH_SQUARED_DEPTH;
  else if (strcmp(valor,"maintain log random with depth adjusted")==0)  Options.Multitree.SuspendedNodesForgetting = MAINTAIN_LOG_RANDOM_WITH_DEPTH_ADJUSTED;
  else cout << "suspended node forgetting method does not exist in the option file"<<endl;
}



void funcSNV(char *valor,options &Options) {
  int prof,i;
  for(i=0; valor[i]; i++) {
    if (!isdigit(valor[i])) {
      cout << "invalid value suspended node forgetting maintain constant" << endl;
      exit(1);
    }
  }
  prof=strtol(valor, (char **)NULL, 10);
  Options.Multitree.SuspendedNodesConstMaintain= prof;
}


void funcMTR(char *valor,options &Options)
{
int prof,i;
for(i=0;valor[i];i++)
 if (!isdigit(valor[i])) {cout<<"invalid value multitree maxim number fo trees"<<endl;exit(1);}
prof=strtol(valor, (char **)NULL, 10);
Options.Multitree.MaxNumTrees= prof;
if ((prof==0)||(prof==1))Options.Multitree.Construction= ONE_AND_FORGET;
else Options.Multitree.Construction= MANY_AND_MAINTAIN;
}


void funcMBT(char *valor,options &Options)
{
if(strcmp(valor,"occam best")==0)  Options.Multitree.BestTreeCrit= OCCAM_BEST;
else  if(strcmp(valor,"test cost best")==0)   Options.Multitree.BestTreeCrit= TEST_COST_BEST;
else  if(strcmp(valor,"occam and test cost best")==0)   Options.Multitree.BestTreeCrit= OCCAM_AND_TEST_COST_BEST;
else  if(strcmp(valor,"coverage best")==0)   Options.Multitree.BestTreeCrit= COVERAGE_BEST;
else  if(strcmp(valor,"cross coverage")==0) Options.Multitree.BestTreeCrit= CROSS_COVERAGE;
else  if(strcmp(valor,"expected error best")==0) Options.Multitree.BestTreeCrit= EXPECTED_ERROR_BEST;
else  if(strcmp(valor,"split optimality best")==0) Options.Multitree.BestTreeCrit= SPLIT_OPTIMALITY_BEST;
else  if(strcmp(valor,"mdl best")==0) Options.Multitree.BestTreeCrit= MDL_BEST;
else {cout<<"multitree selection tree method does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcSTO(char *valor,options &Options)
{
if(strcmp(valor,"split optimality")==0) Options.Multitree.SecondTreeCrit.Met= SPLIT_OPTIMALITY;
else  if(strcmp(valor,"optimality rival ratio")==0)  Options.Multitree.SecondTreeCrit.Met= OPTIMALITY_RIVAL_RATIO;
//else  if(strcmp(valor,"roc distance")==0) Options.Multitree.BestTreeCrit= ROC_DISTANCE;
//else  if(strcmp(valor,"roc samling")==0) Options.Multitree.BestTreeCrit= ROC_SAMPLING;
else  if(strcmp(valor,"optimality rival ratio depth")==0) second_tree_criterion(OPTIMALITY_RIVAL_RATIO_DEPTH, 50);
else  if(strcmp(valor,"optimality rival ratio component")==0) second_tree_criterion(OPTIMALITY_RIVAL_RATIO_COMPONENT, 40);
else  if(strcmp(valor,"optimality rival ratio component random")==0)
  Options.Multitree.SecondTreeCrit= second_tree_criterion(OPTIMALITY_RIVAL_RATIO_COMPONENT_RANDOM,10);
else  if(strcmp(valor,"second topmost")==0)  Options.Multitree.SecondTreeCrit.Met= SECOND_TOPMOST;
else  if(strcmp(valor,"second bottommost")==0) Options.Multitree.SecondTreeCrit.Met= SECOND_BOTTOMMOST;
else  if(strcmp(valor,"second random")==0)  Options.Multitree.SecondTreeCrit.Met=SECOND_RANDOM;
else  if(strcmp(valor,"second random depth")==0) Options.Multitree.SecondTreeCrit= second_tree_criterion(SECOND_RANDOM_DEPTH, 0.1);
else {cout<<"multitree second  opening tree method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcTCM(char *valor,options &Options) {
  if(strcmp(valor,"no test costs")==0) Options.TestCostMethod = NO_TEST_COSTS;
  else  if(strcmp(valor,"uniform test costs")==0) Options.TestCostMethod = UNIFORM_TEST_COSTS;
  else  if(strcmp(valor,"test costs from file")==0) Options.TestCostMethod = TEST_COSTS_FROM_FILE;
  else { cout << "test cost method does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcTCU(char *valor,options &Options) {
  if (strcmp(valor,"test costs no use")==0) Options.TestCostUse = TEST_COSTS_NO_USE;
  else  if(strcmp(valor,"test costs linear plus1 without repetition")==0) Options.TestCostUse = TEST_COSTS_LINEAR_PLUS1_WITHOUT_REPETITION;
  else  if(strcmp(valor,"test costs linear plus1 with repetition")==0) Options.TestCostUse = TEST_COSTS_LINEAR_PLUS1_WITH_REPETITION;
  else { cout << "test cost use does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcWMT(char *valor,options &Options)
{
if(strcmp(valor,"no costs")==0)Options.Weights.Method= NO_COSTS;
else  if(strcmp(valor,"uniform weights")==0) Options.Weights.Method= UNIFORM_WEIGHTS;  // The same as NO_COSTS
else  if(strcmp(valor,"weights from file")==0)Options.Weights.Method= WEIGHTS_FROM_FILE;
else  if(strcmp(valor,"inverse frequency weights")==0){ Options.Weights.Method= INVERSE_FREQUENCY_WEIGHTS;  Options.Weights.Parameter= 1.0;}
else  if(strcmp(valor,"costs from matrix")==0) Options.Weights.Method= COSTS_FROM_MATRIX;
else {cout<<"weights method does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcCDP(char *valor,options &Options)
{
if(strcmp(valor,"direct")==0)Options.Cost_Derived_Probability_Method = DIRECT;
else  if(strcmp(valor,"with smoothing")==0) Options.Cost_Derived_Probability_Method = WITH_SMOOTHING;
else {cout<<"cost derived probability method does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcMSC(char *valor,options &Options) {
  if(strcmp(valor,"no combination")==0) Options.Comb.Comb=NO_COMBINATION;	
  else  if(strcmp(valor,"cross coverage combination")==0)   Options.Comb.Comb= CROSS_COVERAGE_COMBINATION;
  else  if(strcmp(valor,"majority crisp")==0)  Options.Comb.Comb= MAJORITY_CRISP;
  else  if(strcmp(valor,"majority absolute stochastic")==0) Options.Comb.Comb= MAJORITY_ABSOLUT_STOCHASTIC;
  else  if(strcmp(valor,"majority relative stochastic")==0) Options.Comb.Comb= MAJORITY_RELATIVE_STOCHASTIC;
  else  if(strcmp(valor,"majority cost stochastic")==0) Options.Comb.Comb= MAJORITY_COST_STOCHASTIC;
  else {cout<<"multitree combination method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcCST(char *valor,options &Options) {
  if(strcmp(valor,"true")==0)    Options.Comb.Smoothing= true;	
  else  if(strcmp(valor,"false")==0)      Options.Comb.Smoothing= false;
  else {cout<<"multitree combination smoothing must be true or false: " << valor << endl; exit(0);;}
}


void funcCVE(char *valor,options &Options) {
  if(strcmp(valor,"absolute")==0) Options.Comb.CombinationVector=ABSOLUTE;	
  else  if(strcmp(valor,"relative")==0)   Options.Comb.CombinationVector= RELATIVE;
  else {cout<<"combination vector unknown in the option file: " << valor << endl; exit(0);;}
}


void funcCVM(char *valor,options &Options) {
  if (strcmp(valor,"original")==0) Options.Comb.CombinationVectorMethod=ORIGINAL;	
  else  if(strcmp(valor,"good loser")==0)   Options.Comb.CombinationVectorMethod= GOOD_LOSER;
  else  if(strcmp(valor,"bad loser")==0)  Options.Comb.CombinationVectorMethod= BAD_LOSER;
  else  if(strcmp(valor,"difference")==0) Options.Comb.CombinationVectorMethod= DIFFERENCE;
  else  if(strcmp(valor,"majority")==0) Options.Comb.CombinationVectorMethod= MAJORITY;
  else  if(strcmp(valor,"squared")==0) Options.Comb.CombinationVectorMethod= SQUARED;
  else {cout<<"combination vector method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcCFM(char *valor,options &Options) {
  if (strcmp(valor,"sum")==0) Options.Comb.FusionMethod=SUM;	
  else  if(strcmp(valor,"prod")==0)   Options.Comb.FusionMethod= PROD;
  else  if(strcmp(valor,"arithmean")==0)  Options.Comb.FusionMethod= ARITHMEAN;
  else  if(strcmp(valor,"geomean")==0) Options.Comb.FusionMethod= GEOMEAN;
  else  if(strcmp(valor,"max")==0) Options.Comb.FusionMethod= MAXIMUM;
  else  if(strcmp(valor,"min")==0) Options.Comb.FusionMethod= MINIMUM;
  else  if(strcmp(valor,"median")==0) Options.Comb.FusionMethod= MEDIAN;
  else {cout<<"combination fusion method does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcC2S(char *valor,options &Options) {
  if (strcmp(valor,"no extraction")==0) Options.Comb2Single.Method = NO_EXTRACTION;	
  else  if(strcmp(valor,"invented dataset")==0)   Options.Comb2Single.Method = INVENTED_DATASET;
  else {cout << "combination to single solution method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcIDL(char *valor,options &Options) {
  long prof;
  int i;
  for(i=0;valor[i];i++)
   if (!isdigit(valor[i])) {
     cout<<"invalid number for invented dataset length"<<endl;exit(1);
   }
  prof=strtol(valor, (char **)NULL, 10);
  Options.Comb2Single.InventedDataSetLength = prof;
}


void funcSIM2(char *valor,options &Options) {

  double val;

  val=strtod(valor, (char **)NULL);
  Options.Comb2Single.SimilarityImportanceFactor  =  val;	
}


void funcTIM(char *valor,options &Options) {

  double val;

  val=strtod(valor, (char **)NULL);
  Options.Comb2Single.TestCostImportanceFactor  =  val;	
}


void funcOIM(char *valor,options &Options) {

  double val;

  val=strtod(valor, (char **)NULL);
  Options.Comb2Single.OccamImportanceFactor  =  val;	
}


void funcOUT(char *valor,options &Options)
{
if(strcmp(valor,"functional-logic")==0) Options.Output.ShowRulesMode= FUNCLOGIC;
else {cout<<"output method does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcCDT(char *valor,options &Options)
{
if(strcmp(valor,"no")==0) Options.Output.ShowClassDist= NO;
else {cout<<"output class selection not valid: " << valor << endl; exit(0);;}
}	






void funcSAR(char *valor,options &Options)
{
if(strcmp(valor,"no show")==0) Options.Output.ShowAllMultiTreeRules= NO_SHOW_ALL_RULES;
else if(strcmp(valor,"show")==0) Options.Output.ShowAllMultiTreeRules= SHOW_ALL_RULES;
else if(strcmp(valor,"to file")==0) Options.Output.ShowAllMultiTreeRules= ALL_FILES_TO_FILE;
else {cout<<"show all multitree rules not valid: " << valor << endl; exit(0);;}
}	

void funcSAK(char *valor,options &Options)
{
if(strcmp(valor,"no show")==0) Options.Output.ShowAllKBestSolutions= NO_SHOW_SOLUTIONS;
else if(strcmp(valor,"show")==0) Options.Output.ShowAllKBestSolutions= SHOW_SOLUTIONS;
else if(strcmp(valor,"to file")==0) Options.Output.ShowAllKBestSolutions= ALL_SOLUTIONS_TO_FILE;
else {cout<<"show all k-best solutions option not valid: " << valor << endl; exit(0);;}
}	


void funcSSC(char *valor,options &Options)
{
if(strcmp(valor,"no show")==0) Options.Output.ShowSolutionsComponents= NO_SHOW_COMPONENTS;
else if(strcmp(valor,"show")==0) Options.Output.ShowSolutionsComponents= SHOW_COMPONENTS;
else {cout<<"show solutions components option not valid: " << valor << endl; exit(0);;}
}	

void funcSST(char *valor,options &Options) {
  if(strcmp(valor,"absolute statistics")==0) Options.Output.ShowStatistics= ABSOLUTE_STATISTICS;
  else if(strcmp(valor,"relative statistics")==0) Options.Output.ShowStatistics= RELATIVE_STATISTICS;
  else if(strcmp(valor,"both statistics")==0) Options.Output.ShowStatistics= BOTH_STATISTICS;
   else if(strcmp(valor,"just accuracy")==0) Options.Output.ShowStatistics= JUST_ACCURACY;
 else {cout<<"show statistics option not valid: " << valor << endl; exit(0);;}
}	



void funcSCM(char *valor,options &Options)
{
if(strcmp(valor,"no show")==0) Options.Output.ShowConfusionMatrix= NO_SHOW_CONFUSION_MATRIX;
else if(strcmp(valor,"show")==0) Options.Output.ShowConfusionMatrix= SHOW_CONFUSION_MATRIX;
else if(strcmp(valor,"show only if costs")==0) Options.Output.ShowConfusionMatrix= SHOW_ONLY_IF_COSTS;
else {cout<<"show confusion matrix option not valid: " << valor << endl; exit(0);;}
}	




void funcCRP(char *valor,options &Options)
{
if(strcmp(valor,"no")==0) Options.Output.ComputeROCPoints= NO_COMPUTE_ROC_POINTS;
else if(strcmp(valor,"yes")==0) Options.Output.ComputeROCPoints= COMPUTE_ROC_POINTS;
else {cout<<"compute ROC points not valid: " << valor << endl; exit(0);;}
}	

void funcSRP(char *valor,options &Options)
{
if(strcmp(valor,"no")==0) Options.Output.ShowROCPoints= NO_SHOW_ROC_POINTS;
else if(strcmp(valor,"yes")==0) Options.Output.ShowROCPoints= SHOW_ROC_POINTS;
else {cout<<"show ROC points not valid: " << valor << endl; exit(0);;}
}	

void funcCRA(char *valor,options &Options)
{
if(strcmp(valor,"no")==0) Options.Output.ComputeROCArea= NO_COMPUTE_ROC_AREA;
else if(strcmp(valor,"yes")==0) Options.Output.ComputeROCArea= COMPUTE_ROC_AREA_BY_HAND;
else {cout<<"compute ROC area: " << valor << endl; exit(0);}
}	

void funcGRF(char *valor,options &Options)
{
if(strcmp(valor,"no")==0) Options.Output.GenerateROCCurveFile= NO_GENERATE_ROC_CURVE_FILE;
else if(strcmp(valor,"yes")==0) Options.Output.GenerateROCCurveFile= GENERATE_ROC_CURVE_FILE;
else {cout<<"generate ROC curve file: " << valor << endl; exit(0);;}
}	





void funcSKB(char *valor,options &Options)
{
if(strcmp(valor,"k-best less visited")==0)Options.SelectKBest.Met = K_BEST_LESS_VISITED;
else  if(strcmp(valor,"k-best less visited plus")==0) Options.SelectKBest.Met = K_BEST_LESS_VISITED_PLUS; // Not implemented
else  if(strcmp(valor,"k-best less visited then different components")==0) Options.SelectKBest.Met = K_BEST_LESS_VISITED_THEN_DIFFERENT_COMPONENTS;
else  if(strcmp(valor,"k-best different components")==0) Options.SelectKBest.Met = K_BEST_DIFFERENT_COMPONENTS;
else  if(strcmp(valor,"k-best random")==0) Options.SelectKBest.Met = K_BEST_RANDOM;
else {cout<<"k-best criterion does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcKBN(char *valor,options &Options)
{
int prof,i;
for(i=0;valor[i];i++)
 if (!isdigit(valor[i])) {cout<<"invalid k-best maximum number of solutions"<<endl;exit(1);}
prof=strtol(valor, (char **)NULL, 10);
Options.SelectKBest.NumSolutions = prof;
}

	void funcSCG(char *valor,options &Options)
{
if(strcmp(valor,"component accumulate")==0)Options.SelectKBest.ComponentsGeneration = COMP_ACCUMULATE;
else  if(strcmp(valor,"component random generated from start")==0)   Options.SelectKBest.ComponentsGeneration = COMP_RANDOM_GENERATE_FROM_START;
else  if(strcmp(valor,"component random generated from second")==0) Options.SelectKBest.ComponentsGeneration = COMP_RANDOM_GENERATE_FROM_SECOND;
else  if(strcmp(valor,"k-best different components")==0) Options.SelectKBest.Met = K_BEST_DIFFERENT_COMPONENTS;
else  if(strcmp(valor,"k-best random")==0) Options.SelectKBest.Met = K_BEST_RANDOM;

else {cout<<"k-best criterion does not exist in the option file: " << valor << endl; exit(0);;}
}




// ESTOS FACTORS TINDREM QUE VORE QUE FEM AMB ELLS


void funcCRF(char *valor,options &Options)
{

double val;

val=strtod(valor, (char **)NULL);
Options.ComponentsRandomFactor= val;
}



void funcMRC(char *valor,options &Options)
{
if(strcmp(valor,"true")==0)   Options.Components_Reckoning= true; // true if component matrix is to be used. Put false instead.
else  if(strcmp(valor,"false")==0) Options.Components_Reckoning= false;
else {cout<<"multitree components reckoning method does not exist in the option file: : " << valor << endl; exit(0);;}
}


void funcSFR(char *valor,options &Options)
{

double val;

val=strtod(valor, (char **)NULL);
Options.SelectKBest.FactorForRandomComponents= val;	
}

//
//

void funcSNS(char *valor,options &Options)
{
int i;
int val;
for(i=0;valor[i];i++)
 if (!isdigit(valor[i])) {cout<<"invalid k best number of solutions"<<endl;exit(1);}
val=strtol(valor, (char **)NULL, 10);
Options.SelectKBest.NumSolutions =  val;	
}




void funcSTS(char *valor,options &Options) {
  if (strcmp(valor,"no sample training set")==0)   Options.SampleTrainingSet= NO_SAMPLE_TRAINING_SET;
  else  if(strcmp(valor,"sample training set")==0) Options.SampleTrainingSet= SAMPLE_TRAINING_SET;
  else {cout<<"sample training set option does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcSTP(char *valor,options &Options) {

  double val;

  val=strtod(valor, (char **)NULL);
  if ((val <= 0) || (val > 1))  {
     cout<<"invalid sample training set proportion: it must be between 0 (strictly) and 1"<<endl;
     exit(1);
  }
  Options.SampleTrainingSetProportion =  val;	
}



void funcCRV(char *valor,options &Options) {
  if (strcmp(valor,"use separate test set")==0)   Options.CrossValidation= USE_SEPARATE_TEST_SET;
  else if(strcmp(valor,"cross validation")==0) Options.CrossValidation= CROSS_VALIDATION;
  else if(strcmp(valor,"kfold cross validation")==0) Options.CrossValidation= KFOLD_CROSS_VALIDATION;
  else if(strcmp(valor,"repeated kfold cross validation")==0) Options.CrossValidation= REPEATED_KFOLD_CROSS_VALIDATION;
  else {cout<<"cross validation option does not exist in the option file: " << valor << endl; exit(0);;}
}

void funcKFV(char *valor,options &Options) {
  int prof,i;
  for(i=0;valor[i];i++)
   if (!isdigit(valor[i])) {
     cout<<"invalid number for k-fold cross validation"<<endl;exit(1);
   }
  prof=strtol(valor, (char **)NULL, 10);
  Options.kFoldValidation = prof;
}

void funcRKF(char *valor,options &Options) {
  int prof,i;
  for(i=0;valor[i];i++)
   if (!isdigit(valor[i])) {
     cout<<"invalid number for repeat k-fold cross validation"<<endl;exit(1);
   }
  prof=strtol(valor, (char **)NULL, 10);
  Options.RepeatKFold = prof;
}




void funcMNV(char *valor,options &Options) {
  if (strcmp(valor,"ignore examples with missing numeric values")==0)   Options.MissingNumericValues= IGNORE_EXAMPLES_WITH_MISSING_NUMERIC_VALUES;
  else if(strcmp(valor,"substitute missing numeric values")==0) Options.MissingNumericValues= SUBSTITUTE_MISSING_NUMERIC_VALUES;
  else {cout<<"missing numeric values option does not exist in the option file: " << valor << endl; exit(0);;}
}


void funcSMS(char *valor,options &Options) {
  if (strcmp(valor,"yes")==0)   Options.ShowNumberSolutions= SHOW_NUMBER_SOLUTIONS;
  else if(strcmp(valor,"no")==0) Options.ShowNumberSolutions= DO_NOT_SHOW_NUMBER_SOLUTIONS;
  else {cout<<"show number possible solutions of the multitree option does not exist in the options file: " << valor << endl; exit(0);;}
}




void funcTWC(char *valor,options &Options) {
  if (strcmp(valor,"yes")==0)
    Options.AllowTestWithoutOneClass= true;
  else if(strcmp(valor,"no")==0)
    Options.AllowTestWithoutOneClass= false;
  else {
    cout << "allow test without one class option does not exist in the option file: " << valor << endl;
    exit(0);
  }
}


void funcPIC(char *valor,options &Options) {
  if (strcmp(valor,"yes")==0)
    Options.Comb.PruneInCombination= true;
  else if(strcmp(valor,"no")==0)
    Options.Comb.PruneInCombination= false;
  else {
    cout << "post-pruning in combination option does not exist in the option file: " << valor << endl;
    exit(0);
  }
}


void funcSIM(char *valor,options &Options) {
  if (strcmp(valor,"kappa")==0)
    Options.Comb2Single.SimilarityMethod = KAPPA;
  else if(strcmp(valor,"kappa1")==0)
    Options.Comb2Single.SimilarityMethod = KAPPA1;
  else if(strcmp(valor,"qstat")==0)
    Options.Comb2Single.SimilarityMethod = QSTAT;
  else {
    cout << "similarity method for combination to single option does not exist in the option file: " << valor << endl;
    exit(0);
  }
}




void tracta_linia(char *tira,options &opcions)
{
 char *op,*valor;
 // separem el token en valor i opcio
 valor=strchr(tira,'=');
	if(valor==NULL)
		{ printf("Error in options file. Looking for a '='."); exit(1); }

		
	// SE HAURIEN DE CONTROLLAR ELS BLANCS QUE PUGUEN HI HAURE DARRERE	
		
	op=valor;
	valor++;
	*op='\0';
	
	 if (strcmp(_EEM,tira)==0) funcEEM(valor,opcions);
	 else if (strcmp(_SMM,tira)==0) funcSMM(valor,opcions);
	 else if (strcmp(_NSM,tira)==0) funcNSM(valor,opcions);
	 else if (strcmp(_PSP,tira)==0) funcPSP(valor,opcions);
	 else if (strcmp(_OUT,tira)==0) funcOUT(valor,opcions);
//	 else if (strcmp(_MCN,tira)==0) funcMCN(valor,opcions);
   else if (strcmp(_SNF,tira)==0) funcSNF(valor,opcions);
   else if (strcmp(_SNV,tira)==0) funcSNV(valor,opcions);
	 else if (strcmp(_MTR,tira)==0) funcMTR(valor,opcions);
	 else if (strcmp(_MBT,tira)==0) funcMBT(valor,opcions);
	 else if (strcmp(_CRF,tira)==0) funcCRF(valor,opcions);
	 else if (strcmp(_MSC,tira)==0) funcMSC(valor,opcions);
	 else if (strcmp(_CST,tira)==0) funcCST(valor,opcions);
	 else if (strcmp(_CVE,tira)==0) funcCVE(valor,opcions);
	 else if (strcmp(_CVM,tira)==0) funcCVM(valor,opcions);
	 else if (strcmp(_CFM,tira)==0) funcCFM(valor,opcions);
		 else if (strcmp(_C2S,tira)==0) funcC2S(valor,opcions);
	 else if (strcmp(_IDL,tira)==0) funcIDL(valor,opcions);
	 else if (strcmp(_SIM2,tira)==0) funcSIM2(valor,opcions);
	 else if (strcmp(_TIM,tira)==0) funcTIM(valor,opcions);
	 else if (strcmp(_OIM,tira)==0) funcOIM(valor,opcions);
	
	 else if (strcmp(_CSM,tira)==0) funcCSM(valor,opcions);
	 else if (strcmp(_REC,tira)==0) funcREC(valor,opcions);
   else if (strcmp(_CPV,tira)==0) funcCPV(valor,opcions);
   else if (strcmp(_MRC,tira)==0) funcMRC(valor,opcions);
    else if (strcmp(_STO,tira)==0) funcSTO(valor,opcions);
  // else if (strcmp(_SP1,tira)==0) funcSP1(valor,opcions);
  // else if (strcmp(_SP2,tira)==0) funcSP2(valor,opcions);
  // else if (strcmp(_SP3,tira)==0) funcSP3(valor,opcions);
  // else if (strcmp(_SP4,tira)==0) funcSP4(valor,opcions);
   else if (strcmp(_NMI,tira)==0) funcNMI(valor,opcions);
	 else if (strcmp(_SPC,tira)==0) funcSPC(valor,opcions);
	 else if (strcmp(_SKB,tira)==0) funcSKB(valor,opcions);
	 else if (strcmp(_SCG,tira)==0) funcSCG(valor,opcions);

   else if (strcmp(_KBN,tira)==0) funcKBN(valor,opcions);

	 else if (strcmp(_SFR,tira)==0) funcSFR(valor,opcions);
	 else if (strcmp(_PPM,tira)==0) funcPPM(valor,opcions);
	 else if (strcmp(_PST,tira)==0) funcPST(valor,opcions);
	 else if (strcmp(_SPP,tira)==0) funcSPP(valor,opcions);
	 else if (strcmp(_TCM,tira)==0) funcTCM(valor,opcions);
	 else if (strcmp(_TCU,tira)==0) funcTCU(valor,opcions);
	 else if (strcmp(_WMT,tira)==0) funcWMT(valor,opcions);
	 else if (strcmp(_CDP,tira)==0) funcCDP(valor,opcions);
	 else if (strcmp(_CDT,tira)==0) funcCDT(valor,opcions);
//	 else if (strcmp(_MCM,tira)==0) funcMCM(valor,opcions);
   else if (strcmp(_SNS,tira)==0) funcSNS(valor,opcions);

   else if (strcmp(_SAR,tira)==0) funcSAR(valor,opcions);
   else if (strcmp(_SAK,tira)==0) funcSAK(valor,opcions);
   else if (strcmp(_SSC,tira)==0) funcSSC(valor,opcions);
   else if (strcmp(_SST,tira)==0) funcSST(valor,opcions);
   else if (strcmp(_SCM,tira)==0) funcSCM(valor,opcions);
   else if (strcmp(_CRP,tira)==0) funcCRP(valor,opcions);
   else if (strcmp(_SRP,tira)==0) funcSRP(valor,opcions);
   else if (strcmp(_CRA,tira)==0) funcCRA(valor,opcions);
   else if (strcmp(_GRF,tira)==0) funcGRF(valor,opcions);

   else if (strcmp(_STS,tira)==0) funcSTS(valor,opcions);
   else if (strcmp(_STP,tira)==0) funcSTP(valor,opcions);
   else if (strcmp(_CRV,tira)==0) funcCRV(valor,opcions);
   else if (strcmp(_KFV,tira)==0) funcKFV(valor,opcions);
   else if (strcmp(_MNV,tira)==0) funcMNV(valor,opcions);
   else if (strcmp(_SMS,tira)==0) funcSMS(valor,opcions);
   else if (strcmp(_RKF,tira)==0) funcRKF(valor,opcions);

   else if (strcmp(_TWC,tira)==0) funcTWC(valor,opcions);
   else if (strcmp(_PIC,tira)==0) funcPIC(valor,opcions);
   else if (strcmp(_SIM,tira)==0) funcSIM(valor,opcions);

   else {cout<<"option unknown: " << tira << endl;exit(0);}
	}

void ompli_options(options &opcions)
{

 FILE *in=NULL;
// char filein[800];
 char exemple[EXAMPLESIZE];
// word i=1;
// bool iswe=false;
 //strcpy (filein,fitxer);

	
  if ((in=fopen(OFILE,"rt"))==NULL)
	{ printf("Error opening options\n"); exit(1); }

   // tracta_exemple(exemple);
  while(fgets((char *) exemple,EXAMPLESIZE,in))
    {
    exemple[strlen(exemple)-1]='\0';// eliminem /n
    if (exemple[strlen(exemple)-1]=='\r') exemple[strlen(exemple)-1]='\0';// eliminem /r si hi ha
      	
      if (exemple[0]!=COMENTARI) tracta_linia(exemple,opcions);
    }
 fclose(in);
}

