/***************************************************************************
                          parser.cpp  -  description
                             -------------------

    functionality        : parser for the dataset files
    relevant classes     :
    relevant functions   : ompli_memoria

    begin                : 9 Oct 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "parser.h"
#define EXAMPLESIZE 8000

#define MISSING_VALUE '?'

#include "options.h"

enum  what_to_do_with_numeric_missing_values { SUBSTITUTE_MISSING_VALUES, IGNORE_EXAMPLE };

what_to_do_with_numeric_missing_values What_To_Do_With_Numeric_Missing_Values; // Local variable to the .cpp


/*****************************************************************************************************/
// i assigna els tipus (!TYPES line)
void calcula_tamany_exemple(char *exemple, dataset &dset)
{
  char *aux,buf[EXAMPLESIZE],*tok;
  int i=0,tip;
  argtable taula;
  //  printf (" tamanys n:%d,w:%d,b:%d",snumeric,sword,sbyte);
  aux=buf;
  
  strcpy(aux,exemple);
  while ((tok=strtok(aux,","))!=NULL)
    {
       if (i==0) {
       			if (strncmp("!TYPES:",tok,7)!=0)
            					{
            						cout<<"Error: first line does not contain the types"<<endl;
            						exit(0);
            					}
            				}
        i++;
      dset.num_arguments++ ;
      aux=NULL;
      // creem una entrada a la taula dels indexs
      //taula=ne w(argtable);
      dset.exetable.push_back(taula);
    }
  i=0;
  aux=buf;
  strcpy(aux,&exemple[7]);
  //dset.argoffset = new (argoffsett) [dset.num_arguments];
  dset.size_example= 0;
  while ((tok=strtok(aux,","))!=NULL) {
    dset.argoffset.push_back(dset.size_example);

    if ((tok[0]=='I') || (tok[0]=='i'))
	  {/* es numeric*/
			dset.profile.types.push_back(IGNORE_KIND);	
	    dset.profile.kinds.push_back(IGNORE_KIND);
	    dset.size_example+=SIGNORE;	
    }
    else if(tok[0]=='0')
	  {/* es numeric*/
		  if (tok[1]=='U' || tok[1]=='u') {
	      dset.profile.types.push_back(NUMERIC_WITH_UNKNOWN_KIND);	
	     	dset.profile.kinds.push_back(NUMERIC_WITH_UNKNOWN_KIND);
	      dset.size_example+=SNUMERIC;		
		  }
		  else {
		  	dset.profile.types.push_back(NUMERIC_KIND);	
	     	dset.profile.kinds.push_back(NUMERIC_KIND);
	      dset.size_example+=SNUMERIC;	
	    }			
	  }
    else /* no numeric*/
	  {
	    dset.profile.kinds.push_back(DISCRETE_KIND);
	    tip=atoi(tok);
	    dset.profile.types.push_back(tip);
	    dset.size_example+=SBYTE;
	  }
    // printf("\n desplaament: %d",(*dset).argoffset[i]);
    aux=NULL;
    i++;
  }
}

/*****************************************************************************************************/
// busca en la llista de simbols si existeix
// si no els afegeix al final
byte is_index(argtable *taula, char *arg)
{
  unsigned int cont;
  
  for (cont=0;cont<(*taula).size();)
    { 
      if ((*taula)[cont]==arg) 
	{
          return cont;
	}
      cont++;
    }
  // printf (" %d, size %d ",cont,(*taula).size());
  if (cont >= 255) {
    cout << endl << endl << "ERROR in function is_index in parser.cpp: " << endl;
    cout << " This dataset has more than 256 different values for one attribute" << endl;
    cout << " This system is designed to handle nominal values with at least 256 values." << endl;
    exit(-1);
  }
  (*taula).push_back(arg);
  return cont;
}


/*****************************************************************************************************/

// retorna el index corresponent a eixe simbol i argument, si no existeix, crea
// un index nou actualitzan la taula exetab
byte troba_index(char * arg,int numarg, exetablet &exetab)
{
  
  //  int i,cont=1;
  argtable *taula;
  byte res;
  
  taula=&(exetab[numarg]);
  if ((*taula).size()==0)
    {   
      /* empty for this argument*/
      (*taula).push_back(arg);
      return 0;
    }
  else
    {
      res=is_index(taula,arg);
    }
  return res;
}


/***********************************************************************/
// Troba equivalncies de classes
string Class_Of(string Classe, dataset &dset) {

  for (unsigned int i=0; i< dset.class_equivalence.size(); i++) {
    pair<string,string> P= dset.class_equivalence[i];
    if (P.first == Classe)
      return P.second;
  }

  return Classe; // Not found, we return as it was

}


/*****************************************************/

bool Valid_Line(char *exemple) {

  if ((exemple[0] == '!') || (exemple[0] == '%'))
    return false;

  int i=0;
  while(exemple[i]) {
    if (exemple[i] == ',')   // at least one comma
      return true;
    i++;
  }
  return false;
}


/*****************************************************************************************************/
// Ompli la memoria i dinamicament assigna indexs als diferents valors dels arguments
// tamb recull informaci de la distribuci de la classe
bool tracta_exemple(char *exemple, dataset &dset)
{
  char *aux,buf[EXAMPLESIZE],*tok;
  word  i=0,ind;
  //  numeric ftemp;
//  example tex(dset.size_example);
  example &tex = *new example(dset.size_example);
  // tex.v= new (byte) [dset.size_example];


  // printf ("cadena %s",exemple);
  strcpy(buf,exemple);
  aux=buf;

  bool ClassAssigned= false;

  //printf ("\n cadena %s",exemple);
  while ((tok=strtok(aux,","))!=NULL)
  {
  	if (i==dset.num_arguments)
    {
    	if (tok[0]!=13)
     	{
   	  	cout << "Error in train data: Too many arguments. Argument " << i << " already read and more commas found" <<endl;
   	  	exit(-1);
     	}
    }
    else
    {
     	if ((dset.profile.kinds[i]==NUMERIC_KIND) || (dset.profile.kinds[i]==NUMERIC_WITH_UNKNOWN_KIND))
		  {/* es numeric*/
		    numeric V;
		
		    if (tok[0] == MISSING_VALUE) {
		      if (dset.profile.kinds[i]==NUMERIC_WITH_UNKNOWN_KIND) {
		        V= MISSING_NUMERIC; // V= numeric_limits<float>::signaling_NaN();
		        cout << "Missing value in numeric argument (" << i << ") found in trainset file: taken into account. Numeric splits will have three children" << endl;
		      }
		      else {
		        if (What_To_Do_With_Numeric_Missing_Values == SUBSTITUTE_MISSING_VALUES) {
		          cout << "Missing value in numeric argument (" << i << ") found in trainset file: substituted by 0.0" << endl;
		          V= 0.0;
		        }
		        else {
		          cout << "Missing value in numeric argument (" << i << ") found in trainset file: whole example ignored" << endl;
			        if(tex.v != NULL)	
                delete [] tex.v; // this is done in this way because the example destructor doesn't free the memory.
			        delete &tex;
			        return false;   // Ignores the example
			      }
			    }
		    }
		    else {
		      V= (numeric)atof(tok);
		    }
			  *(numeric*)&(tex.v[dset.argoffset[i]])=V;
		  }
     	else if (dset.profile.kinds[i]==IGNORE_KIND)
		  {/* es IGNORE kind*/
		    // nothing to do
      }
      else /* no numeric, guardem com un byte*/
		  {	
		    char CName[100];
		
		    if (i==dset.num_arguments-1) { // si es la classe mirem les equivalencies
			    string tok2= Class_Of(tok, dset);
			    strcpy(CName, tok2.c_str());
        }
        else {
          strcpy(CName, tok);
        }
		

			  ind=troba_index(CName,i,dset.exetable);
	  		tex.v[dset.argoffset[i]]=ind;
			
			  if (i==dset.num_arguments-1) // si es la classe magatzemem la distr
				{
				  ClassAssigned= true;
				  if (ind==dset.vdist.size())
					   dset.vdist.push_back(1);
				  else dset.vdist[ind]=dset.vdist[ind]+1;
				}
		  }
      aux=NULL;
      i++;
    }
  }


  dset.data.push_back(tex);

  delete &tex;  // doesn't destroy the tex.v space, just the shell of the object.


  if (!ClassAssigned) {
    cout << endl << "An example does not contain the class. It only has " << i << " arguments (separated by commas)" << endl;
    cout << endl << "The line is: " << endl << exemple << endl;

    exit(-1);
  }

  return true;
}
/*****************************************************************************************************/

void ompli_types(dataset &dset)
{
  int i=0;
  for (i=0;i<dset.num_arguments;i++)
     dset.profile.mvalues.push_back(dset.exetable[i].size());

}

/*****************************************************************************************************/

// retorna el index corresponent a eixe simbol i argument, si no existeix,
// mostra missatge de error
byte troba_index_test(char * arg,word numarg, const exetablet &exetab)
{
  unsigned cont=1;
 const argtable *taula;
 //  byte res;
  if (numarg>exetab.size())
  	{
  		cout<<" \n ERROR incorrect number of arguments in test data ";
  		exit(-1);
  	}
  else
  	{
  		taula=&(exetab[numarg]);
  		 for (cont=0;cont<(*taula).size();)
    			{
		      if ((*taula)[cont]==arg) return cont;
		              cont++;
    			}
	  	cout<<" \n ERROR unknowed symbol :"<<arg<<"in argument "<<numarg;
  		exit(-1);
	}
}
/*****************************************************************************************************/
// Ompli la memoria a partir del indexs que es passen al camp exetable de dste
// si apareix un simbol que no est a la taula dels indexs mostrem error
// normalment s'utilitzar per a omplir test data sets
// tamb recull informaci de la distribuci de la classe
bool tracta_exemple_test(char *exemple,  dataset &dset)
{
   char *aux,buf[EXAMPLESIZE],*tok;
   word i=0,ind;
   //  numeric ftemp;
  example &tex = *new example(dset.size_example);
  // tex.v= new (byte) [dset.size_example];

  bool ClassAssigned= false;

  // printf ("cadena %s",exemple);
  strcpy(buf,exemple);
  aux=buf;
  //printf ("\n cadena %s",exemple);
  while ((tok=strtok(aux,","))!=NULL)
    {
    	if (i==dset.num_arguments)
    		  {
    		  	if (tok[0]!=13)
    		  	{
	            cout << "Error in train data: Too many arguments. Argument " << i << " already read and more commas found" <<endl;
   	
    		  	exit(0);
    		  	}
    		  }
      else
      {
      	 if (dset.profile.kinds[i]==NUMERIC_KIND)
	 	{/* es numeric*/
			  numeric V;
		
	    if (tok[0] == MISSING_VALUE) {
		    if (dset.profile.kinds[i]==NUMERIC_WITH_UNKNOWN_KIND) {
		      V= MISSING_NUMERIC; // V= numeric_limits<float>::signaling_NaN();
		      cout << "Missing value in numeric argument (" << i << ") found in trainset file: taken into account. Numeric splits will have three children" << endl;
		    }
		    else {
   	      if (What_To_Do_With_Numeric_Missing_Values == SUBSTITUTE_MISSING_VALUES) {
		        cout << "Missing value in numeric argument (" << i << ") found in testset file: substituted by 0.0" << endl;
		        V= 0.0;
		      }
		      else {
		        cout << "Missing value in numeric argument (" << i << ") found in testset file: whole example ignored" << endl;
			      return false;   // Ignores the example
			    }
			  }
		  }
		  else {
		    V= (numeric)atof(tok);
		  }
		
		  *(numeric*)&(tex.v[dset.argoffset[i]])=V;
	 		 	}
	 		
	 		else if (dset.profile.kinds[i]==IGNORE_KIND)
		  {/* es IGNORE kind*/
		    // nothing to do
      }
	 		 	
       	else /* no numeric, guardem com un byte*/
	 	{	
	 	
	 	
		   char CName[100];
		
		   if (i==dset.num_arguments-1) { // si es la classe mirem les equivalencies
			   string tok2= Class_Of(tok, dset);
			   strcpy(CName, tok2.c_str());
       }
       else {
         strcpy(CName, tok);
       }
		
			ind=dset.GetIndex(i,CName);
	 	  	tex.v[dset.argoffset[i]]=ind;
	 		if (i==dset.num_arguments-1) // si es la classe magaztenem la distribuci
				{
				dset.vdist[ind]=dset.vdist[ind]+1;
				ClassAssigned= true;
				}
		}
      	aux=NULL;
      	i++;
      	}
    }
  dset.data.push_back(tex);


  if (!ClassAssigned) {
    cout << endl << "An example does not contain the class. It only has " << i << " arguments (separated by commas)" << endl;
    cout << endl << "The line is: " << endl << exemple << endl;

    exit(-1);
  }

  return true;
}

/*****************************************************************************************************/
// ompli la memoria de un data set a partir de un fitxer i un altre dataset
// normalment se utilitzar per a omplir un dataset de un test
void ompli_memoria(const string & fitxer,dataset  & dset, const dataset  & dtrain, const options &Options  )
{
   if (Options.MissingNumericValues == IGNORE_EXAMPLES_WITH_MISSING_NUMERIC_VALUES) {
     What_To_Do_With_Numeric_Missing_Values = IGNORE_EXAMPLE;
   }
   else if (Options.MissingNumericValues == SUBSTITUTE_MISSING_NUMERIC_VALUES) {
     What_To_Do_With_Numeric_Missing_Values = SUBSTITUTE_MISSING_VALUES;
   }


FILE *in=NULL;
char filein[800];
 char exemple[EXAMPLESIZE];
 unsigned int i=0;
  //copiem la infomaci del dataset de train al de test
	dset.argoffset=dtrain.argoffset;;
	dset.exetable=dtrain.exetable;
	dset.num_arguments=dtrain.num_arguments;
	dset.profile=dtrain.profile;
	dset.size_example=dtrain.size_example;
 // tracta_exemple(exemple);
//  strcpy (filein,fitxer);
  for (i=0;i<fitxer.size();i++)
       filein[i]=fitxer[i];
    filein[i]='\0';
 if (filein[0])
   {
    if ((in=fopen(filein,"rt"))==NULL)
	{ cout << "Error when opening file : \"" << filein << "\"" << endl;
	  exit(1);
	   }
//   fgets((char *) exemple,EXAMPLESIZE,in);
   }
  while(fgets((char *) exemple,EXAMPLESIZE,in))
    {
     exemple[strlen(exemple)-1]='\0';// eliminem /n
     if (exemple[strlen(exemple)-1]=='\r') exemple[strlen(exemple)-1]='\0';// eliminem /r si hi h
      // (*dset).num_exemples++;
     if (Valid_Line(exemple))
       tracta_exemple_test(exemple,dset);
    }
 fclose(in);
// ompli_types(dset);
	//for (i=0;i<(*dset).vdist.size();i++)
 	 // printf(" \n classe %d: %d",i,(*dset).vdist[i]);
  /* per cada example el fica en memoria
     assignant indexs al valors
     i controlant la taula dels indexs*/
  /* el primer exemple mirem lo que ocupa*/
}
/*******************************************************************************************/
// Ompli la informaci sobre el noms dels arguments i classes

void set_names (char *exemple, dataset & dset)
{
  char *aux,buf[EXAMPLESIZE],*tok;
int	i=0;
  aux=buf;
  strcpy(aux,exemple);
  while ((tok=strtok(aux,","))!=NULL)
    {
    	dset.names.push_back(tok);
    	i++;
    	aux=NULL;
    }
	if (i!=dset.NumArguments()+1) {
	cout<<"Error:  Number of Names incorrect"<<endl;
	exit(0);}
}

/*******************************************************************************************/
// Reassinga equivalncies entre classes (servix per a ajuntar classes)

void set_classes (char *exemple, dataset & dset)
{

  char *aux,*aux1=NULL,buf[EXAMPLESIZE],*tok;
  int i=0,pos=0;

  cout << endl << "!RECLASSIFY directive found in training set:" << endl;
  cout << "Some classes are to be fusioned in order to reduce the number of classes: " << endl;

  aux=buf;
  i=0;
  strcpy(aux,exemple);
  while ((tok=strtok(aux,";"))!=NULL)
    {
     aux1=strchr(tok,'>');
     if (!aux1)
     			{
     				cout<<"Error:Incorrect format of class equivalence"<<endl;
     			}
     pos=aux1-tok;// recuperem posici del '>'
     tok[pos]='\0'; // xapua per recuperar el contingut de la classe
     string Left= tok;
      tok[pos]='>';

      string Right= &tok[pos+1];

      pair<string,string> P;

      P= pair<string,string>(Left,Right);

      dset.class_equivalence.push_back(P);

      cout << "  Class \"" << Left << "\" is joined/added to class \"" << Right << "\"" << endl;
     	
     	i++;
    	aux=NULL;
    }


	
}
/*******************************************************************************************/
// Ompli la informaci sobre els pesos de les clases.

void set_weights(char *exemple, dataset & dset)
{
  char *aux,*aux1=NULL,buf[EXAMPLESIZE],*tok;
int	i=0,pos=0,ind=0;
// inicialitzaci del pesos
 for (i=0;i<dset.NumClasses();i++)   // A EST MAL PERQU NO SE SAP EL NMERO DE CLASSES
     dset.weights.push_back(0);
  aux=buf;
  i=0;
  strcpy(aux,exemple);
  while ((tok=strtok(aux,","))!=NULL)
    {
     aux1=strchr(tok,'=');
     if (!aux1)
     			{
     				cout<<"Error:Incorrect format of weight"<<endl;
     			}
     pos=aux1-tok;// recuperem posici del '='
     tok[pos]='\0'; // xapua per recuperar el contingut de la classe
     ind=dset.GetIndex(dset.NumArguments(),tok);//recuperem l'index corresponent a eixa classe
    dset.weights[ind]=(numeric)atof(&tok[pos+1]); //li assignem el pes
      tok[pos]='=';
    	i++;
    	aux=NULL;
    }
	if (i!=dset.NumClasses()) {
	cout<<"Error:  Number of weights incorrect"<<endl;
	exit(0);}
}


/*****************************************************************************************************/

// Ompli un data set amb tota la informacio a partir de un fitxer de dades
void ompli_memoria(const string & fitxer,dataset & dset, const options &Options) {
   if (Options.MissingNumericValues == IGNORE_EXAMPLES_WITH_MISSING_NUMERIC_VALUES) {
     What_To_Do_With_Numeric_Missing_Values = IGNORE_EXAMPLE;
   }
   else if (Options.MissingNumericValues == SUBSTITUTE_MISSING_NUMERIC_VALUES) {
     What_To_Do_With_Numeric_Missing_Values = SUBSTITUTE_MISSING_VALUES;
   }

   FILE *in=NULL;
   char filein[800];
   char exemple[EXAMPLESIZE],bweight[EXAMPLESIZE-9];

   bool iswe=false;
   //strcpy (filein,fitxer);
   unsigned int i;
   for (i=0;i<fitxer.size();i++)
     filein[i]=fitxer[i];
   filein[i]='\0';
   if (fitxer[0]) {
     if ((in=fopen(filein,"rt"))==NULL)	{
       cout << "Error when opening file : \"" << filein << "\"" << endl;
       exit(1);
     }
     fgets((char *) exemple,EXAMPLESIZE,in);
   }
   exemple[strlen(exemple)-1]='\0';// eliminem /n
   if (exemple[strlen(exemple)-1]=='\r')
     exemple[strlen(exemple)-1]='\0';// eliminem /r si hi ha;
   calcula_tamany_exemple(exemple, dset);  // assigna el tipus al dataset (!TYPES line)
   // tracta_exemple(exemple);
   long l=1;
   while(fgets((char *) exemple,EXAMPLESIZE,in)) {
     exemple[strlen(exemple)-1]='\0';// eliminem /n
     if (exemple[strlen(exemple)-1]=='\r')
       exemple[strlen(exemple)-1]='\0';// eliminem /r si hi h
     // (*dset).num_exemples++;
   //  if ((l==1)||(l==2)) {
       if (strncmp("!NAMES:",exemple,7)==0) {
         set_names(&exemple[7],dset);
         continue;
       }
       if (strncmp("!RECLASSIFY:",exemple,12)==0) {
         set_classes(&exemple[12],dset);
         continue;
       }
       if (strncmp("!WEIGHTS:",exemple,9)==0)	{
         strcpy(bweight,&exemple[9]);// ho guarde no tinc encara els indexs
				 iswe=true;
         continue;
       }				
   //  }
          				
     if (Valid_Line(exemple))
       tracta_exemple(exemple,dset);
   //  cout << "\rExample in line: " << l << " read from the training set file";
     l++;
   }

   cout << endl;

   fclose(in);
   ompli_types(dset);
   if (iswe)
     set_weights(bweight,dset);		// ara ja tinc els indexs ;-), omplic els pesossi hi han
   else
     dset.SetWeightsTo1();
	 //for (i=0;i<(*dset).vdist.size();i++)
 	 // printf(" \n classe %d: %d",i,(*dset).vdist[i]);
   /* per cada example el fica en memoria
     assignant indexs al valors
     i controlant la taula dels indexs*/
   /* el primer exemple mirem lo que ocupa*/
}

