/***************************************************************************
                          rules.h  -  description
                             -------------------

    functionality        : defines basics of rules
    relevant classes     : condition, rule
    relevant functions   :

    begin                : 9 Oct 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
#ifndef RULES_H
#define RULES_H

#include "estructures.h"
#include <iostream> 
#include <strstream>
// #include "exemples.h"
class dataset; // Per a evitar inclusions creuades


class condition {
public:
  cond_op Operator;
  argument Value;

  condition() : Operator(EMPTY) {
  }

  condition(cond_op o, argument v) : Operator(o), Value(v) {
  }

  // CHECK IF TWO CONDITIONS ARE ADDED OVER THE SAME ARGUMENT
  bool Specialises(const condition &c2) const {
    cond_op OldOp= c2.Operator;
      if ((OldOp == CONST_NEQ)  || (OldOp == VAR_NEQ)) {
        if (Operator == CONST_EQ)  {
          return true; // OK (encara que un <>a pot ser carregat amb un =a que fara el node buit
        }
       else if (Operator == CONST_NEQ)  {
          if ((OldOp == CONST_NEQ) && (Value.Value_d == c2.Value.Value_d) )
            return false;
          else
            return true; // OK ((esta es perfectament bona)
        }
        else if (Operator == VAR_EQ)  {
          return true; // OK ((esta es perfectament bona)
        }
        else if (Operator == VAR_NEQ)  {
          if ((OldOp == VAR_NEQ) && (Value.Value_v == c2.Value.Value_v) )
            return false;
          else
            return true; // OK ((esta es perfectament bona)
        }
      }
      else if (OldOp == NUM_LT)  {
        if ((Operator == NUM_LT)  || (Operator == NUM_GE)) {
          if  (Value.Value_n < c2.Value.Value_n)  {
            return true;
          }
        }
      }
      else if (OldOp == NUM_GE)  {
       if ((Operator == NUM_LT)  || (Operator == NUM_GE)) {
          if  (Value.Value_n > c2.Value.Value_n)  {
             return true;
          }
        }
      }
      else if (OldOp == NUM_LT_WU)  {
        if ((Operator == NUM_LT_WU)  || (Operator == NUM_GE_WU)) {
          if  (Value.Value_n < c2.Value.Value_n)  {
            return true;
          }
        }
      }
      else if (OldOp == NUM_GE_WU)  {
       if ((Operator == NUM_LT_WU)  || (Operator == NUM_GE_WU)) {
          if  (Value.Value_n > c2.Value.Value_n)  {
             return true;
          }
        }
      }
  //    else if (OldOp == NUM_MISSING)
  //      return false;  // NORES ESPECIALITZA UN VALOR MISSING
    return false;
  }
};


class rule {
public:
  vector<vector<condition> > V_condition;
private:
  int Num_arg;
public:
  rule(int n) : V_condition(n+1) {
    Num_arg= n+1;  // S'afegix 1 per al resultat de la funci
  }

  rule(const rule &r) {
    Num_arg= r.Num_arg;
    V_condition = r.V_condition;
  }

  ~rule() {
  }

  bool Add_Condition(int a, const condition &c) {
    vector<condition> &OldCondVector= V_condition[a];
    int NumCond= OldCondVector.size();
    if (NumCond >= 0) {
       for (int i=0; i<NumCond; i++) {
         if (!c.Specialises(OldCondVector[i])) {
            #ifdef DEBUG_RULES
            cout << "Error en funci Add_Condition de la classe rule. La condici s incompatible amb alguna de les anteriors o entraria en un bucle infinit" << endl;
            #endif
            return false;
         }
       }
    }
    OldCondVector.push_back(c);
    return true;
  }

  void ShowCutre() ;

  string ToStringUnfolded(char * FunctionName, const dataset & DataSet, bool Short= false) ;

  friend class dataset;
  //  friend bool Covers(const rule & r, int e);
};



#endif
