/***************************************************************************
                          trees.cpp  -  description
                             -------------------

    functionality        : largest source file with the main learning functions
    relevant classes     : tree, and_tree, or_tree, multi_tree
    relevant functions   :

    begin                : 23 Oct 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include "trees.h"


#ifdef TREE_DEBUG
long GiveMeANumber() {     // This function is to number the nodes
  static long n;
  return n++;
}
#endif

// Function for comparison to be used to select the best in the list of or children.
bool Compare_ExpectedError(tree *o1, tree *o2) {

  if ( (((or_tree *)o1)->State == PROGRESS) &&  ((or_tree *)o2)->State == SUSPENDED)
    return false;
  else  if ( (((or_tree *)o2)->State == PROGRESS) &&  ((or_tree *)o1)->State == SUSPENDED)
    return true;

  if ( ((or_tree *)o1)->BackupError >= ((or_tree *)o2)->BackupError)
// EN REALITAT CALCULA >= EN LLOC DE < PERQUE ESTEM COMPARANT L'ERROR
    return true;
  else
    return false;
}



// This function is used as a < for calling max_element for selecting the best Solution
bool Compare_AccExpected_Error(tree *o1, tree *o2)  {
  if ( (((or_tree *)o1)->State == PROGRESS) &&  ((or_tree *)o2)->State == SUSPENDED)
    return false;
  else  if ( (((or_tree *)o2)->State == PROGRESS) &&  ((or_tree *)o1)->State == SUSPENDED)
    return true;

  if ( ((or_tree *)o1)->Accumulated.AccExpectedError >= ((or_tree *)o2)->Accumulated.AccExpectedError)
// EN REALITAT CALCULA >= EN LLOC DE < PERQUE ESTEM COMPARANT num regles i quan menys millor
    return true;
  else
    return false;
}



// This function is used as a < for calling max_element for selecting the best Solution
bool Compare_TestCost(tree *o1, tree *o2)  {
  if ( (((or_tree *)o1)->State == PROGRESS) &&  ((or_tree *)o2)->State == SUSPENDED)
    return false;
  else  if ( (((or_tree *)o2)->State == PROGRESS) &&  ((or_tree *)o1)->State == SUSPENDED)
    return true;

  if ( ((or_tree *)o1)->Accumulated.TestCost >= ((or_tree *)o2)->Accumulated.TestCost)
// EN REALITAT CALCULA >= EN LLOC DE < PERQUE ESTEM COMPARANT num regles i quan menys millor
    return true;
  else
    return false;
}

double Compare_TestCost_And_NumRules_Factor= 0.5;

void Initialise_Compare_TestCost_And_NumRules_Factor(double f) {
  Compare_TestCost_And_NumRules_Factor= f;
}


// This function is used as a < for calling max_element for selecting the best Solution
bool Compare_TestCost_And_NumRules(tree *o1, tree *o2) {
  if ( (((or_tree *)o1)->State == PROGRESS) &&  ((or_tree *)o2)->State == SUSPENDED)
    return false;
  else  if ( (((or_tree *)o2)->State == PROGRESS) &&  ((or_tree *)o1)->State == SUSPENDED)
    return true;

  if ( pow(((or_tree *)o1)->Accumulated.TestCost, Compare_TestCost_And_NumRules_Factor)
       * pow(((or_tree *)o1)->Accumulated.AccNumRules, 1 - Compare_TestCost_And_NumRules_Factor)
        >=
       pow(((or_tree *)o2)->Accumulated.TestCost, Compare_TestCost_And_NumRules_Factor)
       * pow(((or_tree *)o2)->Accumulated.AccNumRules, 1 - Compare_TestCost_And_NumRules_Factor)
     )
// EN REALITAT CALCULA >= EN LLOC DE < PERQUE ESTEM COMPARANT num regles i quan menys millor
    return true;
  else
    return false;
}

// This function is used as a < for calling max_element for selecting the best Solution
bool Compare_NumRules(tree *o1, tree *o2)  {
  if ( (((or_tree *)o1)->State == PROGRESS) &&  ((or_tree *)o2)->State == SUSPENDED)
    return false;
  else  if ( (((or_tree *)o2)->State == PROGRESS) &&  ((or_tree *)o1)->State == SUSPENDED)
    return true;

  if ( ((or_tree *)o1)->Accumulated.AccNumRules >= ((or_tree *)o2)->Accumulated.AccNumRules)
// EN REALITAT CALCULA >= EN LLOC DE < PERQUE ESTEM COMPARANT num regles i quan menys millor
    return true;
  else
    return false;
}

// This function is used as a < for calling max_element for selecting the best Solution
bool Compare_Cross_Coverage(tree *o1, tree *o2)  {
  if ( (((or_tree *)o1)->State == PROGRESS) &&  ((or_tree *)o2)->State == SUSPENDED)
    return false;
  else  if ( (((or_tree *)o2)->State == PROGRESS) &&  ((or_tree *)o1)->State == SUSPENDED)
    return true;

  if ( ((or_tree *)o1)->Coverage_Validation < ((or_tree *)o2)->Coverage_Validation)
    return true;
  else
        if ( ((or_tree *)o1)->Coverage_Validation == ((or_tree *)o2)->Coverage_Validation)
              if ( ((or_tree *)o1)->Accumulated.AccNumRules >= ((or_tree *)o2)->Accumulated.AccNumRules)
                 return true;
    return false;
}


// This function is used as a < for calling max_element for selecting the best Solution
bool Compare_MDL(tree *o1, tree *o2)  {
  if ( (((or_tree *)o1)->State == PROGRESS) &&  ((or_tree *)o2)->State == SUSPENDED)
    return false;
  else  if ( (((or_tree *)o2)->State == PROGRESS) &&  ((or_tree *)o1)->State == SUSPENDED)
    return true;

  if ( ((or_tree *)o1)->Accumulated.AccMDL >= ((or_tree *)o2)->Accumulated.AccMDL)
// EN REALITAT CALCULA >= EN LLOC DE < PERQUE ESTEM COMPARANT num regles i quan menys millor
    return true;
  else
    return false;
}	


bool (*Global_Comp_Function)(tree *, tree *);  // Global variable used in the following two functions

void Initialise_CompFunction_For_Compare_Meta_For_Different(bool (*Comp_Function2)(tree *, tree *)) {

  Global_Comp_Function= Comp_Function2;

}


// Used as < for calling max_element for selecting the n-best different solutions. It reuses all the others
bool Compare_Meta_For_Different_Using_Visits(tree *o1, tree *o2) {



  if ( (((or_tree *)o1)->State == PROGRESS) &&  ((or_tree *)o2)->State == SUSPENDED)
    return false;
  else  if ( (((or_tree *)o2)->State == PROGRESS) &&  ((or_tree *)o1)->State == SUSPENDED)
    return true;

  if ( ((or_tree *)o1)->Visits > ((or_tree *)o2)->Visits)
  // QUAN MS VISITES PITJOR, PER AIX > I NO <
    return true;
  else if ( ((or_tree *)o1)->Visits == ((or_tree *)o2)->Visits) {
    return Global_Comp_Function(o1, o2);
  }
  else
   return false;
}




// Function for comparison to be used to select the best in the list of or children.
bool Compare_Split_Optimality(tree *o1, tree *o2) {
  if (
      // NO FA FALTA (((or_tree *)o2)->State == SUSPENDED) &&
         (((or_tree *)o1)->Split_Optimality < ((or_tree *)o2)->Split_Optimality) )
    return true;
  else
    return false;
}

bool Compare_Rival_Ratio(tree *o1, tree *o2) {
  if (
      // NO FA FALTA (((or_tree *)o2)->State == SUSPENDED) &&
         (((or_tree *)o1)->Second_Optimality < ((or_tree *)o2)->Second_Optimality) )
    return true;
  else
    return false;
}


// Function for compare depth to select the best second or
bool Compare_TopMost(tree *o1, tree *o2) {

  float v1= float(((and_tree*)((or_tree *)o1)->Parent)->Depth);
  float v2= float(((and_tree*)((or_tree *)o2)->Parent)->Depth);
  if (
      // NO FA FALTA (((or_tree *)o2)->State == SUSPENDED) &&
         v1 > v2 )   // > perque a menor produnditat millor
    return true;
  else
    return false;

}


// Function for compare depth to select the best second or
bool Compare_BottomMost(tree *o1, tree *o2) {

  float v1= float(((and_tree*)((or_tree *)o1)->Parent)->Depth);
  float v2= float(((and_tree*)((or_tree *)o2)->Parent)->Depth);
  if (
      // NO FA FALTA (((or_tree *)o2)->State == SUSPENDED) &&
         v1 < v2 )
    return true;
  else
    return false;

}



float fact;

void Initialise_fact_For_Compare_Rival_Ratio_Depth(float f) {
  fact=f;
}


bool Compare_Rival_Ratio_Depth(tree *o1, tree *o2) {
  float v1= float(((and_tree*)((or_tree *)o1)->Parent)->Depth);
  float v2= float(((and_tree*)((or_tree *)o2)->Parent)->Depth);
/*
  if ((v1 == 0) || (v2 == 0))
    cout << "depth zeros FISTRO en trees.cpp" << endl;
*/

  float f1= pow(2, -v1 * fact);
  float f2= pow(2, -v2 * fact);



  if (
      // NO FA FALTA (((or_tree *)o2)->State == SUSPENDED) &&
         (((or_tree *)o1)->Second_Optimality * f1) <
         (((or_tree *)o2)->Second_Optimality * f2)
        )
    return true;
  else
    return false;
}


component_matrix *Component_Matrix;

void Initialise_fact_For_Compare_Rival_Ratio_Component(float f, component_matrix &C) {
  fact= f;
  Component_Matrix = &C;
}

bool Compare_Rival_Ratio_Component(tree *o1, tree *o2) {
  and_tree *a1Pa, *a2Pa;
  a1Pa= (and_tree*)(((or_tree *)o1)->Children.front());
  a2Pa= (and_tree*)(((or_tree *)o2)->Children.front());

  float v1= Component_Matrix->AccessRP(a1Pa->ActualArg, get_partition(a1Pa->ActualCond));
  float v2= Component_Matrix->AccessRP(a2Pa->ActualArg, get_partition(a2Pa->ActualCond));
/*
  if ((v1 == 0) || (v2 == 0))
    cout << "matrix zeros FISTRO en trees.cpp" << endl;
*/
  float f1= pow(2, -v1 * fact);
  float f2= pow(2, -v2 * fact);

  if (
      // NO FA FALTA (((or_tree *)o2)->State == SUSPENDED) &&
         (((or_tree *)o1)->Second_Optimality * f1) <
         (((or_tree *)o2)->Second_Optimality * f2)
        )
    return true;
  else
    return false;
}



  // Show all Rules of Multitree
  string  multitree::ShowAllRules(char *FunctionName, const dataset &DataSet, bool Short) const {
  // Short is for not using names of arguments. It gives a shorter output
    long NumRules= 0;
    string s= ShowNodeBest(Root, FunctionName, "", DataSet, Short, NumRules, true);
    static char Buf[20];
    strstream sst(Buf, 20);
    sst << NumRules << endl << '\0';
    s+= "Number of Rules of All The Multitree: " + string(Buf);
    return s;
  }

  // Show tree in the form of rules
  string multitree::ShowTreeBest(char *FunctionName, const dataset &DataSet, const dataset &Validation_Data,bool Short) const {
  // Short is for not using names of arguments. It gives a shorter output

    if (Options.Multitree.BestTreeCrit == OCCAM_BEST)  {
      cout << "Number of Rules Expected:" << Root->Accumulated.AccNumRules << endl;
    }
    else if (Options.Multitree.BestTreeCrit ==CROSS_COVERAGE)  {
     cout << "Number of examples Train:" << DataSet.Cardinality() << endl;
     cout << "Number of examples Cross Validation:" << Validation_Data.Cardinality() << endl;
     cout << "Number of examples of Validation_test covered:" << Root->Coverage_Validation<< endl;
     cout << "Number of Rules Expected:" << Root->Accumulated.AccNumRules << endl;
    }
    long NumRules= 0;

  //  Options.Multitree.BestTreeCrit= OCCAM_BEST;

    string s= ShowNodeBest(Root, FunctionName, "", DataSet, Short, NumRules, false);
    static char Buf[20];
    strstream sst(Buf, 20);
    sst << NumRules << endl << '\0';
    s+= "Number of Rules of This Solution: " + string(Buf);
    return s;
  }

  string multitree::ShowNodeBest(and_tree *Actual_AND_Node, char *FunctionName, string Desp,
                                                   const dataset &DataSet, bool Short, long &NumRules, bool All) const {
    string s;
    if (!Actual_AND_Node)
      cout << "ERROR DELS GROSSOS en multitree:ShowNodeBest";
    else {
      if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned) ) { // || (Actual_AND_Node->Children.size() == 0)) {    // NO TINDRIA QUE N'HI HAURE DISCCREPANCIES ACI, PERO N'HI HAN
        static char Buf[100];
        strstream sst(Buf, 100);
        int  cl= Actual_AND_Node->PredictedClass;
        if (Short)
          sst <<  cl;
        else
           sst << DataSet.NameClassValue(cl) ;

        cl=  Actual_AND_Node->Cardinality;
        sst << " , Card: " << cl;

        /* sst << " , ";

        int NumClasses= TrainSet.NumClasses();
        for(int i=0; i<NumClasses; i++) {  // Class will be the majority class
           indext iValue= Actual_AND_Node->ClassDistribution[i];
           sst << "C" << i << ": " << iValue << ",";
        }
        */
        sst << '\0';

        s= Desp + Actual_AND_Node->Rule.ToStringUnfolded(FunctionName, DataSet, false)
               + string(" [class: ") + string(Buf) + string("]") + string("\n") ;  // This line is because we dont use another node for the condition for the class (We use attribute 'Class').
        NumRules++;
      }
      else {
        if (All == true) {
          list<tree *>::iterator ItAnd, ItAndEnd;
          ItAndEnd= Actual_AND_Node->Children.end();
          for(ItAnd= Actual_AND_Node->Children.begin(); ItAnd != ItAndEnd; ItAnd++) {

            or_tree *Actual_OR_Node = (or_tree *)*ItAnd;
            if (Actual_OR_Node->State == SUSPENDED)
              continue;

            list<tree *>::iterator It, ItEnd;
            ItEnd= Actual_OR_Node->Children.end();
            for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
              s+= ShowNodeBest((and_tree *)*It, FunctionName, Desp + "", DataSet, Short, NumRules, true);
            }
          }
        }
        else {
          or_tree *Actual_OR_Node = Actual_AND_Node->Select_Best_OR_For_Show(Options);

          list<tree *>::iterator It, ItEnd;
          ItEnd= Actual_OR_Node->Children.end();
          for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
            s+= ShowNodeBest((and_tree *)*It, FunctionName, Desp + "", DataSet, Short, NumRules, false);
          }
       }
      }
    }
    return s;
  }



