/***************************************************************************
                          trees.h  -  description
                             -------------------

    functionality        : largest source file with the main learning functions
    relevant classes     : tree, and_tree, or_tree, multi_tree
    relevant functions   :

    begin                : 15 Oct 2001
    copyright            : (C) 2001 by Cesar Ferri, Jose Hernandez
    email                : {cferri,jorallo}@dsic.upv.es
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef TREES_H
#define TREES_H

#include "utils.h"
#include "estructures.h"
#include "rules.h"
#include "exemples.h"
#include "criteria.h"
#include "evaluate.h"
#include "options.h"
#include "roc.h"

#include <strstream>
#include <list>
#include <vector>
#include <cstdlib>
#include <algorithm>


/* Macro per a mostrar com es va creant el multitree */
// #define TREE_DEBUG

typedef enum { PROGRESS, NEEDED, COMPLETED, PRUNED, SUSPENDED } state;


#ifdef TREE_DEBUG
long GiveMeANumber() ;    // This function is to number the nodes
#endif

class tree {
public:                                   // A HAURIA DE SER PROTECTED PER NO HO AGARRA
  const tree *Parent;
  list<tree *> Children;
  #ifdef TREE_DEBUG
  long Num;
  #endif

public:
  tree(const tree *Daddy) {
    Parent= Daddy;
    #ifdef TREE_DEBUG
    Num= GiveMeANumber();
    # endif
  }

  virtual ~tree() {
    list<tree *>::iterator It, ItEnd;
    ItEnd= Children.end();
    for(It= Children.begin(); It != ItEnd; It++) {
      // (*It)->~tree();
     delete *It;
    }

    #ifdef TREE_DEBUG
      cout << "Node genric #" << Num << " destruit" << endl;
    # endif
  }

  void Add_Child(tree *Child) {
    #ifdef TREE_DEBUG
      if ((Child != NULL))
    # endif
       Children.push_back(Child);
    #ifdef TREE_DEBUG
      else cout << "Afegint nodes buits";
    # endif
  } 
};



class and_tree;


// void   Initialise_AND_for_Propagation(and_tree *, const options &);   // forward declaration

class or_tree;
class multitree;

class and_tree : public tree {
friend class multitree;
friend class or_tree;
public:
  rule Rule;                       // Whole rule up to this node
  int Depth;											 // Depth of this node.
  condition ActualCond;            // Condition added to this node
  int ActualArg;                   // To which argument this condition is applied
protected:
  vector<bool> OpenVariables;      // State of variables in this node
  // bit_vector Covered;           // Covered Examples. Efficient option in the beginning but expensive for deep trees
  vector<indext> Covered;           // Covered Examples. Better option for deep trees (although can be improved)
public:
  indext Cardinality;               // The number of examples that fall under this node
  //  state State;                     // State of this node
public:
  vector<indext> ClassDistribution; // Distribution of classes in this node (absolute frequencies)
  vector<float> SmoothDistribution; // Distribution of classes in this node (relative frequencies with smoothing if activated)
  vector<float> CostClass; // Expected cost assuming each class
  vector<float> CostProb; // Probabilities derived using costs
  vector<indext> Disc_Matrix;     // Discrepancy used for comparison with Invented Dataset
  or_tree *BestChild;              // Used to know which is the child with best optimality. This is needed for ObtainFirst function
public:
  vector<indext> TestDistribution; // Distribution of classes in this node for the TESTSET (absolute frequencies). Used for ROC analysis.
                                   // THIS VECTOR IS ONLY INITIALISED IF SOME ELEMENT IS >0

protected:
  bool Is_Leaf;                    // true if leaf
  bool Is_PostPruned;              // true if this node has been postpruned
  discrete PredictedClass;         // if it is a leaf, this value is the class chosen for prediction under this node.

  and_criteria_structure Crit;
  float ExpectedError;     // De moment a soles es gasta per al pruning Laplace o m-estimate
  and_accumulated_structure Accumulated;
  indext Coverage_Validation; // Number of examples of the Validation_Test that are covered this node
  vector <indext> Coverage_Validation_Vector; // Number of examples of the Validation_Test that fall in this node
public:

  // Constructor for root node only
  and_tree(int nArg, rule EmptyRule, const dataset & Train, const options & Options)
 : tree(NULL), Rule(nArg) {
    Depth= 0;

    for(int i=0; i < Train.NumArguments(); i++)
      OpenVariables.push_back(true);
    //    ActualCond = NO_CONDITION;
    // State= NEEDED;
    Is_Leaf= true;
    Is_PostPruned= false;

    // Com es l'arrel es pot calcular facil
    // This is like Apply_Condition in the rest o fnodes
        Cardinality = Train.Cardinality();
        ClassDistribution= Train.ClassDistribution();
        for(indext l=0; l < Cardinality; l++)  // All examples covered by the root
          Covered.push_back(l);

     ComputeSmoothing(Train, Options);

     ComputeCosts(Train, Options);

     // Sets Predicted Class
     SetsPredictedClass(Train, Options);

    ComputeExpectedError(Train, Options);

    Compute_Criteria_AND(*this, Train, Options);

    Initialise_AND_for_Propagation(this, Options);
     Coverage_Validation=0;
   #ifdef TREE_DEBUG
        cout << "Node #" << Num <<" and creat: " << Rule.ToStringUnfolded("f", Train, false)  << " Card: " << Cardinality << ". " ;
        int NumClasses= Train.NumClasses();
        for(int i=0; i<NumClasses; i++) {
           indext iValue= ClassDistribution[i];
           cout << "C" << i << ": " << iValue << ",";
        }
        cout << endl;
    #endif

    BestChild= NULL;
  }

  // Constructor for no root nodes
  and_tree(const or_tree *Daddy, condition Cond, int Arg, const dataset &Train, const options & Options )
  : tree((const tree *)Daddy), Rule(((and_tree *)(Parent->Parent))->Rule) {

    and_tree *GrandPa= (and_tree *)Parent->Parent;
    Depth= GrandPa->Depth + 1;
    OpenVariables = GrandPa->OpenVariables; 
    ActualCond = Cond;
    ActualArg= Arg;
    if (!Rule.Add_Condition(ActualArg, Cond))
      cout << "ERROR: Condici no vlida no detectada" << endl;
    // State= SUSPENDED;
    Is_Leaf= true;
    Is_PostPruned= false;

    Apply_Condition(Train);
    // This function applies the condition and computes:
    //    Cardinality; 
    //    ClassDistribution;
    //    Covered;

    #ifdef TREE_DEBUG
        cout << "Node #" << Num <<" and comena a crear-se. " << " Card: " << Cardinality << ". "  << endl;
    #endif

    // si la cardinality == 0 moltes d'estes funcions no fan res i aco es podria millorar.

    ComputeSmoothing(Train, Options);

    ComputeCosts(Train, Options);

    // Sets Predicted Class
    SetsPredictedClass(Train, Options) ;

    ComputeExpectedError(Train, Options);

    Compute_Criteria_AND(*this, Train, Options);

    Initialise_AND_for_Propagation(this, Options);
    Coverage_Validation=0;
    #ifdef TREE_DEBUG
        cout << "Node #" << Num <<" and creat: " << Rule.ToStringUnfolded("f", Train, false) << " Card: " << Cardinality << ". " ;
        int NumClasses= Train.NumClasses();
        for(int i=0; i<NumClasses; i++) {
           indext iValue= ClassDistribution[i];
           cout << "C" << i << ": " << iValue << ",";
        }
        cout << endl;
    #endif

    BestChild= NULL;
  }

  virtual ~and_tree() {
    // It has to be virtual. If not, the destructor of tree doesn't work properly

    #ifdef TREE_DEBUG
      cout << "AND_NODE #" << Num << " destruit" << endl;
    #endif
  }

  // Generates the OR children, adding them to Suspended_Nodes
  // It also generates its corresponding AND and stops at this level
  void Procreate(const  dataset & Train, list<or_tree *> & Suspended_Nodes, const options & Options ) ;

  // Selects the best tree according to the split selection criterion
  or_tree *Select_Best_OR_For_Split(const options &Options ) ;

  // Selects the best tree according to the best tree criterion
  or_tree *Select_Best_OR_For_Show(const options &Options);

  // Selects the best tree according to the n-best different tree criterion
  or_tree *Select_Best_OR_Different_Using_Visits(const options &Options);

  // Selects the best tree according to the n-best different tree criterion
  or_tree *Select_Best_OR_Different_Using_Components(const options &Options, const component_matrix &M);

private:
  // This function applies the new condition of an and_tree
  // and computes the cardinality and the class distribution
  void Apply_Condition(const dataset &Train);

   // si la cardinality == 0 moltes d'estes funcions no fan res i aco es podria millorar.

  // Sets Predicted Class
  void SetsPredictedClass(const dataset &Train, const options & Options) ;

  // Computes Smoothing of class frequencies
  void ComputeSmoothing(const dataset &Train, const options & Options) ;


  // Computes Costs and class frequencies derived from Costs
  void ComputeCosts(const dataset &Train, const options & Options) ;

  // Computes ExpectedError
  void ComputeExpectedError(const dataset &Train, const options & Options) ;

// Computes the different criteria according to the values: SSC, ...
friend void Compute_Criteria_AND(and_tree &AT, const dataset &Train, const options &Options) ;

// Computes the different criteria according to the values: SSC, ...
friend void Compute_Criteria_OR(or_tree &AT, const dataset &Train, const options &Options ) ;

friend void Initialise_AND_for_Propagation(and_tree *, const options &Options);

/*compute points for the ROC analysis*/
friend void compute_ROC_points(const vector<and_tree *> &solucio,vector<vector<double> > &points, bool Roc_With);
};




class or_tree : public tree {
friend class multitree;
friend class and_tree;
protected:
  condition *Partition;
  state State;  
  int Visits;                      // Number of visits. Only used when more than one solutions is output
  bool Marked;							// Used for choosing different solutions on the multitree.

  or_criteria_structure Crit;
  float Split_Optimality; // This is the variable every criterion must have updated.
  float BackupError;     // De moment a soles es gasta per al pruning Laplace o m-estimate or PEP-PREPRUNING  i per al SelectBest quan s'acumula

  or_accumulated_structure Accumulated;
  float Second_Optimality;     // Per a seleccionar el segent millor arbre OR que obrir (rival_ratio, ...).
   indext Coverage_Validation; // per a l'opcio de CROSS_COVERAGE magatzena informaci sobre la cobertura
   vector  < indext>  Coverage_Validation_Vector; // Number of examples of the Validation_Test that fall in this node
  vector<indext> Disc_Matrix;     // Discrepancy used for comparison with Invented Dataset

public:

  // Constructor. Creates the corresponding and_trees depending on the partition given on argument a
  or_tree(const and_tree *Daddy, const dataset &Train, int a, split_partition Part,
               const options & Options, argument Arg= argument(0) ): tree(Daddy) {

    Visits= 0;
    State= SUSPENDED;
     Coverage_Validation=0;
    if (Part == PARTITION_MANY_CONST_EQ) {

      // S'HAURIA DE MIRAR LA TAULA D'INDEXT PER NO OBRIR AQUELLS QUE ESTAN TANCATS O NO TENEN EXEMPLES

      discrete NumConst= Train.NumConstants(a);
      for (int c=0; c<NumConst; c++) {
        condition co= condition(CONST_EQ, c) ;
        and_tree *AT= new and_tree(this, co, a, Train, Options);
        AT->OpenVariables[a]= false;
        Add_Child(AT);
      }
    } 
    else if (Part == PARTITION_ONE_CONST_EQ) {

      // S'HAURIA DE MIRAR LA TAULA D'INDEXT PER NO OBRIR AQUELLS QUE ESTAN TANCATS O NO TENEN EXEMPLES

      and_tree *AT1= new and_tree(this, condition(CONST_EQ, Arg), a, Train, Options);
      AT1->OpenVariables[a]= false;
      Add_Child(AT1);

      and_tree *AT2= new and_tree(this, condition(CONST_NEQ, Arg), a, Train, Options);
      //  AT2->OpenVariables[a]= false; No PERQUE ES POT OBRIR MES
      Add_Child(AT2);    
    }
    else if (Part ==  PARTITION_ONE_VARIABLE_EQ) {
      and_tree *AT1= new and_tree(this, condition(VAR_EQ, Arg), a, Train, Options);
      AT1->OpenVariables[a]= false;
      Add_Child(AT1);

      and_tree *AT2= new and_tree(this, condition(VAR_NEQ, Arg), a, Train, Options);
      //  AT2->OpenVariables[a]= false; No PERQUE ES POT OBRIR MES
      Add_Child(AT2);
    }
    else if (Part == PARTITION_ONE_CONST_DISEQ) {
      and_tree *AT1= new and_tree(this, condition(NUM_LT, Arg), a, Train, Options);
      Add_Child(AT1);

      and_tree *AT2= new and_tree(this, condition(NUM_GE, Arg), a, Train, Options);
      Add_Child(AT2);
    }
    else if (Part == PARTITION_ONE_CONST_DISEQ_WITH_UNKNOWN) {
      and_tree *AT1= new and_tree(this, condition(NUM_LT_WU, Arg), a, Train, Options);
      Add_Child(AT1);

      and_tree *AT2= new and_tree(this, condition(NUM_GE_WU, Arg), a, Train, Options);
      Add_Child(AT2);

      and_tree *AT3= new and_tree(this, condition(NUM_MISSING, Arg), a, Train, Options);
      AT3->OpenVariables[a]= false;
      Add_Child(AT3);
    }
    else {
      // ERROR. 
      cout << "PARTITION " << Part << " NOT IMPLEMENTED" << endl;
    }


    BackupError= 0;
    float ParCard= ((and_tree *)Parent)->Cardinality;
    list<tree *>::iterator It, ItEnd;
    ItEnd= Children.end();
    for(It= Children.begin(); It != ItEnd; It++) {
      and_tree *Child= (and_tree *)*It;
      float  Fact= Child->Cardinality / ParCard;
      BackupError += Fact * Child->ExpectedError;
    }


    // Computes different criteria according to values: SSC, ...
    Compute_Criteria_OR(*this, Train, Options);


  };

  virtual ~or_tree() {
    #ifdef TREE_DEBUG
      cout << "OR_NODE #" << Num << " destruit" << endl;
    #endif
  }

  void Change_State(state S) {
    State= S;
  }

private:

// This function is used as a < for calling max_element function in Select_Best_OR
friend bool Compare_Split_Optimality(tree *, tree *);

// This function is used as a < for calling max_element for selecting the best solution
friend bool Compare_TestCost(tree *, tree *);

// This function is used as a < for calling max_element function in Select_Best_OR
friend bool Compare_ExpectedError(tree *o1, tree *o2) ;

// This function is used as a < for calling max_element for selecting the best Solution
friend bool Compare_NumRules(tree *o1, tree *o2);

// This function is used as a < for calling max_element for selecting the best Solution
friend bool Compare_TestCost_And_NumRules(tree *o1, tree *o2);

// Function for compare rival ratio to select the best second or
friend bool Compare_Rival_Ratio(tree *o1, tree *o2);

// Function for compare depth to select the best second or
friend bool Compare_TopMost(tree *o1, tree *o2);

// Function for compare depth to select the best second or
friend bool Compare_BottomMost(tree *o1, tree *o2);

// Function for compare rival ratio and using depth to select the best second or
friend bool Compare_Rival_Ratio_Depth(tree *o1, tree *o2);

// An auxiliary function for the previous one
friend void Initialise_fact_For_Compare_Rival_Ratio_Depth(float f);

// This function is used as a < for calling max_element for selecting the best Solution
friend bool Compare_Cross_Coverage(tree *o1, tree *o2);

// Used as < for calling max_element for selecting the n-best different solutions. It reuses all the others
friend bool Compare_Meta_For_Different_Using_Visits(tree *o1, tree *o2);

// An auxiliary function for the previous one
friend void Initialise_CompFunction_For_Compare_Meta_For_Different(bool (*Comp_Function2)(tree *, tree *));

// Function for compare rival ratio and using component matrix to select the best second or
friend bool Compare_Rival_Ratio_Component(tree *o1, tree *o2);

// An auxiliary function for the previous one
friend void Initialise_fact_For_Compare_Rival_Ratio_Component(float f, component_matrix &C);

// Computes different criteria according to values: SSC, ...
friend void Compute_Criteria_OR(or_tree &OT, const dataset &Train, const options &) ;

// Computes the different criteria according to the values: SSC, ...
friend void Compute_Criteria_AND(and_tree &AT, const dataset &Train, const options &) ;


// This function is used as a < for calling max_element for selecting the best Solution
friend bool Compare_MDL(tree *o1, tree *o2);

// This function is used as a < for calling max_element function in Select_Best_OR
friend bool Compare_AccExpected_Error(tree *o1, tree *o2);

};

class multitree {
private:
  // list<and_tree *> Needed_Nodes;    // Nodes to be explored for a solution
  list<or_tree *> Suspended_Nodes; // Nodes to be explored for a second solution
  and_tree *Root;
  long NumTrees; // Number of trees in the multitrees (ALTERNATIVE OR TREES EXPLORED)
   const dataset &TrainSet;
   options &Options;
public:
  // Constructs the multitree (just adds a root and_node)
  multitree(const dataset & Train, options &Opt)
                : TrainSet(Train), Options(Opt) {
    NumTrees= 0;
     int I= Train.NumArguments();
    Root= new and_tree(I, rule(I), Train, Options);
    Root->ActualArg= -1;  //  This is because the root does not use this value.
    // Needed_Nodes.push_back(Root);
  }

  ~multitree() {
   delete Root;
  }

  long SuspendedNodesExplored() {
    return NumTrees;
  }


  // Postprune the multitree
  // Marks the nodes as POSTPRUNE if it is enabled
  void PostPrune() {
    PostPrune1(Root);
  }

  // Postprune the multitree
  // Marks the nodes as POSTPRUNE if it is enabled
  void PostPrune1(and_tree *Actual_AND_Node) {

    double Card= Actual_AND_Node->Cardinality;

    // For the moment there is just one method implemented
    if (!(Actual_AND_Node->Is_Leaf) && (Card > 0) && (Options.PostPruningMethod == PEP_POSTPRUNING)) {
        double PEPFACTOR = 0.5;
        double E_Parent = Actual_AND_Node->ExpectedError * Card + PEPFACTOR;
        double E_Children = Actual_AND_Node->Accumulated.AccExpectedError * Card
                          + PEPFACTOR * Actual_AND_Node->Accumulated.AccNumRules;
        double SE = sqrt ( E_Children * (Card - E_Children) / Card );
        if (E_Parent <= (E_Children + SE))
          Actual_AND_Node->Is_PostPruned= true;
    }

    if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)      // post-pruning
          ) {
      return;
    }
    else {
      or_tree *Actual_OR_Node;

      list<tree *>::iterator It, ItEnd;
      ItEnd= Actual_AND_Node->Children.end();
      for(It= Actual_AND_Node->Children.begin(); It != ItEnd; It++) {
        Actual_OR_Node= (or_tree *)(*It);
        // A SOLES S'HAURIA DE FER SI ESTA EN ESTAT "PROGRESS"

        list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_OR_Node->Children.end();
        for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
          PostPrune1((and_tree *)*It);
        }
      }
    }
  }




  // Obtains a vector of vector of nodes, i.e. a vector of solutions
  // Initially the vector is empty
  void ObtainNBestDifferent(int n, vector<vector<and_tree *> > &Solutions) {
    component_matrix AccMatrix= Options.AccComponents;
    AccMatrix.Reset();
    component_matrix NewMatrix= AccMatrix;

    if (Options.SelectKBest.ComponentsGeneration == COMP_RANDOM_GENERATE_FROM_START) {
      AccMatrix.Randomise1(Options.SelectKBest.FactorForRandomComponents);
    }
//    Options.SelectKBest.ComponentsGeneration == COMP_ACCUMULATE;
//    Options.SelectKBest.ComponentsGeneration == COMP_RANDOM_GENERATE_FROM_SECOND;


    for(int c=0; c<n; c++) {

      if (Options.SelectKBest.Met == K_BEST_LESS_VISITED_PLUS) {
         // TO DO    if (Root.Unused_ORs == 0)
         // TO DO      break;
      }
      vector<and_tree *> Sol;
      Obtain1BestDifferent(Root, Sol, AccMatrix, NewMatrix);
      Solutions.push_back(Sol);
      /*
        cout << "Matriu nova: " << endl;
       NewMatrix.Show();
      */
//      Pause();

      if ((Options.SelectKBest.ComponentsGeneration == COMP_RANDOM_GENERATE_FROM_START) ||
         (Options.SelectKBest.ComponentsGeneration == COMP_RANDOM_GENERATE_FROM_SECOND)) {
        AccMatrix.Randomise1(Options.SelectKBest.FactorForRandomComponents);
      }
      else if (Options.SelectKBest.ComponentsGeneration == COMP_ACCUMULATE)
        AccMatrix.Add(NewMatrix);

      NewMatrix.Reset();
      /*
      cout << "Matriu acumulada: " << endl;
      AccMatrix.Show();
      */
    }
  }

  // Function to traverse the tree.
  // Initially the vector is empty
  void Obtain1BestDifferent(and_tree *Actual_AND_Node, vector<and_tree *> &Sol,
                            component_matrix & AccMatrix, component_matrix & NewMatrix) {

    NewMatrix.fill_matrix_cond(Actual_AND_Node->ActualArg, Actual_AND_Node->ActualCond,
                                 Actual_AND_Node->Depth);

    /* THIS IS ALREADY DONE IN FUNCTION PostPrune
    double Card= Actual_AND_Node->Cardinality;
    if (!(Actual_AND_Node->Is_Leaf) && (Card > 0) && (Options.PostPruningMethod == PEP_POSTPRUNING)) {
        double PEPFACTOR = 0.5;
        double E_Parent = Actual_AND_Node->ExpectedError * Card + PEPFACTOR;
        double E_Children = Actual_AND_Node->Accumulated.AccExpectedError * Card
                          + PEPFACTOR * Actual_AND_Node->Accumulated.AccNumRules;
        double SE = sqrt ( E_Children * (Card - E_Children) / Card );
        if (E_Parent <= (E_Children + SE))
          Actual_AND_Node->Is_PostPruned= true;
    }
    */

    if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)      // post-pruning
          ) {
      Sol.push_back(Actual_AND_Node);
      return;
    }
    else {

      or_tree *Actual_OR_Node;

      if ((Options.SelectKBest.Met == K_BEST_LESS_VISITED) || (Options.SelectKBest.Met == K_BEST_LESS_VISITED_PLUS))
        Actual_OR_Node= Actual_AND_Node->Select_Best_OR_Different_Using_Visits(Options);
      else if ((Options.SelectKBest.Met == K_BEST_DIFFERENT_COMPONENTS) || (Options.SelectKBest.Met == K_BEST_LESS_VISITED_THEN_DIFFERENT_COMPONENTS))
        Actual_OR_Node= Actual_AND_Node->Select_Best_OR_Different_Using_Components(Options, AccMatrix);
      else if (Options.SelectKBest.Met == K_BEST_RANDOM) {
         list<tree *>::iterator It, ItEnd;
         ItEnd= Actual_AND_Node->Children.end();
         long Size= Actual_AND_Node->Children.size();
         long Max= -10;
         for(It= Actual_AND_Node->Children.begin(); It != ItEnd; It++) {
           or_tree *Este= (or_tree *)(*It);
           if (Este->State == PROGRESS) {
             long Rand= (long)RandUniform(0, Size);
             if (Max < Rand) {
               Max= Rand;
               Actual_OR_Node= Este;
             }
           }
         }
         if (Max == -1) {  // ACI NO ENTRA MAI PERO A MI EM VA DONAR ERROR (coredump) ESTA OPCI (SECOND_RANDOM i K_BEST_RANDOM), PERO COM ERA ALEATORIA PASSAVA RARES VEGADES I NO VAIG DESCUBRIR PERQU.
           cout << "IEP. SuperError en ObtainqBestDifferent en trees.h " << endl;
           Pause();
         }

      }
      else if (Options.SelectKBest.Met == K_BEST_LESS_VISITED_PLUS) {
         // TO DO     Actual_OR_Node->Unused_ORs--; // Decrement and propagate
         // TO DO    Propagate_Unused_OR_Up(Actual_OR_Node);
         cout << "Option not implemented in function Obtain1BestDifferent" << endl;
      }
      else {
        cout << "Option unknown in function Obtain1BestDifferent" << endl;
      }

      Actual_OR_Node->Visits++;
      list<tree *>::iterator It, ItEnd;
      ItEnd= Actual_OR_Node->Children.end();
      for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
        Obtain1BestDifferent((and_tree *)*It, Sol, AccMatrix, NewMatrix);
      }
    }
  }


  void ObtainFirst(vector<and_tree *> &Sol) {
    ObtainFirst2(Root, Sol);
  }

  // Function to traverse the tree and obtain the first solution.
  // Initially the vector is empty
  // It also marks the nodes as POSTPRUNE if it is enabled
  void ObtainFirst2(and_tree *Actual_AND_Node, vector<and_tree *> &Sol) {

    /* THIS IS ALREADY DONE IN PostPrune

     double Card= Actual_AND_Node->Cardinality;
    if (!(Actual_AND_Node->Is_Leaf) && (Card > 0) && (Options.PostPruningMethod == PEP_POSTPRUNING)) {
        double PEPFACTOR = 0.5;
        double E_Parent = Actual_AND_Node->ExpectedError * Card + PEPFACTOR;
        double E_Children = Actual_AND_Node->Accumulated.AccExpectedError * Card
                          + PEPFACTOR * Actual_AND_Node->Accumulated.AccNumRules;
        double SE = sqrt ( E_Children * (Card - E_Children) / Card );
        if (E_Parent <= (E_Children + SE))
          Actual_AND_Node->Is_PostPruned= true;
    }
    */

    if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)      // post-pruning
          ) {
      Sol.push_back(Actual_AND_Node);
      return;
    }
    else {

      or_tree *Actual_OR_Node = NULL;


      Actual_OR_Node= Actual_AND_Node->BestChild;

      // Actual_OR_Node = Actual_AND_Node->Select_Best_OR_For_Split(Options);   // A NO FUNCIONA PERQUE ELS PROGRESS NO ELS MIRA
      	
      /*
  	 		list<tree *>::iterator It1, ItEnd1;
			  ItEnd1= Actual_AND_Node->Children.end();
			  for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++) {
  			   if (((or_tree*)(*It1))->State==PROGRESS) {
  			    Actual_OR_Node = (or_tree *)(*It1);   // THE LEFTMOST ONE
			      break;
				   } // del if
		    } // del for
		  */

      //      Actual_OR_Node= (or_tree *)*((Actual_AND_Node->Children).begin());
       /*
      if (Actual_OR_Node == NULL) {
        cout << "Iep ACI ESTEM!!" << endl;
      }
      */
      list<tree *>::iterator It, ItEnd;
      ItEnd= Actual_OR_Node->Children.end();
      for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
        ObtainFirst2((and_tree *)*It, Sol);
      }
    }
  }


  // Used by Obtain_Best_wrt_Combination for propagating Disc_Matrix Upwards
  void Accumulate_Disc_Matrix(and_tree *Actual_AND_Node) {
    int NClasses= TrainSet.NumClasses();
    int SizeMatrix= NClasses * NClasses;

    if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)      // post-pruning
          ) {

     // if (Actual_AND_Node->Disc_Matrix.size() > 0)
     //   cout << endl << Actual_AND_Node->Disc_Matrix.size();

      return;
    }
    else {

      or_tree *Actual_OR_Node;

      Actual_OR_Node= (or_tree *)*Actual_AND_Node->Children.begin();
      double BestK= -2000000000.0;


      bool FirstTime= true;
      list<tree *>::iterator It, ItEnd;
      ItEnd= Actual_AND_Node->Children.end();
      for(It= Actual_AND_Node->Children.begin(); It != ItEnd; It++) {
        or_tree *This_Or = (or_tree *)*It;
        vector<unsigned long> Acc_Disc_Matrix(SizeMatrix);
        {
          list<tree *>::iterator It, ItEnd;
          ItEnd= This_Or->Children.end();
           // Se SUMEN LES MATRIUS DELS AND
          for(It= This_Or->Children.begin(); It != ItEnd; It++) {
            Accumulate_Disc_Matrix((and_tree *)*It);
            if (((and_tree *)*It)->Disc_Matrix.size() != 0) {
              for (int i=0; i < SizeMatrix; i++) {
                Acc_Disc_Matrix[i]+= ((and_tree *)*It)->Disc_Matrix[i];
              }
            }
          }
        }

        This_Or->Disc_Matrix= Acc_Disc_Matrix;

        // SE SELECCIONA LA MILLOR SIMILARITY (e.g. KAPPA) DELS OR
        double K= Similarity(Acc_Disc_Matrix, TrainSet.NumClasses(), Options);
        if ((K > BestK) || (FirstTime)) {
          FirstTime= false;
          BestK = K;
          Actual_OR_Node= This_Or;
        }
      }
      Actual_AND_Node->Disc_Matrix=Actual_OR_Node->Disc_Matrix;

    }
  }

  // Obtains a set o nodes (a Solution) such that are most similar to the combined solution
  // It also may use an invented dataset
  void Obtain_Best_wrt_Combination(vector<and_tree *> &Sol, const options &Options) {
    Accumulate_Disc_Matrix(Root);
    Obtain_Best_wrt_Combination2(Root, Sol, Options);
  }

  void Obtain_Best_wrt_Combination2(and_tree *Actual_AND_Node, vector<and_tree *> &Sol, const options &Options) {

    if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)      // post-pruning
          ) {
      Sol.push_back(Actual_AND_Node);
      return;
    }
    else {

      or_tree *Actual_OR_Node;

      Actual_OR_Node= (or_tree *)*Actual_AND_Node->Children.begin();
      double BestK= -2000000000.0;

       // S'AGARRA EL QUE SUMA MS K DELS SEUS NODES PER BAIX.
      // S'HAURIEN DE SUMAR LES MATRIUS DELS AND I DESPRS CALCULAR LES similarities (KAPPAS..)

      if (Actual_AND_Node->Children.size() <= 1)
        ; // There is no need to select, there is just one child.
      else
      { list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_AND_Node->Children.end();
        bool FirstTime= true;
        for(It= Actual_AND_Node->Children.begin(); It != ItEnd; It++) {
          or_tree *This_Or = (or_tree *)*It;

          if (This_Or->State == SUSPENDED)
            continue;

          vector<unsigned long> V= This_Or->Disc_Matrix;

          double K= Similarity(V, TrainSet.NumClasses(), Options);

         // if (K != 0)
         //   cout << "Iep...";


          double TC= 1.0 / This_Or->Accumulated.TestCost;
          double NR= 1.0 / This_Or->Accumulated.AccNumRules;

          double f1= Options.Comb2Single.SimilarityImportanceFactor;  // Similarity
          double f2= Options.Comb2Single.TestCostImportanceFactor;    // TestCost
          double f3= Options.Comb2Single.OccamImportanceFactor;       // OCCAM
          double Mixture= pow(K, f1) * pow(TC, f2) * pow(NR, f3);


          // if (Mixture < -10000)
          //   cout << "iep";

          if ((Mixture > BestK) || (FirstTime)) {
            FirstTime= false;
            BestK = Mixture;
            Actual_OR_Node= This_Or;
          }

          // if (FirstTime)
          //   cout << "iep";

        }
      }

      if (Actual_OR_Node->State == SUSPENDED) {
        cout << endl << "SUSPENDED NODE SELECTED IN FUNCTION \"Obtain_Best_wrt_Combination2\"" << endl;
        exit(-1);
      }

      list<tree *>::iterator It, ItEnd;
      ItEnd= Actual_OR_Node->Children.end();
      for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
        Obtain_Best_wrt_Combination2((and_tree *)*It, Sol, Options);
      }
    }
  }


  // Unmark a multitree
  void UnMark(and_tree *Actual_AND_Node) {
    if (Actual_AND_Node->Is_Leaf) {
      // we can also put "|| Is_Postpruned" to make it more efficient
      return;
    }
    else {
      or_tree *Actual_OR_Node;
      list<tree *>::iterator It1, ItEnd1;
      ItEnd1= Actual_AND_Node->Children.end();
      for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++) {
        Actual_OR_Node= (or_tree *)*It1;
        Actual_OR_Node->Marked= false;
        list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_OR_Node->Children.end();
        for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
          UnMark((and_tree *)*It);
        }
      }
    }
  }

  // Mark a node and its antecedents
  void Mark(and_tree *Actual_AND_Node) {
    if (Actual_AND_Node->Parent == NULL)
      return;
    else {
      or_tree *Dad= (or_tree *)Actual_AND_Node->Parent;
      if (Dad->Marked == false) {  // if it is not marked (not that if it is marked there is no need to mark it again and its ancestors)
        Dad->Marked= true;
        Mark((and_tree *)(Dad->Parent));
      }
    }
  }

  // Unmarks and marks a multitree according to a solutions (set of leaf nodes)
  void Mark(const vector<and_tree *> Sol) {

    UnMark(Root);  // put all the Marked attributes to false

    long Size= Sol.size();
    for (int i=0; i<Size; i++) {
      Mark(Sol[i]);
    }
  }


  // Shows the solutions of a multitree given as argument (set of leaf nodes)
  void Show(const vector<and_tree *> Sol, char *FunctionName) {
    long Size= Sol.size();
    for (int i=0; i<Size; i++) {
      string s;
      s= Sol[i]->Rule.ToStringUnfolded(FunctionName, TrainSet);
      discrete Cl= Sol[i]->PredictedClass;
      cout << s << " [class: " << TrainSet.NameClassValue(Cl) << "]" << endl;
    }
  }

  // Obtains a vector of class results using the marks in the OR nodes
  void Result_as_Marked(const dataset &Test, const options &Options, vector<discrete> &Results, vector<unsigned long> &ArgumentUseVector) {
    unsigned long Len= Test.Cardinality();
    float Reliability;
    vector<float> ClassProbVector;

    for (unsigned long i=0; i<Len; i++) {
      int NArg= Test.NumArguments();
      vector<unsigned long> ArgumentUseVector2(NArg);
      discrete Cla= PredictMarked(Root, i, Test, Reliability, ClassProbVector, ArgumentUseVector2);

      for(int a=0; a < NArg; a++) {
        if (ArgumentUseVector2[a]>0)  // repeated use of the same argument are not taken into account
          ArgumentUseVector[a]++;
      }

      Results.push_back(Cla);
    }
  }

  // Predict the class using the marks in the OR nodes
  discrete PredictMarked(and_tree *Actual_AND_Node, indext i, const dataset & Test, float &Reliability, vector<float> &ClassProbVector, vector<unsigned long> &ArgumentUseVector) {
    if (!Actual_AND_Node)
      {
      	cout << "ERROR DELS GROSSOS en multitree:PredictMarked";
	      exit(0);
      }
    else {
      if ((Actual_AND_Node->Is_Leaf)
            || (Actual_AND_Node->Is_PostPruned) ) {

        if (Options.ReliabilityCalculation == FREQUENCY_RELIABILITY) {
          if (Actual_AND_Node->Cardinality)
            Reliability= Actual_AND_Node->ClassDistribution[Actual_AND_Node->PredictedClass] / double(Actual_AND_Node->Cardinality);
          else
            Reliability= 0;
        }
        else // if (Options.ReliabilityCalculation == LAPLACE_RELIABILITY)
          Reliability= (Actual_AND_Node->ClassDistribution[Actual_AND_Node->PredictedClass] + 1) / double((Actual_AND_Node->Cardinality + Test.NumClasses()));

        for (unsigned int i= 0; i < Test.NumClasses(); i++) {
          if (Actual_AND_Node->Cardinality) {
            if (Options.ClassProbabilityVectorCalculation == FREQUENCY_CLASS_PROBABILITY_VECTOR)
              ClassProbVector.push_back(Actual_AND_Node->ClassDistribution[i] / float(Actual_AND_Node->Cardinality));
            else // if (Options.ClassProbabilityVectorCalculation == LAPLACE_CLASS_PROBABILITY_VECTOR)
              ClassProbVector.push_back((Actual_AND_Node->ClassDistribution[i] + 1)
                                         / float(Actual_AND_Node->Cardinality + Test.NumClasses())  );
          }
          else
            ClassProbVector.push_back(1/float(Test.NumClasses()));
        }

        return Actual_AND_Node->PredictedClass;  // This is so easy because we dont use another node for the condition for the class (We use attribute 'Class').
      }
      else {

        or_tree *Actual_OR_Node= NULL;

        list<tree *>::iterator It1, ItEnd1;  // Selects the OR which is marked
        ItEnd1= Actual_AND_Node->Children.end();
        for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++) {
          or_tree *Actual_OR_Node2= (or_tree *)*It1;
          if (Actual_OR_Node2->Marked) {
           Actual_OR_Node= Actual_OR_Node2;
           break;
          }
        }
        if (Actual_OR_Node == NULL) {
          cout << endl << "ERROR! Function multitree::PredictMarked has arrived to an AND node whose nodes are not marked. Exiting" << endl;
          exit(-1);
        }

        list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_OR_Node->Children.end();
        for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
          int Arg= ((and_tree *)*It)->ActualArg;
          if (Test.FollowsCondition(Arg, ((and_tree *)*It)->ActualCond, i)) {  // Selects the branch following the condition.
            ArgumentUseVector[Arg]++;
            return PredictMarked((and_tree *)*It, i, Test, Reliability, ClassProbVector, ArgumentUseVector);
          }
        }
      }
    }
    return 0; // Never arrives here
  }


  // Fill all leaf nodes with their "TestDistribution"
  // Nodes with no examples under them are NOT INITIALISED
  void FillTestDistribution(const dataset &Test) {
    indext Len= Test.Cardinality();
    for (indext i=0; i<Len; i++) {
      #ifdef VERBOSE
        cout << "\r  Example " << i << " from " << Len;
        cout.flush();
      #endif
      IncrementTestDistribution(Root, i, Test);
    }
    cout << "\r " << Len << " examples done.                     "; // << endl;
  }

  void IncrementTestDistribution(and_tree *Actual_AND_Node, indext i, const dataset &Test) {
    if (!Actual_AND_Node)
      {
      	cout << "ERROR DELS GROSSOS en multitree:IncrementTestDistribution";
	      exit(0);
      }
    else {
      if ((Actual_AND_Node->Is_Leaf) || (Actual_AND_Node->Is_PostPruned)) {
        if (Actual_AND_Node->TestDistribution.size() == 0) {  // if it is not initialised
          for(int j=0; j< Test.NumClasses(); j++) {
            Actual_AND_Node->TestDistribution.push_back(0);
          }
        }

        discrete Classe= Test.ClassOf(i);
        Actual_AND_Node->TestDistribution[Classe]++;
      }
      else {
        or_tree *Actual_OR_Node;

        list<tree *>::iterator It1, ItEnd1;
        ItEnd1= Actual_AND_Node->Children.end();
        for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++) {
          Actual_OR_Node= (or_tree *)*It1;
          list<tree *>::iterator It, ItEnd;
          ItEnd= Actual_OR_Node->Children.end();
          for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
            if (Test.FollowsCondition(((and_tree *)*It)->ActualArg, ((and_tree *)*It)->ActualCond, i))  // Selects the branch following the condition.
              IncrementTestDistribution((and_tree *)*It, i, Test);
          }
        }
      }
    }
  }



  // Show all Rules of Multitree
  string ShowAllRules(char *FunctionName, const dataset &DataSet, bool Short) const;

  // Show tree in the form of rules
  string ShowTreeBest(char *FunctionName, const dataset &DataSet, const dataset &Validation_Data,bool Short) const;

  discrete PredictBest(indext i, const dataset & Test, float &Reliability, vector<float> &ClassProbVector) {
    return PredictBest2(Root, i, Test, Reliability, ClassProbVector);
  }

  discrete PredictBest2(and_tree *Actual_AND_Node, indext i, const dataset & Test, float &Reliability, vector<float> &ClassProbVector) {
    if (!Actual_AND_Node) {
	    cout << "ERROR DELS GROSSOS en multitree:PredictBest2";
	    exit(0);
    }
    else {
      if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)  // post-pruning
          ) {

        if (Options.ReliabilityCalculation == FREQUENCY_RELIABILITY) {
          if (Actual_AND_Node->Cardinality)
            Reliability= Actual_AND_Node->ClassDistribution[Actual_AND_Node->PredictedClass] / double(Actual_AND_Node->Cardinality);
          else
            Reliability= 0;
        }
        else // if (Options.ReliabilityCalculation == LAPLACE_RELIABILITY)
          Reliability= (Actual_AND_Node->ClassDistribution[Actual_AND_Node->PredictedClass] + 1) / double(Actual_AND_Node->Cardinality + Test.NumClasses());

        for (unsigned int i= 0; i < Test.NumClasses(); i++) {
          if (Actual_AND_Node->Cardinality) {
            if (Options.ClassProbabilityVectorCalculation == FREQUENCY_CLASS_PROBABILITY_VECTOR)
              ClassProbVector.push_back(Actual_AND_Node->ClassDistribution[i] / float(Actual_AND_Node->Cardinality));
            else // if (Options.ClassProbabilityVectorCalculation == LAPLACE_CLASS_PROBABILITY_VECTOR)
              ClassProbVector.push_back((Actual_AND_Node->ClassDistribution[i] + 1)
                                         / float(Actual_AND_Node->Cardinality + Test.NumClasses())  );
          }
          else
            ClassProbVector.push_back(1/float(Test.NumClasses()));
        }


        return Actual_AND_Node->PredictedClass;  // This is so easy because we dont use another node for the condition for the class (We use attribute 'Class').
      }
      else {
        or_tree *Actual_OR_Node = Actual_AND_Node->Select_Best_OR_For_Show(Options);
  
        list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_OR_Node->Children.end();
        for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
          if (Test.FollowsCondition(((and_tree *)*It)->ActualArg, ((and_tree *)*It)->ActualCond, i))  // Selects the branch following the condition.
            return PredictBest2((and_tree *)*It, i, Test, Reliability, ClassProbVector);
        }
      }
    }
    return 0; // Never arrives here
  }


  // USED WHEN Options.Comb.Comb==MAJORITY_CRISP;
   discrete PredictComb(indext i, const dataset & Test, float &Reliability, vector<float> &ClassProbVector) {
    return PredictComb2(Root, i, Test, Reliability, ClassProbVector);
  }


  void Positivise(vector<double> &results) { // positivise a vector. Ex (-1, -3, 1) turns into (2, 0, 4)
    double Min= 0;
    for(unsigned long i=0; i<results.size(); i++) {
      if (Min > results[i])
        Min= results[i];
    }
    for(unsigned long i=0; i<results.size(); i++) {
      results[i]+= -Min;
    }
  }


  void Positivise(vdistt &results) { // positivise a vector. Ex (-1, -3, 1) turns into (2, 0, 4)
    unsigned long Min= 0;
    for(unsigned long i=0; i<results.size(); i++) {
      if (Min > results[i])
        Min= results[i];
    }
    for(unsigned long i=0; i<results.size(); i++) {
      results[i]+= -Min;
    }
  }


  discrete PredictComb2(and_tree *Actual_AND_Node, indext i, const dataset & Test, float &Reliability, vector<float> &ClassProbVector) {
    if (!Actual_AND_Node) {
      cout << "ERROR DELS GROSSOS en multitree:PredictBest2";
      return 0;
    }
    else {
      if ( (Actual_AND_Node->Is_Leaf)
           || (Actual_AND_Node->Is_PostPruned && Options.Comb.PruneInCombination)      // post-pruning
         ) {

        if (Options.ReliabilityCalculation == FREQUENCY_RELIABILITY) {
          if (Actual_AND_Node->Cardinality)
            Reliability= Actual_AND_Node->ClassDistribution[Actual_AND_Node->PredictedClass] / double(Actual_AND_Node->Cardinality);
          else
            Reliability= 0;
        }
        else // if (Options.ReliabilityCalculation == LAPLACE_RELIABILITY)
          Reliability= (Actual_AND_Node->ClassDistribution[Actual_AND_Node->PredictedClass] + 1) / double(Actual_AND_Node->Cardinality + Test.NumClasses());

        for (unsigned int i= 0; i < Test.NumClasses(); i++) {
          if (Actual_AND_Node->Cardinality) {
            if (Options.ClassProbabilityVectorCalculation == FREQUENCY_CLASS_PROBABILITY_VECTOR)
              ClassProbVector.push_back(Actual_AND_Node->ClassDistribution[i] / float(Actual_AND_Node->Cardinality));
            else // if (Options.ClassProbabilityVectorCalculation == LAPLACE_CLASS_PROBABILITY_VECTOR)
              ClassProbVector.push_back((Actual_AND_Node->ClassDistribution[i] + 1)
                                         / float(Actual_AND_Node->Cardinality + Test.NumClasses())  );
          }
          else
            ClassProbVector.push_back(1/float(Test.NumClasses()));
        }

        return Actual_AND_Node->PredictedClass;  // This is so easy because we dont use another node for the condition for the class (We use attribute 'Class').
      }
      else {
      		vector<double> results(Test.NumClasses());// vector per magatzenar el resultats de les classes OR
      			//results(NumClasses());
 				  list<tree *>::iterator It1, ItEnd1;
  				ItEnd1= Actual_AND_Node->Children.end();
  				long Num_ORs= 0;
  				for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++) {
  				  if (((or_tree*)(*It1))->State==PROGRESS) {
			        Num_ORs++;
				    } // del if
				  } // del for
				
				  for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++) {
  				  if (((or_tree*)(*It1))->State==PROGRESS) {
			        		list<tree *>::iterator It, ItEnd;
        					ItEnd= (*It1)->Children.end();
        					for(It= (*It1)->Children.begin(); It != ItEnd; It++) {
          			   		if (Test.FollowsCondition(((and_tree *)*It)->ActualArg, ((and_tree *)*It)->ActualCond, i)) { // Selects the branch following the condition.
                        bool METODE_ANTIC= true;
                        if (METODE_ANTIC) {
                         	results[PredictComb2((and_tree *)*It, i, Test, Reliability, ClassProbVector)]++; //actualitzem el vector
                        }
                        else
                         	results[PredictComb2((and_tree *)*It, i, Test, Reliability, ClassProbVector)]+= 1 / double(Num_ORs); //actualitzem el vector
                      }
			 	       		}
				    } // del if
				  } // del for
	
		
    	    Positivise(results);   // this adds the lowest negative value to all the elements in order to obtain an equivalent vector with only positive values
                            // in general, this doesn't modify the vector
		  			
			   word j, mpos=0;
			   double max;
	       max= results[0] - 1;  // In order to begin with a value lower than at least one.
 				
			    for (j=0;j<results.size();j++) {
				    if (results[j]>max)
				    		{
				    			max=results[j];
				    			mpos=j;
				    		}
			   }
         if (max==0.0) {  // THIS IS TO AVOID TO ENTER HERE WHEN "DIFFERENCE" METHOD IS USED THAT CAN HAVE NEGATIVE VALUES
		    //  discrete r= Root->PredictedClass;
	         discrete r2= TrainSet.MajorityClass();
	     /*  if (r != r2)
	        cout << "iep" << endl;
	    */
	        mpos= r2; // if there are no assignments for that example,
	      }                                        //  we return the most popular class of the train dataset
	    		
			  return mpos;
	   }//del else
   }//del else
}

  // USED WHEN Options.Comb.Comb==MAJORITY_ABSOLUT_STOCHASTIC

 discrete PredictCombVect(indext i, const dataset & Test, float &Reliability, vector<float> &ClassProbVector) {
   vector<double> resultats(Test.NumClasses());
   word mpos=0;
   double max=0.0;
   float Acc= 0;

   // ASSIGNANT
/*
   if (i == 6)
     cout << "iepp..." << endl;
*/
   PredictCombVect2(Root, i, Test,resultats);
   // cout<< "Resultats"<<endl;

	 Positivise(resultats);   // this adds the lowest negative value to all the elements in order to obtain an equivalent vector with only positive values
                            // in general, this doesn't modify the vector
	
   max= resultats[0] - 1;  // In order to begin with a value lower than at least one.
	 for (unsigned int j=0; j < resultats.size(); j++) {
	  	//cout<<resultats[j]<<";";
	 	 Acc+= resultats[j];
		 if (resultats[j]>max)	{
		   max=resultats[j];
		   mpos=j;
		 }
	 }
	
   if (max==0.0 && Acc == 0) {  // THIS IS TO AVOID TO ENTER HERE WHEN "DIFFERENCE" METHOD IS USED THAT CAN HAVE NEGATIVE VALUES
	     //  discrete r= Root->PredictedClass;
     discrete r2= TrainSet.MajorityClass();
	     /*  if (r != r2)
	      cout << "iep" << endl;
	     */
     mpos= r2; // if there are no assignments for that example,
	              //  we return the most popular class of the train dataset
   }
	
	 // cout<<"resultat"<<mpos<<";"<<endl;
	
   if (Options.ReliabilityCalculation == FREQUENCY_RELIABILITY) {
  	 if (Acc==0.0) {
	     Reliability= 0;
	   }
	   else {
	     Reliability= max / Acc;
	   }
   }
   else // if (Options.ReliabilityCalculation == LAPLACE_RELIABILITY)
     Reliability= (max + 1) / (Acc + Test.NumClasses());
	

   for (unsigned int i= 0; i < Test.NumClasses(); i++) {
     if (Acc) {
       if (Options.ClassProbabilityVectorCalculation == FREQUENCY_CLASS_PROBABILITY_VECTOR)
         ClassProbVector.push_back(resultats[i] / float(Acc));
       else // if (Options.ClassProbabilityVectorCalculation == LAPLACE_CLASS_PROBABILITY_VECTOR)
         ClassProbVector.push_back((resultats[i] + 1) / float(Acc + Test.NumClasses()));
       }
     else
       ClassProbVector.push_back(1/float(Test.NumClasses()));
   }


   return mpos;
 }


  void PredictCombVect2(and_tree *Actual_AND_Node, indext i, const dataset & Test,
  			vector<double> &resultats) {
  	int NClasses= Test.NumClasses();
  	
  					
    if (!Actual_AND_Node)
      cout << "ERROR DELS GROSSOS en multitree:PredictBest2";
    else {
      if ((Actual_AND_Node->Is_Leaf)
           || (Actual_AND_Node->Is_PostPruned && Options.Comb.PruneInCombination)      // post-pruning
          // DESACTIVAR PRUNING EN COMBINATION
          ) {
        word j;
        double Max= 0, Total= 0;
        for (j=0; j < NClasses; j++) {
          double iValue;
          long Card= Actual_AND_Node->Cardinality;
          if (!Options.Comb.Smoothing)
            iValue= Actual_AND_Node->ClassDistribution[j];
          else {
              if (Card > 0) {
                iValue= Card * Actual_AND_Node->SmoothDistribution[j];	
              }
              else {
                if ((Options.Comb.FusionMethod == PROD) || (Options.Comb.FusionMethod == GEOMEAN) || (Options.Comb.FusionMethod == MINIMUM)) {
                  iValue= 1.0 / double(NClasses);  // (try to avoid a 0)
                }
                else
                  iValue= 0.0;
              }
          }

          if ((Options.Comb.CombinationVector == RELATIVE) && Card) {
            iValue= iValue / Card;
          }

          // if (Options.Comb.CombinationVectorMethod == ORIGINAL) // do nothing
          if (Options.Comb.CombinationVectorMethod == DIFFERENCE) {
              Total+= iValue;
          }
          else if ((Options.Comb.CombinationVectorMethod == GOOD_LOSER) ||
                   (Options.Comb.CombinationVectorMethod == BAD_LOSER) ||
                   (Options.Comb.CombinationVectorMethod == MAJORITY)) {
            if (iValue > Max) {
              Max= iValue;
            }
          }

          resultats[j]= iValue;
				//cout <<" a vore"<<resultats[j]<<Actual_AND_Node->ClassDistribution[j]<<endl;
		    }

		    if (Options.Comb.CombinationVectorMethod == DIFFERENCE) {
		      for (int j=0; j < NClasses; j++) {
            resultats[j] += -Total + resultats[j];
          }
		    }
        else if (Options.Comb.CombinationVectorMethod == MAJORITY) {
          for (int j=0; j < NClasses; j++) {
            if (resultats[j] == Max) {
              resultats[j] = 1.0;
              Max= -1;  // no other can have a 1.0
            }
            else {
              resultats[j] = 0.0;
            }
          }
        }
        else if (Options.Comb.CombinationVectorMethod == GOOD_LOSER) {
          double Acc= 0.0;
          for (int i=0; i < NClasses; i++) {
            Acc+= resultats[i];
          }
          for (int j=0; j < NClasses; j++) {
            if (resultats[j] == Max) {
              resultats[j] = Acc;
              Max= -1;  // no other can have a 1.0
            }
            else {
              resultats[j] = 0.0;
            }
          }
		    }
        else if (Options.Comb.CombinationVectorMethod == BAD_LOSER) {
          for (int j=0; j < NClasses; j++) {
            if (resultats[j] == Max) {
              // NOTHING TO DO. WE MAINTAIN THE VALUE;
              Max= -1;  // no other can have a 1.0
            }
            else {
              resultats[j] = 0.0;
            }
          }
		    }

        else if (Options.Comb.CombinationVectorMethod == SQUARED) {
          for (int j=0; j < NClasses; j++) {
            resultats[j] *= resultats[j];
          }
		    }

		    				
		  }
      else {
    //  	  vector<double> results(Test.NumClasses());// vector per magatzenar el resultats de les classes OR
      		//	res(NumClasses());
 					  list<tree *>::iterator It1, ItEnd1;
  				ItEnd1= Actual_AND_Node->Children.end();
  				long Num_ORs= 0;
  				for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++) {
  				  if (((or_tree*)(*It1))->State==PROGRESS) {
			        Num_ORs++;
				    } // del if
				  } // del for
				
 				
 					for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++) {
  				   if (((or_tree*)(*It1))->State==PROGRESS) {
			        		list<tree *>::iterator It, ItEnd;
        					ItEnd= (*It1)->Children.end();
        					bool FirstTime= true;
 		              vector<double> resultats2(NClasses);
         					for(It= (*It1)->Children.begin(); It != ItEnd; It++) {
          			   	if (Test.FollowsCondition(((and_tree *)*It)->ActualArg, ((and_tree *)*It)->ActualCond, i)) { // Selects the branch following the condition.
                      PredictCombVect2((and_tree *)*It, i, Test,resultats2); //actualitzem el vector
                      // break;  // with this it would be quicker. A set of conditions (and_tree) is exclusive.
                    }
                  }
                  for (int j=0;j<NClasses;j++) {

                      if (Options.Comb.FusionMethod == SUM) {
                        resultats[j]+= resultats2[j];
                      }
                      else if (Options.Comb.FusionMethod == ARITHMEAN) {
                        resultats[j]+= resultats2[j] / double(Num_ORs);
                      }
                      else if (Options.Comb.FusionMethod == PROD) {
                        if (FirstTime)
                          resultats[j]= 1.0;
                        resultats[j]*= resultats2[j];
                      }
                      else if (Options.Comb.FusionMethod == GEOMEAN) {
                        if (FirstTime)
                          resultats[j]= 1.0;
                        resultats[j]*= pow(resultats2[j], 1 / double(Num_ORs));
                      }
                      else if (Options.Comb.FusionMethod == MAXIMUM) {
                        if (resultats[j] < resultats2[j]) {
                          resultats[j] = resultats2[j];
                        }
                      }
                      else if (Options.Comb.FusionMethod == MINIMUM) {
                        if (FirstTime || (resultats[j] > resultats2[j])) {
                          resultats[j] = resultats2[j];
                        }
                      }
                      else if (Options.Comb.FusionMethod == MEDIAN) {
                        cout << endl << endl << "MEDIAN FUSION METHOD NOT IMPLEMENTED" << endl << endl;
                        exit(-1);
                      }

                  }
                  FirstTime= false;

				   }// del if
				 }// del for
			  }//del else
      }//del else
}



  void Increment_Difference_wrt_Combination(indext i, discrete Pred, const dataset & Invented) {

   Increment_Difference_wrt_Combination2(Root, i, Pred, Invented);

  }


  void Increment_Difference_wrt_Combination2(and_tree *Actual_AND_Node, indext i, discrete Pred, const dataset & Invented) {
    if (!Actual_AND_Node)
      cout << "ERROR DELS GROSSOS en multitree::Increment_Difference_wrt_Combination2";
    else {
      int NClasses= Invented.NumClasses();
      if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)) {      // post-pruning
		    if (Actual_AND_Node->Disc_Matrix.size()== 0) {
		      for (int i=0; i < NClasses * NClasses; i++) {
		        Actual_AND_Node->Disc_Matrix.push_back(0);
		      }
		    }
        Actual_AND_Node->Disc_Matrix[Pred*NClasses + Actual_AND_Node->PredictedClass]++;
	    }
	    else {
  			list<tree *>::iterator It1, ItEnd1;
  			ItEnd1= Actual_AND_Node->Children.end();
  			for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++)	{
  				if (((or_tree*)(*It1))->State==PROGRESS) {
			      list<tree *>::iterator It, ItEnd;
        		ItEnd= (*It1)->Children.end();
        		for(It= (*It1)->Children.begin(); It != ItEnd; It++) {
        			if (Invented.FollowsCondition(((and_tree *)*It)->ActualArg, ((and_tree *)*It)->ActualCond, i))  // Selects the branch following the condition.
                 Increment_Difference_wrt_Combination2((and_tree *)*It, i, Pred, Invented); //actualitzem el vector
						  }
				    } // del if
				  } // del for
			  } //del else
      } //del else
}



 discrete PredictCrossCoverage(indext i, const dataset & Test, float &Reliability, vector<float> &ClassProbVector) {
   vdistt resultats(Test.NumClasses());
   word mpos=0;
   word4 max=0;

   float Acc= 0;

   PredictCrossCoverage2(Root, i, Test,resultats);


	 Positivise(resultats);   // this adds the lowest negative value to all the elements in order to obtain an equivalent vector with only positive values
                            // in general, this doesn't modify the vector

   // cout<< "Resultats"<<endl;
	 for (unsigned int j=0;j<resultats.size();j++)	 {
	 //cout<<resultats[j]<<";";
	 	 Acc+= resultats[j];
     if (resultats[j]>max) {
			 max=resultats[j];
			 mpos=j;
     }
	 }

 	 if (max==0.0 && Acc == 00) {
	     //  discrete r= Root->PredictedClass;
     discrete r2= TrainSet.MajorityClass();
	     /*  if (r != r2)
	      cout << "iep" << endl;
	     */
     mpos= r2; // if there are no assignments for that example,
	              //  we return the most popular class of the train dataset
   }
	

	 // cout<<"resultat"<<mpos<<";"<<endl;
		
   if (Options.ReliabilityCalculation == FREQUENCY_RELIABILITY) {
  	 if (Acc==0.0) {
	     Reliability= 0;
	   }
	   else {
	     Reliability= max / Acc;
	   }
   }
   else // if (Options.ReliabilityCalculation == LAPLACE_RELIABILITY)
     Reliability= (max + 1) / (Acc + Test.NumClasses());
	
   for (unsigned int i= 0; i < Test.NumClasses(); i++) {
     if (Acc) {
       if (Options.ClassProbabilityVectorCalculation == FREQUENCY_CLASS_PROBABILITY_VECTOR)
         ClassProbVector.push_back(resultats[i] / float(Acc));
       else // if (Options.ClassProbabilityVectorCalculation == LAPLACE_CLASS_PROBABILITY_VECTOR)
         ClassProbVector.push_back((resultats[i] + 1) / float(Acc + Test.NumClasses()));
       }
     else
       ClassProbVector.push_back(1/float(Test.NumClasses()));
   }


	 // cout<<"resultat"<<mpos<<";"<<endl;
   return mpos;


  }


  void PredictCrossCoverage2(and_tree *Actual_AND_Node, indext i, const dataset & Test,
  			vdistt &resultats) {
    if (!Actual_AND_Node)
      cout << "ERROR DELS GROSSOS en multitree:PredictBest2";
    else {
      if ((Actual_AND_Node->Is_Leaf)
        || (Actual_AND_Node->Is_PostPruned)      // post-pruning
          ) {
		  word j;
          for (j=0;j<Actual_AND_Node->ClassDistribution.size();j++)
          {
				resultats[j]+=Actual_AND_Node->Coverage_Validation_Vector[j];
				//cout <<" a vore"<<resultats[j]<<Actual_AND_Node->ClassDistribution[j]<<endl;
		}
		 }
      else {
      			vector<discrete> results(Test.NumClasses());// vector per magatzenar el resultats de les classes OR
      		//	res(NumClasses());
 				list<tree *>::iterator It1, ItEnd1;
  				ItEnd1= Actual_AND_Node->Children.end();
  				for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++)
  				{
  				if (((or_tree*)(*It1))->State==PROGRESS)
  				{
			        		list<tree *>::iterator It, ItEnd;
        					ItEnd= (*It1)->Children.end();
        					for(It= (*It1)->Children.begin(); It != ItEnd; It++)
        					{
          			   		if (Test.FollowsCondition(((and_tree *)*It)->ActualArg, ((and_tree *)*It)->ActualCond, i))  // Selects the branch following the condition.
                                        PredictCrossCoverage2((and_tree *)*It, i, Test,resultats); //actualitzem el vector
						}
				   }// del if
				 }// del for
			    }//del else
     		 }//del else
}



/* Per a l'opci del CROSS_COVERAGE*/
 void Fill_Coverage_Validation(indext i, const dataset & Validation_Test) {
    return Fill_Coverage_Validation2(Root, i, Validation_Test);
  }

/* Per a l'opci del CROSS_COVERAGE*/
void  Fill_Coverage_Validation2(and_tree *Actual_AND_Node, indext n, const dataset & Validation_Test) {
    if (!Actual_AND_Node) {
      cout << "ERROR DELS GROSSOS en multitree:PredictBest2";
    }
    else {
      if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)      // post-pruning
          ) {
      				   unsigned int i=0;
						if(Actual_AND_Node->Coverage_Validation_Vector.size()==0) //init the vector if it is needed
						          for(i=0;i<Actual_AND_Node->ClassDistribution.size();i++)
										Actual_AND_Node->Coverage_Validation_Vector.push_back(0);
						Actual_AND_Node->Coverage_Validation_Vector[Validation_Test.ClassOf(n)]++; //fill the vector
						if (Actual_AND_Node->PredictedClass==Validation_Test.ClassOf(n))
								Actual_AND_Node->Coverage_Validation++;
						}
      else {
      			//results(NumClasses());
 				list<tree *>::iterator It1, ItEnd1;
  				ItEnd1= Actual_AND_Node->Children.end();
  				for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++)
  				{
  				if (((or_tree*)(*It1))->State==PROGRESS)
  				{
			        		list<tree *>::iterator It, ItEnd;
        					ItEnd= (*It1)->Children.end();
        					for(It= (*It1)->Children.begin(); It != ItEnd; It++)
        					{
          			   		if (Validation_Test.FollowsCondition(((and_tree *)*It)->ActualArg, ((and_tree *)*It)->ActualCond, n))  // Selects the branch following the condition.
                                      	Fill_Coverage_Validation2((and_tree *)*It, n, Validation_Test); //actualitzem el vector
						}
				   }// del if
				 }// del for
			    }//del else
     		 }//del else
}

/* Per a l'opci del CROSS_COVERAGE*/
indext  Propagate_Coverage_Validation(and_tree *Actual_AND_Node, const dataset & Validation_Test,
vector<indext> &Cross_Vector)
{
 unsigned int i=0;
 if (!Actual_AND_Node) {
      cout << "ERROR DELS GROSSOS en multitree:PredictBest2";
      return 0;
    }
    else {
     if ((Actual_AND_Node->Is_Leaf)
          || (Actual_AND_Node->Is_PostPruned)      // post-pruning
          ) {
      			if(Actual_AND_Node->Coverage_Validation_Vector.size()==0) //init the vector if it is needed
			          for(i=0;i<Actual_AND_Node->ClassDistribution.size();i++)
								Actual_AND_Node->Coverage_Validation_Vector.push_back(0);
      			Cross_Vector=Actual_AND_Node->Coverage_Validation_Vector;
						return Actual_AND_Node->Coverage_Validation;
						
				}
      else {
      			//results(NumClasses());
      			 indext maxt=0;
      			 long numrul=30000;
      			
      			if(Actual_AND_Node->Coverage_Validation_Vector.size()==0) //init the vector if it is needed
			          for(i=0;i<Actual_AND_Node->ClassDistribution.size();i++)
								Actual_AND_Node->Coverage_Validation_Vector.push_back(0);
 				list<tree *>::iterator It1, ItEnd1;
  				ItEnd1= Actual_AND_Node->Children.end();
  				for(It1= Actual_AND_Node->Children.begin(); It1 != ItEnd1; It1++)
  				{
  				  indext act=0;
  				
  				if (((or_tree*)(*It1))->State==PROGRESS)
  				{
			        		list<tree *>::iterator It, ItEnd;
        					ItEnd= (*It1)->Children.end();
        					for(It= (*It1)->Children.begin(); It != ItEnd; It++)
        					{
        					            vector <indext> vtemp;
                                      act+=Propagate_Coverage_Validation((and_tree *)*It, Validation_Test,vtemp);
                                       //actualitzem el vector
                                        for(i=0;i<Actual_AND_Node->ClassDistribution.size();i++)
                                      Actual_AND_Node->Coverage_Validation_Vector[i]+=vtemp[i];
						}
						
						((or_tree*)(*It1))->Coverage_Validation=act;
						((or_tree*)(*It1))->Coverage_Validation_Vector=Actual_AND_Node->Coverage_Validation_Vector;
						if ((act>maxt)||((act==maxt)&&(numrul>((or_tree*)(*It1))->Accumulated.AccNumRules)))
						                           {
						                            maxt=act;
						                            numrul=((or_tree*)(*It1))->Accumulated.AccNumRules;
						                             Cross_Vector=((or_tree*)(*It1))->Coverage_Validation_Vector;
						                            }
						
				   }// del if
				 }// del for
				 Actual_AND_Node->Coverage_Validation=maxt;
				Actual_AND_Node->Coverage_Validation_Vector=Cross_Vector;
				   return maxt;
			    }//del else
			
     		 }//del else
     }

  void Evaluate(const dataset & Test, evaluation_statistics & Eval, options &Options) {
    unsigned int NClasses= Test.NumClasses();
    indext Card= Test.Cardinality();

    vector<rank> RankVector(Card);  // vector of pairs index and Reliability to rank examples

    if (!(Eval.TestCardinality= Card)) return; // Eval; // No cardinality. What are we doing here?
    indext k;
    for(k=0; k < Card; k++) {
      discrete Pred;
      float Reliability;

      if (Options.Comb.Comb==NO_COMBINATION)
        Pred = PredictBest(k, Test, Reliability, RankVector[k].ClassProbVector);
      else if (Options.Comb.Comb==MAJORITY_CRISP)
        Pred = PredictComb(k, Test, Reliability, RankVector[k].ClassProbVector);
      else if (Options.Comb.Comb==MAJORITY_ABSOLUT_STOCHASTIC)
        Pred = PredictCombVect(k, Test, Reliability, RankVector[k].ClassProbVector);
	    else if (Options.Comb.Comb==CROSS_COVERAGE_COMBINATION)
	      Pred = PredictCrossCoverage(k, Test, Reliability, RankVector[k].ClassProbVector);
	    else {
	      cout << "Combination Method Unknown in function Evaluate" << endl;
	    }
      //       else     if (Options.Comb==MAJORITY_RELATIVE_STOCHASTIC,)
      //       else         if (Options.Comb==MAJORITY_COST_STOCHASTIC) // TODO

      discrete Real= Test.ClassOf(k);
      bool Correct;
      if (Pred==Real) {
        Eval.AbsAccuracy++;
        Eval.AbsTruePositives[Real]++;
        Correct= true;
      }
      else {
        Eval.AbsFalsePositives[Real]++;
        Correct= false;
      }
      Eval.ConfusionMatrix.AccessRP(Real,Pred)++;


      // Stores ranks for AUC Calculation using Reliability
      // TWO-CLASS ONLY (IT CAN BE DERIVED FROM RELIABILITY
     /*   double Prob0, Prob1;
      if (Pred == 0) {
        Prob0= Reliability;
        Prob1= 1 - Reliability;
      }
      else {
        Prob0= 1 - Reliability;
        Prob1= Reliability;
      }
      RankVector[k].ClassProbVector.push_back(Prob0);
      RankVector[k].ClassProbVector.push_back(Prob1);
     */

      RankVector[k].Real= Real;
    }

    // AUC Calculation using the ClassProbVector
    double Area= 0.0;
    for (unsigned int i= 0; i < NClasses; i++) {
      for (unsigned int j=i+1; j < NClasses; j++) {
        Area+= Compute_AUC_by_Ranks(RankVector, Card, i, j);
      }
    }

    Area/= (NClasses * (NClasses - 1) / 2.0);
    // Area is now AUC
    Eval.AUC= Area;


    Eval.PredCardinality= k;  // TOTS COBERTS
    for(unsigned int c=0; c < NClasses; c++) {
       Eval.TruePositives[c] = Eval.AbsTruePositives[c] / float(Eval.TestCardinality);
       Eval.FalsePositives[c] = Eval.AbsFalsePositives[c] / float(Eval.TestCardinality);
    }
    Eval.Accuracy = Eval.AbsAccuracy / float(Eval.TestCardinality);
    if (Options.Weights.Method!= NO_COSTS) {
    	Eval.ResultMatrix = Eval.ConfusionMatrix * Eval.CostMatrix;
   		Eval.TotalCost= Eval.ResultMatrix.Sum();
 		}
  }



  void Assign_Difference_wrt_Combination(const dataset & Train, const dataset & Invented, options &Options) {
    // HAURIEM D'UTILITZAR LES SEGENTS OPCIONS:
    // Options.CombForSelection.Method ... (MAJORITY_CRISP, ...)
    // Options.CombForSelection.PercentRandom (0: all train, 1: all random)

    double Percent= 1;

    unsigned long LimTrain= static_cast<unsigned long>(Train.Cardinality() * (1 - Percent));
    unsigned long LimInvented= static_cast<unsigned long>(Invented.Cardinality() * (Percent));

    for(unsigned long k=0; k < LimTrain; k++) {
      vector<float> ClassProbVector; // Not used. Just to pass as an argument
      float Reliability; // Not used either.
      discrete Pred;
      if (Options.Comb.Comb==NO_COMBINATION)
        Pred = PredictBest(k, Train, Reliability, ClassProbVector);
      else if (Options.Comb.Comb==MAJORITY_CRISP)
        Pred = PredictComb(k, Train, Reliability, ClassProbVector);
      else if (Options.Comb.Comb==MAJORITY_ABSOLUT_STOCHASTIC)
        Pred = PredictCombVect(k, Train, Reliability, ClassProbVector);
	    else if (Options.Comb.Comb==CROSS_COVERAGE_COMBINATION)
	      Pred = PredictCrossCoverage(k, Train, Reliability, ClassProbVector);
	    else {
	      cout << "Combination Method Unknown in function Assign_Difference_wrt_Combination" << endl;
	    }
      //       else     if (Options.Comb==MAJORITY_RELATIVE_STOCHASTIC,)
      //       else         if (Options.Comb==MAJORITY_COST_STOCHASTIC) // TODO

      // Invented
 //     if (Options.Multitree.BestTreeCrit == SIMILAR_TO_COMBINATION) {
        Increment_Difference_wrt_Combination(k, Pred, Train);
 //     }
    }

    for(unsigned long k=0; k < LimInvented; k++) {
      vector<float> ClassProbVector; // Not used. Just to pass as an argument
      float Reliability; // Not Used either
      discrete Pred;
      if (Options.Comb.Comb==NO_COMBINATION)
        Pred = PredictBest(k, Invented, Reliability, ClassProbVector);
      else if (Options.Comb.Comb==MAJORITY_CRISP)
        Pred = PredictComb(k, Invented, Reliability, ClassProbVector);
      else if (Options.Comb.Comb==MAJORITY_ABSOLUT_STOCHASTIC)
        Pred = PredictCombVect(k, Invented, Reliability, ClassProbVector);
	    else if (Options.Comb.Comb==CROSS_COVERAGE_COMBINATION)
	      Pred = PredictCrossCoverage(k, Invented, Reliability, ClassProbVector);
	    else {
	      cout << "Combination Method Unknown in function Assign_Difference_wrt_Combination" << endl;
	    }
      //       else     if (Options.Comb==MAJORITY_RELATIVE_STOCHASTIC,)
      //       else         if (Options.Comb==MAJORITY_COST_STOCHASTIC) // TODO

      // Invented
 //     if (Options.Multitree.BestTreeCrit == SIMILAR_TO_COMBINATION) {
        Increment_Difference_wrt_Combination(k, Pred, Invented);
 //     }
    }



  }


  void ShowPredictions(const dataset & Test /* FALTEN ELS MODES, DE MOMENT ES EL PREDICTBEST */) {
   for(indext k=0; k < Test.Cardinality(); k++) {
     vector<float> ClassProbVector; // Not shown actually
     float Reliability;
     discrete Pred= PredictBest(k, Test, Reliability, ClassProbVector);
      int Pred2= Pred;
      cout << "Prediction for the example #" << k <<": " << Pred2 << " with reliability: " << Reliability << endl;
    }
  }

// compute possible different solutions in the multitree
// for OR nodes it is added, and for AND nodes multipliyed
double Compute_Solutions(){
    return Compute_Solutions_AND(Root);
			}

double  Compute_Solutions_AND( and_tree *Actual_AND_Node){
			 double acc=0;
			list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_AND_Node->Children.end();
        for(It= Actual_AND_Node->Children.begin(); It != ItEnd; It++) {
  				  if (((or_tree*)(*It))->State==PROGRESS) {
			          acc+=Compute_Solutions_OR((or_tree*)(*It));
				    } // del if
         }
     return acc;
}

double   Compute_Solutions_OR( or_tree *Actual_OR_Node){
			 double  acc=1;
			list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_OR_Node->Children.end();
        for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
				if ((!((and_tree*)(*It))->Is_Leaf)
          && (!((and_tree*)(*It))->Is_PostPruned))      // post-pruning
				      acc*=Compute_Solutions_AND((and_tree*)(*It));
         }
     return acc;
}


  // Learns the multitree
  void Learn(const dataset &Train,const dataset &Validation_Test) {
    Activate_AND(Root, Train);
    NumTrees++;

    if ((Options.Multitree.Construction == ONE_ONLY) || (Options.Multitree.Construction == ONE_AND_FORGET))
       return;
    else if (Options.Multitree.Construction == MANY_AND_FORGET)  {
    }
    else if (Options.Multitree.Construction == MANY_AND_MAINTAIN)  {
       Options.SecondTreeDepthAccStat= 0; // Statistics

       long MaxArbres = Options.Multitree.MaxNumTrees;

       while (NumTrees < MaxArbres) {

          long Remaining= Suspended_Nodes.size() ;

          #ifdef VERBOSE
            cout << '\r' << "Opened Node #" << NumTrees << " learned of " << MaxArbres;
            cout.flush();
          #endif


          or_tree *OT= NULL;
          if (Remaining == 0) {
            cout << endl << "Suspended nodes exhausted. Maybe suspended nodes forgetting is too strict!" << endl; // << " Only " << NumTrees << " learned of " << MaxArbres << endl;
            break;
          }


          if (Options.Components_Reckoning) {
            Options.AccComponents.Add(Options.Components);
            Options.Components.Reset();
            Options.RandomComponents.Randomise1(Options.ComponentsRandomFactor);
/*
            Options.Components.Show();
            Options.AccComponents.Show();
            Options.RandomComponents.Show();
            Pause();
*/
          }

          if (Options.Multitree.SecondTreeCrit.Met == SPLIT_OPTIMALITY) {
            OT= (or_tree *) *max_element(Suspended_Nodes.begin(), Suspended_Nodes.end(), Compare_Split_Optimality);
          }
          else if (Options.Multitree.SecondTreeCrit.Met == OPTIMALITY_RIVAL_RATIO) {
            OT= (or_tree *) *max_element(Suspended_Nodes.begin(), Suspended_Nodes.end(), Compare_Rival_Ratio);
          }
          else if (Options.Multitree.SecondTreeCrit.Met == SECOND_TOPMOST) {
            OT= (or_tree *) *max_element(Suspended_Nodes.begin(), Suspended_Nodes.end(), Compare_TopMost);
          }
          else if (Options.Multitree.SecondTreeCrit.Met == SECOND_BOTTOMMOST) {
            OT= (or_tree *) *max_element(Suspended_Nodes.begin(), Suspended_Nodes.end(), Compare_BottomMost);
          }
          else if (Options.Multitree.SecondTreeCrit.Met == SECOND_FRONT) {
            OT= (or_tree *)Suspended_Nodes.front();
          }
          else if (Options.Multitree.SecondTreeCrit.Met == SECOND_BACK) {
            OT= (or_tree *)Suspended_Nodes.back();
          }

           else if (Options.Multitree.SecondTreeCrit.Met == OPTIMALITY_RIVAL_RATIO_DEPTH) {
           Initialise_fact_For_Compare_Rival_Ratio_Depth(Options.Multitree.SecondTreeCrit.Factor);
            OT= (or_tree *) *max_element(Suspended_Nodes.begin(), Suspended_Nodes.end(), Compare_Rival_Ratio_Depth);
          }

          else if (Options.Multitree.SecondTreeCrit.Met == OPTIMALITY_RIVAL_RATIO_COMPONENT) {
            Initialise_fact_For_Compare_Rival_Ratio_Component(Options.Multitree.SecondTreeCrit.Factor, Options.AccComponents);
            OT= (or_tree *) *max_element(Suspended_Nodes.begin(), Suspended_Nodes.end(), Compare_Rival_Ratio_Component);
          }

          else if (Options.Multitree.SecondTreeCrit.Met == OPTIMALITY_RIVAL_RATIO_COMPONENT_RANDOM) {
            Initialise_fact_For_Compare_Rival_Ratio_Component(Options.Multitree.SecondTreeCrit.Factor, Options.RandomComponents);
            OT= (or_tree *) *max_element(Suspended_Nodes.begin(), Suspended_Nodes.end(), Compare_Rival_Ratio_Component);
          }

          else if (Options.Multitree.SecondTreeCrit.Met == SECOND_RANDOM) {
            long Len= Suspended_Nodes.size();

            long Pos= (long)RandUniform(0, Len);
            if (Pos >= Len) {
              cout << "SUPER ERROR en funci LEARN" << endl << endl;
              Pos = Len - 1;
            }

            list<or_tree *>::iterator It, ItEnd;
            ItEnd= Suspended_Nodes.end();
            long i=0;
            for(It= Suspended_Nodes.begin(); It != ItEnd; It++) {
              if (i==Pos) {
                OT= (*It);
                break;
              }
              i++;
            }
           /*
            if (OT == NULL) {
              cout << "IEP!! SUPER-ERROR IN FUNCTION LEARN" << endl;
              exit(-1);
            }
           */
          }
          else if (Options.Multitree.SecondTreeCrit.Met == SECOND_RANDOM_DEPTH) {
            OT= NULL;
             long Len= Suspended_Nodes.size();
             list<or_tree *>::iterator It, ItEnd;
             ItEnd= Suspended_Nodes.end();
             float MaxOpt= -1;

              for(It= Suspended_Nodes.begin(); It != ItEnd; It++) {
                long Num= (long)RandUniform(0, Len);
                int d= ((and_tree *)((*It)->Parent))->Depth;
                float f= pow(2, -d * Options.Multitree.SecondTreeCrit.Factor);

                float Opt= Num * f;

                if (MaxOpt < Opt) {
                  MaxOpt = Opt;
                  OT = (*It);
                }
              }
           /*
            if (OT == NULL) {
              cout << "IEP!! SUPER-ERROR IN FUNCTION LEARN" << endl;
              exit(-1);
            }
           */
          }

          else  if (Options.Multitree.SecondTreeCrit.Met == ROC_BISECTRIX) {
             cout << "SECOND TREE CRITERION NOT IMPLEMENTED" << endl;
          }
          else {
             cout << "SECOND TREE CRITERION NOT IMPLEMENTED" << endl;
          }

          if (OT == NULL) {
            cout << "IEP!! SUPER-ERROR IN FUNCTION LEARN: or_tree is NULL" << endl;
            exit(-1);
          }
          /*
          if (OT->Parent == NULL) {
            cout << "IEP!! SUPER-ERROR IN FUNCTION LEARN" << endl;
            exit(-1);
          }
          */

          Options.SecondTreeDepthAccStat+= ((and_tree *)(OT->Parent))->Depth;
          Activate_OR(OT, Train);     // Removes the tree from suspended and activates it

          PropagateUpwards(OT);

          NumTrees++;
       }

       cout << '\r' << "Last Opened Node #" << NumTrees << " learned of " << MaxArbres;
       cout.flush();
  	
  		if ((Options.Comb.Comb!=MAJORITY_ABSOLUT_STOCHASTIC)&&((Options.Multitree.BestTreeCrit == CROSS_COVERAGE) ||
  		(Options.Comb.Comb==CROSS_COVERAGE_COMBINATION)))
  		{
  		         indext i1;
  		        for (i1=0;i1<Validation_Test.Cardinality();i1++)
	                 Fill_Coverage_Validation(i1,Validation_Test);
	               vector<indext> vtemp;
                  Propagate_Coverage_Validation(Root,Validation_Test,vtemp);
   }

    }
    else if (Options.Multitree.Construction == THRUSTING)  {
    }
    else
       cout << "Multitree Construction Mode Not Implemented" << endl;

  }

private:

  // Activates an And_node, which creates its or children (and these create its and children)
  // Selects one of the or children, according to the split criterion and activates all of its children
  // Actually, it makes a depth-first (left) traversal of the multitree.
  void Activate_AND(and_tree *Actual_AND_Node, const dataset &Train) {
    if (!Actual_AND_Node)
      cout << "ERROR DELS GROSSOS en multitree:activate";
//    else if (Actual_AND_Node->Is_Leaf)
//      return;
    else {

      if (Options.Components_Reckoning)
        Options.Components.fill_matrix_cond(Actual_AND_Node->ActualArg, Actual_AND_Node->ActualCond, Actual_AND_Node->Depth);




       Actual_AND_Node->Procreate(Train, Suspended_Nodes, Options);
                                                           // Canvia a PROGRESS i genera fills OR, els afegix a Suspended_Nodes
                                                           // Genera els seus corresponents AND, tots SUSPENDED
      if (Actual_AND_Node->Is_Leaf)  {  // NO HA CREAT FILLS.

          bool EXPE= false;
          if (Options.PostPruningMethod == PEP_POSTPRUNING) {
            EXPE= true;
          }
          if ((Options.Multitree.Construction != ONE_ONLY) || (Options.Multitree.Construction != ONE_AND_FORGET)
                || (EXPE)) {
            if ((Options.Multitree.BestTreeCrit == OCCAM_BEST)||(Options.Multitree.BestTreeCrit == CROSS_COVERAGE)
                 || (EXPE) || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                 || (Options.Comb2Single.Use_Occam))  {
              Actual_AND_Node->Accumulated.AccNumRules= 1;
            }
            if ((Options.Multitree.BestTreeCrit == TEST_COST_BEST)
                || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                || (Options.Comb2Single.Use_TestCost))  {
              Actual_AND_Node->Accumulated.TestCost= 0; // (DESPRS ES CALCULA EL COST D'ESTA FULLA);
            }
            else  if (Options.Multitree.BestTreeCrit == MDL_BEST)  {
              Actual_AND_Node->Accumulated.AccMDL= Actual_AND_Node->Crit.GMDL.cost;
            }
            if ((Options.Multitree.BestTreeCrit == EXPECTED_ERROR_BEST) || (EXPE)) {
              Actual_AND_Node->Accumulated.AccExpectedError= Actual_AND_Node->ExpectedError;
            }
            else  if (Options.Multitree.BestTreeCrit == SPLIT_OPTIMALITY_BEST)  {
              // DE MOMENT NO PROPAGA.
            }
            else  if (Options.Multitree.BestTreeCrit == ACCURACY_FALSE_TEST_BEST) {
              cout << "Multitree BestTreeCrit Option Not Implemented" << endl;
            }
       /*     else
               cout << "Multitree BestTreeCrit Option Unknown" << endl;  */
          }

        return;
      }

      // Si que ha creat fills



      or_tree *Actual_OR_Node = Actual_AND_Node->Select_Best_OR_For_Split(Options);
      if (Actual_OR_Node == NULL)  {
        cout << "NO ORs GENERATED UNDER AN AND-NODE !!!" << endl << endl;
        return;
      }
      else {
        if (!Actual_AND_Node->BestChild)
          Actual_AND_Node->BestChild= Actual_OR_Node;
      }


      if (Options.Multitree.Construction == ONE_AND_FORGET) {
         // ONE NODE HAS BEEN SELECTED.
         // WE HAVE TO FORGET ALL THE OTHERS.... (JUST TO FREE MEMORY)

        list<tree *>::iterator It;
        for(It= Actual_AND_Node->Children.begin(); It != Actual_AND_Node->Children.end(); It++) {
          or_tree *OT= (or_tree *)*It;
          if (OT != Actual_OR_Node) {  // only delete the nodes that hasn't been selected
 //           Actual_AND_Node->Children.remove(OT);  // Remove child from list of children
            Suspended_Nodes.remove(OT); // Remove from the suspended ors...
            delete OT;
          }
        }

        Actual_AND_Node->Children.clear();  // Remove all the children
        Actual_AND_Node->Children.push_back(Actual_OR_Node);  // Adds the one that has been selected


        if (Actual_AND_Node->Children.size() != 1) {
          cout << endl  << "ERROR IN METHOD multitree::Activate_AND()" << endl << endl;
          cout << "NOT ALL THE OTHER OR_NODES HAVE BEEN DELETED" << endl;
        }
      }


      if (Options.Multitree.SuspendedNodesForgetting != MAINTAIN_ALL) {
         // ONE NODE HAS BEEN SELECTED.
         // WE HAVE TO FORGET MANY OF THE OTHERS. JUST MAINTAIN A LOGARITHMIC PROPORTION OF THEM
         //   OR A CONSTANT NUMBER.... (JUST TO FREE MEMORY)
         long nChildren= 0;

         if (Options.Multitree.SuspendedNodesForgetting == MAINTAIN_LOG_RANDOM) {
           double nChild= Actual_AND_Node->Children.size();
           nChildren= (long)log(nChild);
         }
         else if (Options.Multitree.SuspendedNodesForgetting == MAINTAIN_LOG_RANDOM_WITH_DEPTH) {
           double nChild= Actual_AND_Node->Children.size();
           nChild /= (Actual_AND_Node->Depth + 1);
           nChildren= (long)log(nChild);
         }
         else if (Options.Multitree.SuspendedNodesForgetting == MAINTAIN_LOG_RANDOM_WITH_DEPTH_ADJUSTED) {
           double nChild= Actual_AND_Node->Children.size();
           nChild+= 8;
           nChild /= (Actual_AND_Node->Depth + 1);
           nChildren= (long)log(nChild);
         }
         else if (Options.Multitree.SuspendedNodesForgetting == MAINTAIN_LOG_RANDOM_WITH_SQUARED_DEPTH) {
           double nChild= Actual_AND_Node->Children.size();
           nChild /= sqrt(Actual_AND_Node->Depth + 1);
           nChildren= (long)log(nChild);
         }
         else if (Options.Multitree.SuspendedNodesForgetting == MAINTAIN_CONST_RANDOM) {
           nChildren= Options.Multitree.SuspendedNodesConstMaintain;
         }

         if (nChildren < 0)
           nChildren= 0;

         list<tree *>::iterator It;
         while ((long)Actual_AND_Node->Children.size() > nChildren +1) {
           long s= Actual_AND_Node->Children.size();
           long j= (long)RandUniform(0,s);
           long i=0;

           for(It= Actual_AND_Node->Children.begin(); It != Actual_AND_Node->Children.end(); It++) {
             if (i == j) {
               or_tree *OT= (or_tree *)*It;
               if (OT != Actual_OR_Node) {  // only delete the nodes that hasn't been selected
                 Actual_AND_Node->Children.remove(OT);  // Remove child from list of children
                 Suspended_Nodes.remove(OT); // Remove from the suspended ors...
                 delete OT;
               }
               break;
             }
             i++;
           }
         }

        if (Actual_AND_Node->Children.size() < 1) {
          cout << endl  << "ERROR IN METHOD multitree::Activate_AND()" << endl << endl;
          cout << "TOO MANY OR_NODES HAVE BEEN DELETED" << endl;
        }

      }

      // Set the Second_Optimality for Multitree
      if ((Options.Multitree.Construction != ONE_ONLY) || (Options.Multitree.Construction != ONE_AND_FORGET)){
        if ((Options.Multitree.SecondTreeCrit.Met == OPTIMALITY_RIVAL_RATIO) ||
             (Options.Multitree.SecondTreeCrit.Met == OPTIMALITY_RIVAL_RATIO_DEPTH) ||
             (Options.Multitree.SecondTreeCrit.Met == OPTIMALITY_RIVAL_RATIO_COMPONENT) ||
             (Options.Multitree.SecondTreeCrit.Met == OPTIMALITY_RIVAL_RATIO_COMPONENT_RANDOM)) {
           float BestOpt= Actual_OR_Node->Split_Optimality;
           list<tree *>::iterator It, ItEnd;
           ItEnd= Actual_AND_Node->Children.end();
           for(It= Actual_AND_Node->Children.begin(); It != ItEnd; It++) {
              or_tree *OT= (or_tree *)*It;
              OT->Second_Optimality= OT->Split_Optimality / BestOpt;
           }
        }
      }


      // TWO-LEVEL PRUNING
      bool Pruned= false;
      if (Options.PC.Met == EXPECTED_ERROR_PRUNING)  {
        if (Actual_AND_Node->ExpectedError < Options.PC.ExpectedError.Factor * Actual_OR_Node->BackupError)
          Pruned=true;
      }
      else if (Options.PC.Met == PEP_PREPRUNING)  {
       // en pre-pruning no t sentit
        double PEPFACTOR = 0.5;
        double Card= Actual_AND_Node->Cardinality;
        double E_Parent = Actual_AND_Node->ExpectedError * Card + PEPFACTOR;
        double E_Children = Actual_OR_Node->BackupError * Card + PEPFACTOR * (Actual_OR_Node->Children).size();
        double SE = sqrt ( E_Children * (Card - E_Children) / Card );
        if (E_Parent <= (E_Children - SE/2))
          Pruned=true;
      }
      else if (Options.PC.Met == MDL_PRUNING)  {
        if (Actual_AND_Node->Crit.GMDL.cost < Options.PC.MDL.Factor * Actual_OR_Node->Crit.GMDL.cost)
           Pruned= true;
      }

      if (Pruned) {
        Actual_AND_Node->Is_Leaf= true;
        Actual_AND_Node->Children.clear(); // Empties the list of children
      }
      else {
        Activate_OR(Actual_OR_Node, Train);
      }


      // PROPAGATION
      Propagate_AND(Actual_AND_Node, Actual_OR_Node);

    }
  }

  // Removes OR from Suspended_Nodes and Activates it
  void Activate_OR(or_tree *Actual_OR_Node, const dataset &Train) {
       Suspended_Nodes.remove(Actual_OR_Node);

        Actual_OR_Node->Change_State(PROGRESS);

        Propagate_OR_To_Zero(Actual_OR_Node);

        list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_OR_Node->Children.end();
        for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
          // ((and_tree *)*It)->Change_State(NEEDED);
          and_tree *AT= (and_tree *)*It;
          #ifdef TREE_DEBUG
             cout << "Activant Node #" << AT->Num << ": " << AT->Rule.ToStringUnfolded("f", Train, false) << " Card: " << AT->Cardinality << endl;
          #endif
          Activate_AND(AT, Train);

          // Propagation
          Propagate_1_OR(Actual_OR_Node, AT);


       }

  }


  void Propagate_AND(and_tree *Actual_AND_Node, or_tree *Actual_OR_Node) {
    bool EXPE= false;
    if (Options.PostPruningMethod == PEP_POSTPRUNING) {
      EXPE= true;
    }

          if ((Options.Multitree.Construction != ONE_ONLY) || (Options.Multitree.Construction != ONE_AND_FORGET)
              || (EXPE)) {
            if ((Options.Multitree.BestTreeCrit == OCCAM_BEST)||(Options.Multitree.BestTreeCrit == CROSS_COVERAGE)
                  || (EXPE) || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                  || (Options.Comb2Single.Use_Occam))  {
              if ((Actual_AND_Node->Accumulated.AccNumRules == -1) ||
                 (Actual_AND_Node->Accumulated.AccNumRules > Actual_OR_Node->Accumulated.AccNumRules))
                 Actual_AND_Node->Accumulated.AccNumRules= Actual_OR_Node->Accumulated.AccNumRules;
            }
            if ((Options.Multitree.BestTreeCrit == TEST_COST_BEST)
               || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
               || (Options.Comb2Single.Use_TestCost)) {
              if ((Actual_AND_Node->Accumulated.TestCost == -1.0) ||
                 (Actual_AND_Node->Accumulated.TestCost > Actual_OR_Node->Accumulated.TestCost))
                 Actual_AND_Node->Accumulated.TestCost= Actual_OR_Node->Accumulated.TestCost;
            }
            else  if (Options.Multitree.BestTreeCrit == MDL_BEST)  {
              if ((Actual_AND_Node->Accumulated.AccMDL == -1.0) ||
                 (Actual_AND_Node->Accumulated.AccMDL > Actual_OR_Node->Accumulated.AccMDL))
                 Actual_AND_Node->Accumulated.AccMDL= Actual_OR_Node->Accumulated.AccMDL;
            }
            if ((Options.Multitree.BestTreeCrit == EXPECTED_ERROR_BEST) || (EXPE))   {
              if ((Actual_AND_Node->Accumulated.AccExpectedError == -1.0) ||
                  (Actual_AND_Node->Accumulated.AccExpectedError > Actual_OR_Node->Accumulated.AccExpectedError))
                 Actual_AND_Node->Accumulated.AccExpectedError= Actual_OR_Node->Accumulated.AccExpectedError;
            }
            else  if (Options.Multitree.BestTreeCrit == SPLIT_OPTIMALITY_BEST)  {
              // DE MOMENT NO PROPAGA.
            }
        /*    else
               cout << "Multitree BestTreeCrit Option Unknows" << endl; */
          }

  }

  void Propagate_OR_To_Zero(or_tree *Actual_OR_Node) {
    bool EXPE= false;
    if (Options.PostPruningMethod == PEP_POSTPRUNING) {
      EXPE= true;
    }

          if ((Options.Multitree.Construction != ONE_ONLY) || (Options.Multitree.Construction != ONE_AND_FORGET)
              || (EXPE))  {
            if ((Options.Multitree.BestTreeCrit == OCCAM_BEST)||(Options.Multitree.BestTreeCrit == CROSS_COVERAGE)
                  || (EXPE) || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                  || (Options.Comb2Single.Use_Occam))  {
               Actual_OR_Node->Accumulated.AccNumRules= 0;
            }
            if ((Options.Multitree.BestTreeCrit == TEST_COST_BEST)
                || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                || (Options.Comb2Single.Use_TestCost)) {
               Actual_OR_Node->Accumulated.TestCost= 0.0;
            }
            else  if (Options.Multitree.BestTreeCrit == MDL_BEST)  {
               Actual_OR_Node->Accumulated.AccMDL= 0.0;
            }
            if ((Options.Multitree.BestTreeCrit == EXPECTED_ERROR_BEST) || (EXPE))   {
               Actual_OR_Node->Accumulated.AccExpectedError= 0.0;
            }
            else  if (Options.Multitree.BestTreeCrit == SPLIT_OPTIMALITY_BEST)  {
              // DE MOMENT NO PROPAGA.
            }
         /*   else
               cout << "Multitree BestTreeCrit Option Unknown" << endl; */
          }
  }

  void Propagate_1_OR(or_tree *Actual_OR_Node, and_tree *AT) {
    bool EXPE= false;
    if (Options.PostPruningMethod == PEP_POSTPRUNING) {
      EXPE= true;
    }

          if ((Options.Multitree.Construction != ONE_ONLY) || (Options.Multitree.Construction != ONE_AND_FORGET)
              || (EXPE)) {
            if ((Options.Multitree.BestTreeCrit == OCCAM_BEST)||(Options.Multitree.BestTreeCrit == CROSS_COVERAGE)
                  || (EXPE) || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                  || (Options.Comb2Single.Use_Occam))  {
               Actual_OR_Node->Accumulated.AccNumRules+= AT->Accumulated.AccNumRules;
            }
            if ((Options.Multitree.BestTreeCrit == TEST_COST_BEST)
                 || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                 || (Options.Comb2Single.Use_TestCost))  {
               // AC ACTUALITZEM EL AT->Accumulated.TestCost i desprs sumem;

               int Arg= AT->ActualArg;
               double TestCost= Options.TestCostVector[Arg];

               TestCost *= AT->Cardinality;  // The cost must be weighted with the cardinality

               // Anem a mirar si este argument l'hem de contar (no siga que ja es gaste dalt
               or_tree *o2 = Actual_OR_Node;
               while(true) {
                 and_tree *a2= (and_tree *)o2->Parent;

                 if (a2->Parent == NULL)  // If it is the root. bye. The root does not have condition.
                   break;

                 if (a2->ActualArg == Arg) {  // this argument has already been used. We don't have to take it into account.
                   TestCost= 0;
                   break;
                 }
                 o2= (or_tree *)a2->Parent;
               }

               // Actualitzem l'acumulat
               AT->Accumulated.TestCost+= TestCost;

               // Ara propaguem
               Actual_OR_Node->Accumulated.TestCost+=  AT->Accumulated.TestCost;
            }
            else  if (Options.Multitree.BestTreeCrit == MDL_BEST)  {
               Actual_OR_Node->Accumulated.AccMDL+=  AT->Accumulated.AccMDL;
            }
            if ((Options.Multitree.BestTreeCrit == EXPECTED_ERROR_BEST)  || (EXPE))  {
               Actual_OR_Node->Accumulated.AccExpectedError+=  (AT->Cardinality / float(((and_tree *)Actual_OR_Node->Parent)->Cardinality)) *
                                                                     AT->Accumulated.AccExpectedError;
            }
            else  if (Options.Multitree.BestTreeCrit == SPLIT_OPTIMALITY_BEST)  {
              // DE MOMENT NO PROPAGA.
            }
         /*   else
               cout << "Multitree BestTreeCrit Option Unknown" << endl; */
          }
  }

  void Propagate_OR(or_tree *Actual_OR_Node) {
       Propagate_OR_To_Zero(Actual_OR_Node);

        list<tree *>::iterator It, ItEnd;
        ItEnd= Actual_OR_Node->Children.end();
        for(It= Actual_OR_Node->Children.begin(); It != ItEnd; It++) {
          and_tree *AT= (and_tree *)*It;

          // Propagation
          Propagate_1_OR(Actual_OR_Node, AT);
      }
  }


  void PropagateUpwards(or_tree *OT)  {
//    cout << "PROPAGATE PER FER" << endl;
    and_tree *AT= (and_tree *)OT->Parent;
    Propagate_AND(AT,OT);
    OT= (or_tree *)AT->Parent;
     if (OT == NULL)
       return;
    Propagate_OR(OT);         // Torna a sumar-los tots. Seria ms eficient si calcularem la diferencia del AT abans de propagar i ara.
    PropagateUpwards(OT);
 }

  string ShowNodeBest(and_tree *Actual_AND_Node, char *FunctionName, string Desp,
                                    const dataset &DataSet, bool Short, long &NumRules, bool All) const;


};



inline void Initialise_AND_for_Propagation(and_tree *Actual_AND_Node, const options &Options) {
    bool EXPE= false;
    if (Options.PostPruningMethod == PEP_POSTPRUNING) {
      EXPE= true;
    }

          if ((Options.Multitree.Construction != ONE_ONLY) || (Options.Multitree.Construction != ONE_AND_FORGET)
              || (EXPE)) {
            if ((Options.Multitree.BestTreeCrit == OCCAM_BEST)||(Options.Multitree.BestTreeCrit == CROSS_COVERAGE)
                 || (EXPE) || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                 || (Options.Comb2Single.Use_Occam))  {
              Actual_AND_Node->Accumulated.AccNumRules = -1;
            }
            if ((Options.Multitree.BestTreeCrit == TEST_COST_BEST)
                || (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST)
                || (Options.Comb2Single.Use_TestCost)) {
              Actual_AND_Node->Accumulated.TestCost= -1.0;
            }
            else  if (Options.Multitree.BestTreeCrit == MDL_BEST)  {
              Actual_AND_Node->Accumulated.AccMDL= -1.0;
            }
            if ((Options.Multitree.BestTreeCrit == EXPECTED_ERROR_BEST) || (EXPE))  {
              Actual_AND_Node->Accumulated.AccExpectedError= -1.0;
            }
            else  if (Options.Multitree.BestTreeCrit == SPLIT_OPTIMALITY_BEST)  {
              // DE MOMENT NO PROPAGA.
            }
       /*     else
               cout << "Multitree BestTreeCrit Option Unknown" << endl; */
          }
}




  // This function applies the new condition of an and_tree
  // and computes the cardinality and the class distribution
  inline void and_tree::Apply_Condition(const dataset &Train) {
    and_tree *at= (and_tree *)(Parent->Parent) ;

    ClassDistribution= at->ClassDistribution;  // Same as grandpa
    Cardinality = 0;
    vector<indext> & CoveredGrandPa = ((and_tree *)(Parent->Parent))->Covered;  // The examples covered by the grandpa (upper and node)
    vector<indext>::iterator It, ItEnd;
    ItEnd= CoveredGrandPa.end();
    for(It= CoveredGrandPa.begin(); It != ItEnd; It++) {
      indext Ind= (*It);
      if (Train.FollowsCondition(ActualArg, ActualCond, Ind)) {
        Covered.push_back(Ind);  // If the example follows the condition we add it to covered
        Cardinality++;
      }
      else { // if the example is not covered. We have to decrement the cardinality for this class
        discrete Class= Train.ClassOf(Ind);
        ClassDistribution[Class]--;
      }
    }
  }
  // Sets Predicted Class
  inline void and_tree::SetsPredictedClass(const dataset &Train, const options & Options) {

    int NumClasses= Train.NumClasses();

    PredictedClass= 0; // ES PODRIA GASTAR UN VALOR ESPECIAL, PERO DE MOMENT ANEM A FER A

    if (Cardinality == 0) {

       if (Options.Weights.ClassSelection == MAJORITY_CLASS) {
           PredictedClass= Train.MajorityClass();
         }
         else if ((Options.Weights.ClassSelection == STRATIFICATION_CLASS)
                  || (Options.Weights.ClassSelection == STRATIFICATION_CLASS_WITHOUT_SMOOTHING)) {
           discrete MaxClass;
           double Max= -1;
           for(int i=0; i<NumClasses; i++) {  //

             double Res= Train.ClassDistribution()[i] * Options.WeightVector[i];
             if (Res > Max) {
               Max= Res;
               MaxClass= i;
             }
           }
           PredictedClass= MaxClass;
         }
        // EN EL CAS DE MIN_COST_CLASS SE AGARRA LA CLASSE AMB MENYS COST (sumant la fila)
        else if ((Options.Weights.ClassSelection == MIN_COST_CLASS)
                    || (Options.Weights.ClassSelection == MIN_COST_CLASS_WITHOUT_SMOOTHING)) {
         float MinCost= 0.0;
         indext MaxClassDist = 0;
         for(int i=0; i<NumClasses; i++) {  // Class will be the less costly
           float CostClassI= 0.0;
           for(int j=0; j<NumClasses; j++) {
             CostClassI += Options.CostMatrix.AccessRP(j,i);
           }
           if ((CostClassI < MinCost) || (i==0)) {
             MinCost= CostClassI;
             PredictedClass= i;
             MaxClassDist= Train.ClassDistribution()[i];
          }
          else if (CostClassI == MinCost) {   // if equal get the one with more frequency
            if (ClassDistribution[i] >= MaxClassDist) {
             MinCost= CostClassI;
             PredictedClass= i;
             MaxClassDist= Train.ClassDistribution()[i];
            }
           }
          }
        }
      return;
    }

    if (Options.Weights.ClassSelection == MAJORITY_CLASS) {
      indext Max= 0;
      for(int i=0; i<NumClasses; i++) {  // Class will be the majority class
        indext iValue= ClassDistribution[i];

        if (iValue > Max) {
          Max= iValue;
          PredictedClass= i;
        }
      }
    }
    else if ((Options.Weights.ClassSelection == STRATIFICATION_CLASS)
              || (Options.Weights.ClassSelection == STRATIFICATION_CLASS_WITHOUT_SMOOTHING))   {
           discrete MaxClass;
           double Max= -1;
           for(int i=0; i<NumClasses; i++) {  //
             double iValue;
              if (Options.Weights.ClassSelection == STRATIFICATION_CLASS)
               iValue  = SmoothDistribution[i];
             else if (Options.Weights.ClassSelection == STRATIFICATION_CLASS_WITHOUT_SMOOTHING)
               iValue  = ClassDistribution[i];

             double Res= iValue * Options.WeightVector[i];
             if (Res > Max) {
               Max= Res;
               MaxClass= i;
             }
           }
           PredictedClass= MaxClass;
     }
    else if ((Options.Weights.ClassSelection == MIN_COST_CLASS)
             || (Options.Weights.ClassSelection == MIN_COST_CLASS_WITHOUT_SMOOTHING)) {
      float MinCost= 0.0;
      double MaxClassDist = 0;
      for(int i=0; i<NumClasses; i++) {  // Class will be the less costly
        float CostClassI= 0.0;
        for(int j=0; j<NumClasses; j++) {
           double jValue;
            if (Options.Weights.ClassSelection == MIN_COST_CLASS)
              jValue  = SmoothDistribution[j];
            else if (Options.Weights.ClassSelection == MIN_COST_CLASS_WITHOUT_SMOOTHING)
              jValue  = ClassDistribution[j];
          CostClassI += jValue * Options.CostMatrix.AccessRP(j,i);   // A JA EST CALCULAT EN UN ALTRE VECTOR. ES PODRIA UTILIZAR.
        }

        double iValue;
            if (Options.Weights.ClassSelection == MIN_COST_CLASS)
              iValue  = SmoothDistribution[i];
            else if (Options.Weights.ClassSelection == MIN_COST_CLASS_WITHOUT_SMOOTHING)
              iValue  = ClassDistribution[i];

        if ((CostClassI < MinCost) || (i==0)) {
           MinCost= CostClassI;
           PredictedClass= i;
           MaxClassDist= iValue;
        }
        else if (CostClassI == MinCost) {   // if equal get the one with more frequency
          if (iValue >= MaxClassDist) {
           MinCost= CostClassI;
           PredictedClass= i;
           MaxClassDist= iValue;
          }
         }
      }
    }
    else {
      cout << "Option for class selection unknown" << endl;
    }

  }


  // Computes Costs vectors and probabilities derived from cost
  inline void and_tree::ComputeCosts(const dataset &Train, const options & Options)  {
     if (Cardinality == 0)
       return;

     int NumClasses= Train.NumClasses();

     if  (Options.Weights.Method != NO_COSTS) {

       float SumInvCost= 0;

        for(int i=0; i<NumClasses; i++) {  // Class will be the less costly
         double CostClassI= 0.0;
          for(int j=0; j<NumClasses; j++) {
            double iValue;
            if (Options.Cost_Derived_Probability_Method == DIRECT)
              iValue= ClassDistribution[j];
            else if (Options.Cost_Derived_Probability_Method == WITH_SMOOTHING)
              iValue= Cardinality * SmoothDistribution[j];
           CostClassI += iValue * Options.CostMatrix.AccessRP(j,i);
          }
          CostClassI+= 0.000001; // Avoid CostClass == 0. It's similar to an epsilon correction.
                                               // It may be interesting to consider this number as a parameter
         if (CostClassI == 0) { // Encara podria ser 1.
              CostClassI= 0.000001;  // MILLORAR ESTE TEMA
         }

         CostClass.push_back();
         CostClass[i]= CostClassI;
         SumInvCost+= 1 / CostClassI;
       }

       for(int i=0; i<NumClasses; i++) {
          CostProb.push_back();
          CostProb[i] = 1 / (CostClass[i] * SumInvCost)  ;
       }
     }
  }

  // Computes Smoothing of class frequencies
  inline void and_tree::ComputeSmoothing(const dataset &Train, const options & Options)  {
       if (Cardinality == 0) {
         // cout << "Error: computing smooth distribution of an empty node" << endl;
         // EN REALITAT NO PASSA RES
       }
       else {


     int NumClasses= Train.NumClasses();

     if (Options.Smoothing.Method == NO_SMOOTHING) {
        for(int i=0; i < NumClasses; i++) {
         SmoothDistribution.push_back();
         SmoothDistribution[i]= ClassDistribution[i] / float(Cardinality);
      }
     }
     else if (Options.Smoothing.Method == LAPLACE) {
      for(int i=0; i < NumClasses; i++) {
       SmoothDistribution.push_back();
       SmoothDistribution[i]= (ClassDistribution[i] + 1) / float(Cardinality + NumClasses);
      }
    }
    else if (Options.Smoothing.Method == K_ESTIMATE) {
      for(int i=0; i < NumClasses; i++) {
       int M= Options.Smoothing.m;
       SmoothDistribution.push_back();
       SmoothDistribution[i]= (ClassDistribution[i] + M) / float(Cardinality + M);
      }
    }
    else if (Options.Smoothing.Method == M_ESTIMATE) {
      int M= Options.Smoothing.m;
      for(int i=0; i < NumClasses; i++) {
       float Freq= Train.FreqOfClass(i);
       SmoothDistribution.push_back();
       SmoothDistribution[i]= (ClassDistribution[i] + Freq * M) / float(Cardinality + M);
      }
   }
    else if (Options.Smoothing.Method == M_ESTIMATE_UNIFORM) {
      int M= Options.Smoothing.m;
     float Freq= 1/float(NumClasses);
     for(int i=0; i < NumClasses; i++) {
       SmoothDistribution.push_back();
       SmoothDistribution[i]= (ClassDistribution[i] + Freq * M) / float(Cardinality + M);
      }
   }
   else {
     cout << "Smoothing method unknown" << endl;
   }

 }

 }

  // Computes ExpectedError
  inline void and_tree::ComputeExpectedError(const dataset &Train, const options & Options) {

   if (Cardinality == 0) {
     ExpectedError= 0;
     return;
   }

    int NumClasses= Train.NumClasses();
    if (Options.ExpectedError.Method == NO_COMPUTE) {
    }
   else if (Options.ExpectedError.Method == RELATIVE_FREQUENCY_WITH_MAJORITY_CLASS) {
        ExpectedError= (1- SmoothDistribution[PredictedClass]);
    }
    else if (Options.ExpectedError.Method == RELATIVE_FREQUENCY_WITH_FREQ_PROB) {
        float Error= 0.0;
        for(int i=0; i<NumClasses; i++) {
          float pi= SmoothDistribution[i];
          Error+= pi * (1-pi);
        }
        ExpectedError= Error;
     }
   else if (Options.ExpectedError.Method == COST_WITH_MINIMUM_CLASS) {
        float Cost= 0.0;
        for(int j=0; j<NumClasses; j++) {
          indext iValue= ClassDistribution[j];
          Cost += iValue * Options.CostMatrix.AccessRP(j,PredictedClass);
        }
        ExpectedError= Cost;
    }
    else if (Options.ExpectedError.Method == COST_WITH_FREQ_PROB) {
        float Cost= 0.0;
        for(int i=0; i<NumClasses; i++) {
          float Cost2= 0.0;
          for(int j=0; j<NumClasses; j++) {
            indext iValue= ClassDistribution[j];
            Cost2 += iValue * Options.CostMatrix.AccessRP(j,i);
          }
          Cost+= Cost2 * SmoothDistribution[i];
        }
        ExpectedError= Cost;
    }
    else if (Options.ExpectedError.Method == COST_WITH_COST_PROB) {
        float Cost= 0.0;
        for(int i=0; i<NumClasses; i++) {
          float Cost2= 0.0;
          for(int j=0; j<NumClasses; j++) {
            indext iValue= ClassDistribution[j];
            Cost2 += iValue * Options.CostMatrix.AccessRP(j,i);
          }
          Cost+= Cost2 * CostProb[i];
        }
        ExpectedError= Cost;
    }
    else {
       cout << "ExpectedError Method Unknown or not implemented" << endl;
    }

  }

  // Generates the OR children, adding them to Suspended_Nodes
  // It also generates its corresponding AND and stops at this level
  inline void and_tree::Procreate(const dataset & Train, list<or_tree *> & Suspended_Nodes, const options & Options) {

  // MILLORA. QUAN SE FA ADD_CHILD S'HAURIA D'INSERTAR ORDENADAMENT DEPENENT DE SUSPENDED/PROGRESS I DEL CRITERI DE SEGON TREE.


    if (Cardinality == 0)  {     // this node has no examples. It's useless!!!!  PERO NO PASSA RES
      Is_Leaf= true;
      #ifdef TREE_DEBUG
        cout << "Warning: Useless Node Generated" << endl;
      #endif
      return;
    }

    if (ClassDistribution[PredictedClass] == Cardinality) {     // There are only examples of one class
      Is_Leaf= true;
      return;
    }

    bool Found= false;
    bool Pruned= false;

      // ONE LEVEL PRUNING
    if (Options.PC.Met == PROPORTIONAL) {
      indext Card= Cardinality;
      indext Max= ClassDistribution[PredictedClass];
      float Accuracy= Max / float(Card);
      float Error= 1-Accuracy;
      indext MisClassified= Card - Max;

       // if ((Error < ((NumClasses-1)/2.0) /float(NumClasses)) && (MisClassified < 10))
       //   Pruned= true;
      if ((const_cast<options &>(Options)).PC.To_Be_Pruned(Error, Train.NumClasses(), MisClassified) )
        Pruned= true;
    }
    else if (Options.PC.Met == STUMP_PRUNING) {
      if (Options.PC.Stump.MaxDepth == Depth)
        Pruned= true;
    }

    if (!Pruned) {
      int NumArg= Train.NumArguments();
	    for(int a=0; a<NumArg; a++) {

	      if (OpenVariables[a]) {
	        kind K= Train.KindOf(a);
 	        if (K==DISCRETE_KIND) {
            int NumConst= Train.NumConstants(a);
            or_tree *OT;

            if (NumConst > 1)  {
              if (Options.Enabled_Partitions[PARTITION_MANY_CONST_EQ]) {
                Found= true;
	              // Partition X=a, X=b, X=c, ...
	              OT= new or_tree(this, Train, a, PARTITION_MANY_CONST_EQ, Options);
	              Add_Child((tree *)OT);
	              Suspended_Nodes.push_back(OT);
              }

              if (Options.Enabled_Partitions[PARTITION_ONE_CONST_EQ]) {
	            // Partition X=a, X!=a
                if ((NumConst > 2) || (!Options.Enabled_Partitions[PARTITION_MANY_CONST_EQ]))  {  // IF THE NUMBER OF VALUES <= 2 IT'S BETTER TO USE THE PREVIOUS PARTITION
                  for (discrete v=0; v < NumConst; v++) {
                    argument Arg= argument(v);

                    rule RuleCopy= Rule;
                    condition Cond=  condition(CONST_NEQ, Arg);
                    if (RuleCopy.Add_Condition(a, Cond) == false) { // Existent condition. It would lead to an infinite loop.
                      // do nothing. Don't create the new node;
                    }
                    else {
                      Found= true;

                      // EN REALITAT S'HAURIA DE MIRAR SI QUEDEN EIXEMPLES AMB EIXA CONSTANT, PER A OPTIMITZAR
	                    OT= new or_tree(this, Train, a, PARTITION_ONE_CONST_EQ, Options, Arg);
	                    Add_Child((tree *)OT);
	                    Suspended_Nodes.push_back(OT);
                    }  // if-else
                  }  // for
                }  // if
              }   // if
            }  // if

            if (Options.Enabled_Partitions[PARTITION_ONE_VARIABLE_EQ]) {
	          // Partition X=Y, X!=Y

              int TypeVar= Train.TypeOf(a);
              for(int b=a+1; b<NumArg; b++) {  // just look for arguments after 'a' (arguments before have been done in the other way)
                if ((OpenVariables[b]) && (Train.TypeOf(b) == TypeVar)) {
                  argument Arg= argument(b);

                  rule RuleCopy= Rule;
                  condition Cond=  condition(VAR_NEQ, Arg);
                  if (RuleCopy.Add_Condition(a, Cond) == false) { // Existent condition. It would lead to an infinite loop.
                    // do nothing. Don't create the new node;
                  }
                  else {
                    Found= true;
    	              OT= new or_tree(this, Train, a, PARTITION_ONE_VARIABLE_EQ, Options, Arg);
	                  Add_Child((tree *)OT);
	                  Suspended_Nodes.push_back(OT);
                  }
                }
              }
            }
	        }
	        else if ((K==NUMERIC_KIND) || (K==NUMERIC_WITH_UNKNOWN_KIND)) {
            or_tree *OT;
            if (Options.Enabled_Partitions[PARTITION_ONE_CONST_DISEQ]) {
              // Partition X<c, X>=c
              set<numeric> s;
              Train.OrderedSetOfRealsOf(a, s, Covered);     // Returns in s a list of reals appeared in argument a (of Covered).

              if (K == NUMERIC_WITH_UNKNOWN_KIND) {
                s.erase((numeric)MISSING_NUMERIC);
                // removes numbers that are representing a missing value
              }

              indext Inc, Iterations;
              indext Size= s.size();

              if (Size != 0) { // if the set is empty there is no case in making a partition.
                indext MaxNumericIntervals= 0;

                if (Options.IntervalCriterion.Met == MAX) {
                  MaxNumericIntervals= Options.IntervalCriterion.Param;
                }
                else if (Options.IntervalCriterion.Met == LOG) {
                  MaxNumericIntervals= (indext)logbase(Options.IntervalCriterion.Base, Size);
                   if (MaxNumericIntervals < (unsigned)Options.IntervalCriterion.Param)
                     MaxNumericIntervals = Options.IntervalCriterion.Param;
                }

                if ((MaxNumericIntervals > 0)  && (MaxNumericIntervals < Size))  {
                   Iterations= MaxNumericIntervals;
                   Inc= Size / Iterations;
                }
                else {
                  Inc= 1;
                  Iterations= Size;
                }
                set<numeric>::iterator i=s.begin();
                numeric Prev= *i;
                i++; Iterations--;
                indext j=0;
                for(; j < Iterations; j++) {   // If there is only one it doesn't enter.
                  numeric Current= *i;
                  numeric Middle= (Prev + Current) / 2.0;   // Add partition on the middle of the interval.
                  argument Arg= argument(Middle);

                  if (K == NUMERIC_KIND) {
                    Found= true;
	                  OT= new or_tree(this, Train, a, PARTITION_ONE_CONST_DISEQ, Options, Arg);
	                }
	                else { // NUMERIC_WITH_UNKNOWN_KIND
	                  rule RuleCopy= Rule;
                    condition Cond=  condition(NUM_MISSING, Arg); // the child condition used to check this is irrelevant
                    if (RuleCopy.Add_Condition(a, Cond) == false) { // Existent condition. It would lead to an infinite loop.
                      OT= NULL; // do nothing. Don't create the new node;
                      // Parent condition is X = NUM_MISSING. It doesn't make sense to use this argument again.
                    }
                    else {
		                  Found= true;
		                  OT= new or_tree(this, Train, a, PARTITION_ONE_CONST_DISEQ_WITH_UNKNOWN, Options, Arg);
		                }
	                }
	
  	              if (OT) {
	                  Add_Child((tree *)OT);
	                  Suspended_Nodes.push_back(OT);
                  }
                  Prev= Current;
                  for(indext k= 0; k<Inc; k++) {        // Increments i as many times as Inc
                    i++;
                  }
                }
              }
            }
	        }
 	        else if (K==IGNORE_KIND) {
 	          // Nothing to do
 	        }
	      }  // if
	    }  // for

    } // if


    if (Found && !Pruned)
      Is_Leaf= false;

    if ((!Found) || (Pruned)) { // NO HI HAN OPEN_VARIABLES O SE PODA
      Is_Leaf= true;
    }

      #ifdef TREE_DEBUG
        if ((Children.size() == 0) != Is_Leaf)
          cout << "Discrepancies between children and Is_Leaf" << endl;
      #endif
  }


void Initialise_Compare_TestCost_And_NumRules_Factor(double );

// Just selects the comparison function to be used for the following functions
inline void Obtain_CompFunction(const options &Options, bool (* &Comp_Function)(tree *, tree *)) {
    if (Options.Multitree.BestTreeCrit == EXPECTED_ERROR_BEST) {
       //  Comp_Function= &Compare_ExpectedError;
       Comp_Function= &Compare_AccExpected_Error;
     }
     else if (Options.Multitree.BestTreeCrit ==  SPLIT_OPTIMALITY_BEST) {
	     Comp_Function= &Compare_Split_Optimality;
     }
     else if (Options.Multitree.BestTreeCrit == TEST_COST_BEST) {
       Comp_Function= &Compare_TestCost;
     }
     else if (Options.Multitree.BestTreeCrit == OCCAM_BEST) {
       Comp_Function= &Compare_NumRules;
     }
     else if (Options.Multitree.BestTreeCrit == OCCAM_AND_TEST_COST_BEST) {
       Comp_Function= &Compare_TestCost_And_NumRules;
       Initialise_Compare_TestCost_And_NumRules_Factor(Options.TestCostRelevanceInSelectBest);
     }
     else if (Options.Multitree.BestTreeCrit == CROSS_COVERAGE) {
       Comp_Function= &Compare_Cross_Coverage;
     }
     else if (Options.Multitree.BestTreeCrit == MDL_BEST) {
	     Comp_Function= &Compare_MDL;
     }
     else
      cout << "Error. Option for Selected the Best Or Unknown" << endl;
}


// Selects the best tree according to the best tree criterion
inline or_tree *and_tree::Select_Best_OR_For_Show(const options &Options) {
   or_tree *OT=NULL;

  if (Children.size() == 0)
    return NULL;
  else {
     bool (*Comp_Function)(tree *, tree *);

     Obtain_CompFunction(Options, Comp_Function); // Depending on OCCAM_BEST, TEST_COST, ...
     // this is a very inefficient way to do it. Because the options doesn't change for all the program execution-
     // it could be better to store Comp_Function in options and use that again and again without recomputing it.

   OT= (or_tree *) *max_element(Children.begin(), Children.end(), Comp_Function);

  }
  return (OT) ;
}

// Selects the best tree according to the n-best different criterion
inline  or_tree *and_tree::Select_Best_OR_Different_Using_Visits(const options &Options) {
  or_tree *OT=NULL;

  if (Children.size() == 0)
    return NULL;
  else {
   bool (*Comp_Function)(tree *, tree *);

   bool (*Comp_Function2)(tree *, tree *);

   Obtain_CompFunction(Options, Comp_Function2);

   Initialise_CompFunction_For_Compare_Meta_For_Different(Comp_Function2);

   Comp_Function= &Compare_Meta_For_Different_Using_Visits; // This function also calls Obtain_CompFunction

   OT= (or_tree *) *max_element(Children.begin(), Children.end(), Comp_Function);

  }
  return (OT) ;

}

// Selects the best tree according to the n-best different criterion
inline  or_tree *and_tree::Select_Best_OR_Different_Using_Components(const options &Options, const component_matrix &M) {
  or_tree *Best_Tree= NULL;

  if (Children.size() == 0)
    return NULL;
  else {
  // bool (*Comp_Function)(tree *, tree *);

   bool (*Comp_Function2)(tree *, tree *);

   Obtain_CompFunction(Options, Comp_Function2);

 	 list<tree *>::iterator It, ItEnd;
 	 It= Children.begin();
 	 ItEnd= Children.end();
 	 	
	 for(; It != ItEnd; It++) {
	  or_tree *Child= (or_tree *)*It;
	
	  if (Child == NULL)
	    cout << "CHILD NULL EN and_tree::Select_Best_OR_Different_Using_Components" << endl;
	
	  if (Child->State == PROGRESS) {
	    if (Best_Tree == NULL)
	      Best_Tree = Child;
	    else {
	     and_tree *a1Pa= (and_tree *)Best_Tree->Children.front();
	     and_tree *a2Pa= (and_tree *)Child->Children.front();
	  	 float v1= M.AccessRP(a1Pa->ActualArg, get_partition(a1Pa->ActualCond));
       float v2= M.AccessRP(a2Pa->ActualArg, get_partition(a2Pa->ActualCond));


	      float OpA;
	      float OpB;

       if (Options.SelectKBest.Met == K_BEST_DIFFERENT_COMPONENTS) {
         OpA= -v1;
	       OpB= -v2;
	     }
       else if (Options.SelectKBest.Met == K_BEST_LESS_VISITED_THEN_DIFFERENT_COMPONENTS) {
         if (Best_Tree->Visits > Child->Visits) {
           OpA = 0; OpB= 1;
         }
         else if (Best_Tree->Visits < Child->Visits) {
           OpA = 1; OpB= 0;
         }
         else {
           OpA= -v1;
           OpB= -v2;
           // OpA= 1; This would ignore components information
           // OpB= 1;
         }

       }

	      if (OpA < OpB) {
	        Best_Tree= Child;
	      }
		    else if ( OpA == OpB) {
		      if (Comp_Function2(Best_Tree, Child))
		        Best_Tree= Child;
		    }
	    }
	  }
	
	 }
/*
	 if (Best_Tree == NULL)
	   cout << "IEPA!" << endl;
	 // Supposed there must be one child whose state is progress
*/
  }
  return (Best_Tree) ;

}





// Selects the best tree according to the split selection criterion
inline or_tree *and_tree::Select_Best_OR_For_Split(const options &Options) {
  
  if (Children.size() == 0)
    return NULL;
  else {
    if (Options.SSC == LEFT_FIRST) { // Split_Optimality has not been set. We use the left first.
      return (or_tree *)Children.front();
    }
    else { // Split_Optimality has been set. We use it to select the best child
      or_tree *OT=NULL, *Best_Tree=NULL;
      float Best_Optimality= -1000000;   // SE TINDRIA QUE FICAR UN NUMERO NEGATIU GRAN. MIRAR LIMITS

      if (Options.SSC == C45) {
	// Compute the mean of GAIN.
	float Sum= 0.0;
	list<tree *>::iterator It, ItEnd;
	ItEnd= Children.end();
	for(It= Children.begin(); It != ItEnd; It++) {
	  or_tree *Child= (or_tree *)*It;
#ifdef TREE_DEBUG
	  if (Child == NULL)
	    cout << "CHILD NULL EN and_tree::Select_Best_OR" << endl;
#endif
	  Sum+= Child->Crit.Gain.GainET;
	}
	float MeanGain = Sum / (float)Children.size();
	//  From the ones higher than the mean, select the best gain_ratio
	ItEnd= Children.end();
	It= Children.begin();
	for(; It != ItEnd; It++) {
	  or_tree *Child= (or_tree *)*It;
	  float ActualOp= Child->Split_Optimality;
	  if ((ActualOp > Best_Optimality) && (Child->Crit.Gain.GainET >= MeanGain)) {
	    Best_Optimality = ActualOp;
	    Best_Tree = Child;
	  }
	}
//	cout<<OT->Num;
     if (Best_Tree==NULL) { // This is because the Mean can lose precision and all the GainEt could be less than the mean
       ItEnd= Children.end();
	   It= Children.begin();
	   Best_Tree= (or_tree *)*It;
	   for(; It != ItEnd; It++) {
	     or_tree *Child= (or_tree *)*It;
	     float ActualOp= Child->Split_Optimality;
	     if ((ActualOp > Best_Optimality) ) {
	      Best_Optimality = ActualOp;
	      Best_Tree = Child;
	    }
	  }
	}
	OT = Best_Tree;
     }
      else {  // ANY OTHER CRITERION
	OT= (or_tree *) *max_element(Children.begin(), Children.end(), Compare_Split_Optimality);
	
#ifdef TREE_DEBUG
	
	list<tree *>::iterator It, ItEnd;
	ItEnd= Children.end();
	for(It= Children.begin(); It != ItEnd; It++) {
	  or_tree *Child= (or_tree *)*It;
	  
	  if (Child == NULL)
	    cout << "CHILD NULL EN and_tree::Select_Best_OR" << endl;
	  
	  float ActualOp= Child->Split_Optimality;
	  if (ActualOp> Best_Optimality) {
	    Best_Optimality = ActualOp;
	    Best_Tree = Child;
	  }
	}
	
	if (OT->Num != Best_Tree->Num)
	  cout << "LES DOS FORMES DE CALCULAR EL MAX NO CONCIDIXEN" << endl;
	
	// OT= Best_Tree;
#endif
      }
      return (OT) ;
    }
  }
}




#endif
